/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**
 * \file 
 */

// config
#if defined( HAVE_CONFIG_H )
#    include "config.h"
#endif



// std c
#include <unistd.h>
#include <stdio.h>
#include <sys/poll.h>

// this header
#include "IEEE1394IO.hh"

// std c++
#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <string>
using std::string;

#include <libavc1394/rom1394.h>
#include <libavc1394/avc1394.h>
#include <libavc1394/avc1394_vcr.h>

// PIAVE
#include "avccontroller.hh"
#include "frame.hh"
#include "bufferpool.hh"
#include "pluginmgr.hh"

using namespace PIAVE;


/**
 *    FrameQ
 */
Raw1394Reader::FrameReciever::FrameReciever( int size, int grow ) : 
  _pool( 0 ),
  _head( 0 ),
  _tail( 0 ),
  _size( size ),
  _grow( grow ),
  _on( false )
{
    _pool = new Buffer*[ size ];
    pthread_mutex_init( &lock, NULL );
    pthread_cond_init( &notempty, NULL );
}

void
Raw1394Reader::FrameReciever::grow()
{
    INFO( "FrameReciever::grow" );

    /*  allocate a new buffer */
    int new_size = _size + _grow;
    Buffer ** new_pool = new Buffer*[ new_size ];

    /* copy over the old pointers */
    int j = 0;
    for ( int i=_tail; i<_size; i++ ) {
        new_pool[j++] = _pool[i];
    }
    int h = _head;
    inc(h);
    for ( int i=0; i<=h; i++ ) {
        new_pool[j++] = _pool[i];
    }
    
    /* delete the old buffers */
    delete [] _pool;

    /* set new bounds */
    _tail = 0;
    _head = _size-1;
    _pool = new_pool;
    _size = new_size;
}

void
Raw1394Reader::FrameReciever::kick()
{
    pthread_mutex_lock( &lock );
    pthread_cond_signal( &notempty );
    pthread_mutex_unlock( &lock );
}

void
Raw1394Reader::FrameReciever::push( Buffer * b )
{
    pthread_mutex_lock( &lock );

    if ( isFull() ) {
        grow();
    }

    inc( _head );
    _pool[ _head ] = b;

    /* Signal that the buffer is now not empty */
    pthread_cond_signal( &notempty );

    /* done, unlock */
    pthread_mutex_unlock( &lock );
}

Buffer *
Raw1394Reader::FrameReciever::pop( AccessMode a )
{
    // Wait until queue is not empty
    if ( isEmpty() ) {
        if ( a == non_block ) {
            return 0;
        } else {
            pthread_mutex_lock( &lock );
            pthread_cond_wait( &notempty, &lock );
        }
    }


    Buffer * ret = 0;
    if ( !isEmpty() ) { 
        // pop a cmnd from the queue
        ret = _pool[ _tail ];
        inc( _tail );
    }

    // unlock queues
    pthread_mutex_unlock( &lock );
    
    return ret;
}

void
Raw1394Reader::FrameReciever::clean( )
{
    while ( !isEmpty() ) {
        Buffer * b = pop( non_block );
        b->dropRef();
    }
}

Raw1394Reader::FrameReciever::~FrameReciever()
{
    for ( int i=0; i<_size; i++ ) {
        delete [] _pool[i];
    }    
    delete [] _pool;
}


/**
 *    Raw1394Reader
 */
Raw1394Reader * Raw1394Reader::_singleton = 0;

Raw1394Reader::Raw1394Reader() :
  InAVStreamIFace(),
  _currentFrame( 0 ),
  _currentFrameSize( 0 ),
  _currentFrameExpectedSize( 0 ),
  _isCapturing( false ),
  _dvDecoder( 0 )
{
    /**
     *    init the IEEE1394 AVC device
     */
    if ( AVCController::inst()->getState() == AVCController::error ) {
        throw stream_error( "raw1394 system error" );
    }
    if ( raw1394_set_iso_handler( AVCController::inst()->getRaw1394Handle(), 
                                  AVC1394_CHANNEL_BROADCAST, 
                                  _iso_handler )<0) 
    {
        WARN( "raw1394_set_iso_handler failed" );
    }

    /**
     *    get the libdv  DV decoder
     */
    PropertyNode * dvDecoderProp = new PropertyNode( "decoder" );
    PropertyNode * f = dvDecoderProp->addDaughter("format");
    f->addAttr("fourcc", "dvsd" );

    Plugin * dvDecoderPlugin = 
        PluginMgr::find( dvDecoderProp, PLUGIN_VIDEO_CODEC );
    if ( !dvDecoderPlugin ) {
        INFO( "NO decoder plugin for DV stream! " );
        throw stream_error( "no codec found" );
    }
    _dvDecoder = (AVDecoderIFace*) dvDecoderPlugin->newElement();

    return;
}

Raw1394Reader::~Raw1394Reader() 
{
}

void
Raw1394Reader::startCapture()
{
    _isCapturing = true;
    pthread_create( &_raw1394_poll_thread, NULL, _raw1394_poll, (void*)this );
}

void
Raw1394Reader::stopCapture()
{
    _isCapturing = false;
    pthread_join( _raw1394_poll_thread, 0 );
}

int
Raw1394Reader::isoHandler( size_t length, quadlet_t * data_q )
{
    if ( length <= 16 ) return 0;
    static int totalframes = 0;
    unsigned char *data_c = (unsigned char*) &data_q[3];

    /* is this the start of a new frame ? */
    if (data_c[0] == 0x1f && data_c[1] == 0x07) {
        raw1394_errcode_t e = 
            raw1394_get_errcode( AVCController::inst()->getRaw1394Handle() );
        bool pal         = ( data_c[ 3 ] & 0x80 );
        int thisFrameExpectedSize = 0;
        if ( pal ) {
            thisFrameExpectedSize = 144000;
        } else {
            thisFrameExpectedSize = 120000;
        }

        if ( _currentFrameSize == _currentFrameExpectedSize ) {
            if ( _currentFrame ) {
                totalframes++;
//                INFO( "reveived complete frame"<<PRINTV( totalframes ) );
                /*  push the complete frame into all running recievers */
                std::vector< FrameReciever* >::iterator c;
                for ( c =  _consumers.begin();
                      c != _consumers.end();
                      ++c )
                {
                    if ( (*c)->isOn() ) {
                        (*c)->push( _currentFrame );
                        /* the recievers take respinsibility from now */
                        _currentFrame->incRef();
                    }
                }
                /* drop ref. 
                 * In case there are recievers, this object is
                 * not the only referencer 
                 */
                _currentFrame->dropRef();
            }
            /* now get a new buffer */
            _currentFrame = BufferPool::getNewBuffer( thisFrameExpectedSize );
        } else if ( _currentFrameSize != 0 ) {
            if ( _currentFrameSize < _currentFrameExpectedSize ) {
                WARN( "received incomplete frame, not enough data"
                      <<_currentFrameSize );
            }
            if ( _currentFrameSize > _currentFrameExpectedSize ) {
                WARN( "received incomplete frame, too much data"
                      <<_currentFrameSize );
            }
        }

        _currentFrameSize = 0;
        _currentFrameExpectedSize = thisFrameExpectedSize;
    }

    /*
     *   have we seen the start of the first full frame?
     */
    if ( !_currentFrame ) return 0;

    int section_type = data_c[0] >> 5; // section type is in bits 5-7
    int dif_sequence = data_c[1] >> 4; // dif sequence number is in bits 4-7
    int dif_block    = data_c[2];
    uint8_t * target = _currentFrame->buf() + dif_sequence * 150 * 80;
    bool section_ok  = true;

    switch ( section_type ) {
      case 0: // 1 header block
          break;
      case 1: // 2 subcode blocks
          target += (1 + dif_block) * 80;
          break;
      case 2: // 3 VAUX blocks
          target += (3 + dif_block) * 80;
          break;
      case 3: // 9 audio blocks
          target += (6 + dif_block * 16) * 80;
          break;
      case 4: // 135 video blocks
          target += (7 + (dif_block / 15) + dif_block) * 80;
          break;
      default:
          WARN( "??? "<<PRINTV(section_type));
          section_ok = false;
    }

    if ( section_ok ) {
        memcpy( target, data_c, 480);
        _currentFrameSize += 480;
    }

    return 0;
}



int 
Raw1394Reader::_iso_handler( raw1394handle_t   handle,
                             int               channel,
                             size_t            length,
                             quadlet_t       * data_q )

{
    if ( channel != AVC1394_CHANNEL_BROADCAST ) {
        WARN( PRINTV( channel ) );
        return 0;
    }
    return Raw1394Reader::inst()->isoHandler( length, data_q );
}


void * 
Raw1394Reader::_raw1394_poll( void * s )
{
    Raw1394Reader * self = static_cast< Raw1394Reader * >( s );

    if ( raw1394_start_iso_rcv( AVCController::inst()->getRaw1394Handle(), 
                                AVC1394_CHANNEL_BROADCAST ) < 0 ) 
    {
        WARN( "raw1394 - couldn't start iso receive" );
    }

    INFO( "start capture" );
    
    /*  - wait for something to happen 
     *    and iterate the raw1394 interface
     */
    struct pollfd pf;
    pf.fd = raw1394_get_fd( AVCController::inst()->getRaw1394Handle() );
    pf.events = POLLIN | POLLPRI;
    while ( self->isCapturing() ) {
        poll( &pf, 1, -1);
        raw1394_loop_iterate( AVCController::inst()->getRaw1394Handle() );
    }

    /* in case we are in the process of recieving a frame,
     * release the half recieved frame fragment 
     */
    if ( self->_currentFrame ) self->_currentFrame->dropRef();

    if (raw1394_stop_iso_rcv( AVCController::inst()->getRaw1394Handle(), 
                              AVC1394_CHANNEL_BROADCAST ) < 0) {
        WARN("raw1394 - couldn't stop iso receive");
    }
    INFO( "stop capture" );

    /* in case there are still consumer waiting for input */
    std::vector< FrameReciever* >::iterator c;
    for ( c =  self->_consumers.begin();
          c != self->_consumers.end();
          ++c )
    {
        (*c)->switchOff();
        (*c)->kick();
    }
    
    return 0;
    
}

void 
Raw1394Reader::decodeMetaData( Frame & f )
{
    _dvDecoder->decodeMetaData( f, _frameProperties );
    f.setProperties( _frameProperties );
}

void 
Raw1394Reader::decodeVideo( Frame & f, uint8_t * dest )
{
    _dvDecoder->decodeVideo( f, dest );
}

void 
Raw1394Reader::decodeAudio( Frame & f, uint8_t * dest )
{
    _dvDecoder->decodeAudio( f, dest );
}

PropertyNode *
Raw1394Reader::getProperties( PropertyNode * p ) const
{
    return _dvDecoder->getProperties( p );
}

Frame
Raw1394Reader::getFrame( Time t )
{
    Frame f;
    ABORT( "???" );
    return f;
}

/*
 *
 *      
 *     Raw1394Consumer
 *
 *
 */

Raw1394Consumer::Raw1394Consumer() :
  _outAVStreamIFace( 0 )
{
    _frameQueue = new Raw1394Reader::FrameReciever();
    Raw1394Reader::inst()->addConsumer( _frameQueue );
}

Raw1394Consumer::~Raw1394Consumer()
{
    void * retval;
    pthread_join( _raw1394_consumer_thread, &retval );
}

void
Raw1394Consumer::switchOn()
{
    _frameQueue->switchOn();
    pthread_create( &_raw1394_consumer_thread, NULL, _raw1394_consume, (void*)this );
}

void
Raw1394Consumer::switchOff()
{
    _frameQueue->switchOff();
    _frameQueue->kick();
    pthread_join( _raw1394_consumer_thread, 0 );
}


void * 
Raw1394Consumer::_raw1394_consume( void * s )
{
    Raw1394Consumer * self = static_cast< Raw1394Consumer * >( s );
    self->raw1394Consume();
}

void 
Raw1394Consumer::raw1394Consume()
{
    static int seen_frames = 0;
    string isNewRec = "";
    while ( _frameQueue->isOn() )
    {
        Frame f;
        Buffer * buffer = _frameQueue->pop( Raw1394Reader::FrameReciever::block );
        if ( buffer ) {
            f.setRawVBuf( buffer ); 
            f.setAVContext( Raw1394Reader::inst() );
            PropertyNode * p = f.getMetaData();
            if ( p ) {
                p->getAttr( "isNewRecording", isNewRec );
                if ( isNewRec == "t" ) {
                    INFO( "found NEW RECORDING" );
                }
            } else {
                INFO( "no metadata" );
            }
            seen_frames++;
//            INFO( "Consumer got buffer"<<PRINTV(seen_frames) );
            if ( _outAVStreamIFace ) {
                _outAVStreamIFace->putFrame( f );
            } else {
                buffer->dropRef();
            }
        }
    }
    _frameQueue->clean();
}

