/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**
 * \file player.hh implements a simple main loop and all infrastructure to try to
 * run this main loop in the requested speed.
 */

#if ! defined( IEEE1394IO_HH )
#define IEEE1394IO_HH

// stdc
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/time.h>

// stdc++
#include <string>

// PIAVE
#include "piave_base.hh"
#include "mediaelements.hh"

// IEEE1394
#include <libraw1394/raw1394.h>
#include <libraw1394/csr.h>
#include "dv1394.h"
// channel is 63 (broadcast) for video
#define  AVC1394_CHANNEL_BROADCAST 63

namespace PIAVE {

    class AVCController;

    class Raw1394Reader : public InAVStreamIFace
    {
      public:
        /**
         *   FIMXE this ringbuffer implementation could maybe merged with the
         *   Raw1394Consumer
         */
        class FrameReciever
        {
          public:
            enum AccessMode { block, non_block };

            FrameReciever( int size = 250, int grow = 50 );
            ~FrameReciever();
            void        kick();
            void        push( Buffer * b );
            Buffer    * pop( AccessMode a );
            Buffer    * get( int i ) const { return _pool[i]; }
            int         getHeadIDX() { return _head; }
            
            bool isEmpty() const { return _head==_tail; }
            bool isFull()  const { int h=_head; inc(h); return h==_tail; }
            void clean();

            void inc( int & i ) const { if ( ++i == _size ) i = 0; }

            void switchOn()  { _on=true;  }
            void switchOff() { _on=false; }
            bool isOn() const { return _on; }

         private:
            void grow();

          private:
            Buffer         ** _pool;      /* the frames */
            int               _head;      /* head of ringbuffer */     
            int               _tail;      /* tail of ring */
            int               _size;      /* current size */
            int               _grow;      /* when buffer full, grow by grow */

            bool              _on;

            pthread_mutex_t   lock;	      /* mutex ensuring exclusive access to buffer */
            pthread_cond_t    notempty;   /* signaled when buffer is not empty */
        };

      public:
        static Raw1394Reader * inst() { 
            if (!_singleton) _singleton = new Raw1394Reader();
            return _singleton; 
        }

        bool         isCapturing() const { return _isCapturing; }

        void startCapture();
        void stopCapture();

        /**
         *  this is called by the raw1394 subsystem to handle 
         *  the data received from the cam
         */
        int isoHandler( size_t, quadlet_t* );


        /**
         *  override the AVStreamIFace
         */
        virtual void decodeMetaData( Frame & f );
        virtual void decodeVideo( Frame & f, uint8_t * dest = 0 );
        virtual void decodeAudio( Frame & f, uint8_t * dest = 0 );
        virtual Frame getFrame( Time t );
        virtual PropertyNode * getProperties( PropertyNode * p=0 ) const;

        /* consumers */
        int     addConsumer( FrameReciever * q ) { _consumers.push_back( q ); }
        
      private:
        Raw1394Reader();
        ~Raw1394Reader();

        /**
         *  a small wrapper around the member function
         */
        static int _iso_handler( raw1394handle_t, int, size_t, quadlet_t* );

      private:
        static Raw1394Reader      * _singleton;
  
        Buffer          * _currentFrame;
        int               _currentFrameSize;
        int               _currentFrameExpectedSize;

        bool              _isCapturing;

        /* the consumers that want to process data */
        std::vector< FrameReciever* >   _consumers;
 
        static void     * _raw1394_poll( void * );
        pthread_t         _raw1394_poll_thread;

        AVDecoderIFace  * _dvDecoder;
    };


    class Raw1394Consumer
    {
      public:
        Raw1394Consumer();
        ~Raw1394Consumer();

//         void addOutStream( OutAVStreamIFace * ostream );
//         void addOutStream( OutAudioStreamIFace * ostream );
        void setOutStream( OutAVStreamIFace * ost ) { _outAVStreamIFace = ost; }

        void switchOn();
        void switchOff();
        void kick() { _frameQueue->kick(); }

      private:
//         std::vector<OutAVStreamIFace *>      _outAVStreamIFaces;
//         std::vector<OutAudioStreamIFace *>   _outAudioStreamIFaces;

        OutAVStreamIFace *      _outAVStreamIFace;
        Raw1394Reader::FrameReciever  * _frameQueue;

        void           raw1394Consume();

        static void * _raw1394_consume( void * );
        pthread_t     _raw1394_consumer_thread;

    };
}

#endif
