/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**
 * \file 
 */

// config
#if defined( HAVE_CONFIG_H )
#    include "config.h"
#endif



// std c
#include <unistd.h>
#include <stdio.h>

// this header
#include "avccontroller.hh"

// std c++
#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <string>
using std::string;

#include <libavc1394/rom1394.h>
#include <libavc1394/avc1394.h>
#include <libavc1394/avc1394_vcr.h>

// PIAVE
#include "frame.hh"
#include "server.hh"

using namespace PIAVE;


/*
 *  \class
 *
 */
AVCController * AVCController::_singleton = 0;

AVCController::AVCController() :
  _controllerState( uninitialized ),
  _raw1394_handle( 0 ),
  _nports( -1 ),
  _port( 0 )

{
    INFO( "AVCController::AVCController()" );

    struct raw1394_portinfo port_info[16];
    
    try {
		
        _raw1394_handle = raw1394_new_handle();

        if ( 0 ==  _raw1394_handle ) {
            WARN( "couldn't get raw1394 handle" );
            setControllerState( error );
            return;
        }
        _nports = raw1394_get_port_info( _raw1394_handle, port_info, 16 );
        INFO( PRINTV( _nports ) );

        for ( int i=0; i<_nports; i++ ) {
            INFO( "Port #"<<i<<" nodes="<<port_info[i].nodes
                  <<" name=\""<<port_info[i].name<<"\"" );
        }

        if ( _nports>0 ) {
            setControllerState( not_connected );
        } else {
            setControllerState( error );
        }

    } catch( string e ) {
        if ( _raw1394_handle ) {
            raw1394_destroy_handle( _raw1394_handle );
        }
        _raw1394_handle = 0;
        INFO( "Caught exception: "<<e );
        setControllerState( error );
    };

    if ( _controllerState == error ) {
        throw stream_error( "raw1394 system error" );
    }

    return;
        
}
    
void
AVCController::init()
{
    if ( _controllerState == not_connected ) {
        raw1394_set_port( _raw1394_handle, _port );
        raw1394_set_bus_reset_handler( _raw1394_handle, reset_handler );
        _avc_node = findAVCNode( -1 );
        if ( _avc_node >= 0 ) {
            setControllerState( connected );
        } else {
            setControllerState( not_connected );
        }            
    }
}


int
AVCController::findAVCNode( int phyID )
{
    if ( !_raw1394_handle ) return -1;

 	int avc_node = -1;
    rom1394_directory rom1394_dir;

//   	pthread_mutex_lock(&avc_mutex);
    int nnodes = raw1394_get_nodecount( _raw1394_handle );
    INFO( PRINTV( nnodes ));
	
    if ( phyID >= 0 && phyID < nnodes) {
        rom1394_get_directory( _raw1394_handle, phyID, &rom1394_dir);
        if ( rom1394_get_node_type( &rom1394_dir) == ROM1394_NODE_TYPE_AVC ) {
            if ( avc1394_check_subunit_type( _raw1394_handle, phyID, 
                                             AVC1394_SUBUNIT_TYPE_VCR ) ){
                avc_node = phyID;
                INFO( PRINTV( rom1394_dir.label ) );
            }
        }
        rom1394_free_directory( &rom1394_dir );
    }
    INFO( PRINTV( avc_node ));
		
    for ( int node = 0; 
          node<nnodes; 
          node++ ) 
    {
        rom1394_get_directory( _raw1394_handle, node, &rom1394_dir );
        string type;
        switch(rom1394_get_node_type(&rom1394_dir)) {
          case ROM1394_NODE_TYPE_UNKNOWN:
              type = "unknown";
              break;
          case ROM1394_NODE_TYPE_DC:
              type = "DC";
              break;
          case ROM1394_NODE_TYPE_AVC:
              type = "AVC";
              if ( avc1394_check_subunit_type( _raw1394_handle, node, 
                                               AVC1394_SUBUNIT_TYPE_VCR) )
              {
                  avc_node = node;
                  type = "AVC (subtype=VCR)";
              }
              break;
          case ROM1394_NODE_TYPE_SBP2:
              type = "SBP2";
              break;
          case ROM1394_NODE_TYPE_CPU:
              type = "CPU";
              break;
        }
        INFO( "Found node: "<< rom1394_dir.label<<"  type="<<type );
        rom1394_free_directory( &rom1394_dir);
    }
    if ( avc_node >= 0 ) {
        INFO( PRINTV( avc_node ));
    } else {
        WARN( "Could not find any ieee1394 VCR!" );
    } 
    
//    pthread_mutex_unlock(&avc_mutex);
    return avc_node;
}

quadlet_t
AVCController::getCamState()
{
    return avc1394_vcr_status( _raw1394_handle,_avc_node );
}

string
AVCController::getCamStateName()
{
    /* Get a textual description of the status */
    quadlet_t state = getCamState();
    return string( avc1394_vcr_decode_status( state ) );
}

string
AVCController::getTimeCode()
{
    return string( avc1394_vcr_get_timecode( _raw1394_handle, _avc_node ) );
}


void             
AVCController::vcrStop()
{
    if ( _controllerState != connected ) return;
    avc1394_vcr_stop( _raw1394_handle, _avc_node );
}

void             
AVCController::vcrPause()
{
    if ( _controllerState != connected ) return;
    avc1394_vcr_pause( _raw1394_handle, _avc_node );
}

void             
AVCController::vcrPlay()
{
    if ( _controllerState != connected ) return;
    if ( avc1394_vcr_is_playing( _raw1394_handle, _avc_node) != 
         AVC1394_VCR_OPERAND_PLAY_FORWARD) 
    {
        avc1394_vcr_play( _raw1394_handle, _avc_node );
    }
}

void             
AVCController::vcrPlaySlow()
{
    if ( _controllerState != connected ) return;
    avc1394_vcr_play( _raw1394_handle, _avc_node );
    avc1394_vcr_play( _raw1394_handle, _avc_node );
}

void             
AVCController::vcrReverse()
{
    if ( _controllerState != connected ) return;
    avc1394_vcr_reverse( _raw1394_handle, _avc_node );
}

void             
AVCController::vcrReverseSlow()
{
    if ( _controllerState != connected ) return;
    avc1394_vcr_reverse( _raw1394_handle, _avc_node );
    avc1394_vcr_reverse( _raw1394_handle, _avc_node );
}

void             
AVCController::vcrTrickPlay( int speed )
{
    if ( _controllerState != connected ) return;
    avc1394_vcr_trick_play( _raw1394_handle, _avc_node, speed );
}

void             
AVCController::vcrRewind()
{
    if ( _controllerState != connected ) return;
    avc1394_vcr_rewind( _raw1394_handle, _avc_node );
}

void             
AVCController::vcrForward()
{
    if ( _controllerState != connected ) return;
    avc1394_vcr_forward( _raw1394_handle, _avc_node );
}

void             
AVCController::vcrNext()
{
    if ( _controllerState != connected ) return;
    avc1394_vcr_next( _raw1394_handle, _avc_node );
}

void             
AVCController::vcrPrevious()
{
    if ( _controllerState != connected ) return;
    avc1394_vcr_previous( _raw1394_handle, _avc_node );
}

int 
AVCController::reset_handler( raw1394handle_t handle, unsigned int generation )
{
    WARN( "reset handler called" );
    AVCController::inst()->setControllerState( not_connected );
    AVCController::inst()->init();
    return 0;
}

    
