/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/** 
 * \file frame.cc implements frame.hh
 *
 * based on:
 *
 * frame.cc -- utilities for process digital video frames
 * Copyright (C) 2000 Arne Schirmacher <arne@schirmacher.de>
 */

// config
#if defined( HAVE_CONFIG_H )
#    include "config.h"
#endif



// C includes
#include <pthread.h>

// std c++
#include <string>
using std::string;
#include <iostream>
using std::cout;
using std::cerr;
using std::endl;

// this header
#include "frame.hh"

// PIAVE
#include "piave_base.hh"
#include "avstream.hh"
#include "bufferpool.hh"

using namespace PIAVE;

/*
 *
 *      VideoBuffer
 *
 */

VideoBuffer::VideoBuffer( int width, int height, bool allocate )  :
  _width( width ),
  _height( height ),
  _lineStride( width * 2 ),
  _sizeBytes( width * height * 2 ),
  _pixels( 0 ),       
  _alphaPixels( 0 ),       
  _is_locked( false ),
  _is_allocated( allocate )
{
    // \todo FIXME rethink allocation strategy
    if ( allocate ) 
    {
        // \todo FIXME  factor * 2 is format dependent
        _pixels = BufferPool::getNewBuffer( width * height * 2 );
    }
};


VideoBuffer::~VideoBuffer()
{
    if ( _is_allocated ) 
    {
        // \todo FIXME rething alocation strategy
    }
};

const VideoBuffer & 
VideoBuffer::operator=( const VideoBuffer & o )
{
    ABORT( "FIXME" );
    return (*this);
}


void 
VideoBuffer::dropRef() 
{
    if ( _pixels ) _pixels->dropRef(); 
    if ( _alphaPixels ) _alphaPixels->dropRef(); 
}


/*
 *
 *      AudioBuffer
 *
 */
AudioBuffer::AudioBuffer() :
  _nChannels( -1 ),
  _isInterleaved( false ),
  _nSamples( -1 )
{
    memset( _channels, 0, sizeof( _channels ) );
}

AudioBuffer::~AudioBuffer()
{
    // nothing to do
}

Buffer *
AudioBuffer::interleave()
{
    if ( _isInterleaved ) return _channels[0];
    if ( _nChannels < 2 ) return _channels[0];

    /*  get a buffer for the interleaved audio data */
    Buffer * buf = BufferPool::getNewBuffer( 
        sizeof(int16_t) * _nSamples * _nChannels );
    int16_t * dest = (int16_t*)buf->buf();

    /*  do the actual interleave */
    int j = 0;
    for( int i=0; i < _nSamples; i++) {
        for( int ch=0; ch < _nChannels; ch++) {
            dest[ j++ ] = ((int16_t*)_channels[ch]->buf())[i];
        }
    }

    /*  drop the non interleaved buffers */
    for ( int i=0; i<MAX_CHANNELS; i++ ) {
        if ( _channels[i] ) {
            _channels[i]->dropRef();
            _channels[i] = 0;
        }
    }
    _isInterleaved = true;
    _channels[0] = buf;
    return buf;
}


void 
AudioBuffer::dropRef()
{
    for ( int i=0; i<MAX_CHANNELS; i++ ) {
        if ( _channels[i] ) {
            _channels[i]->dropRef();
            INFO( PRINTV( _channels[i]->getRefCount() ));
        }
    }
}



/*
 *
 *      Frame
 *
 */


Frame::Frame()
  : _vBuf( 0 ), 
    _aBuf( 0 ),
    _rawVBuf( 0 ),
    _rawABuf( 0 ),
    _videoIsDecoded( false ),
    _audioIsDecoded( false ),
    _avContext( 0 ),
    _audioContext( 0 ),
    _properties( 0 )
{};

Frame::Frame( const Frame & o )
  : _vBuf( o._vBuf ), 
    _aBuf( o._aBuf ),
    _rawVBuf( o._rawVBuf ),
    _rawABuf( o._rawABuf ),
    _pos( o._pos ),
    _length( o._length ),
    _videoIsDecoded( o._videoIsDecoded ),
    _audioIsDecoded( o._audioIsDecoded ),
    _avContext( o._avContext ),
    _audioContext( o._audioContext ),
    _properties( o._properties )
{
};

const Frame & 
Frame::copy( const Frame & o )
{
    _vBuf           = o._vBuf; 
    _aBuf           = o._aBuf;
    _rawVBuf        = o._rawVBuf; 
    _rawABuf        = o._rawABuf;
    _pos            = o._pos;
    _length         = o._length;
    _videoIsDecoded = o._videoIsDecoded;
    _audioIsDecoded = o._audioIsDecoded;
    _avContext      = o._avContext;
    _audioContext   = o._audioContext;
    _properties     = o._properties;
    return (*this);
}

const Frame & 
Frame::copyVideo( const Frame & o )
{
    _vBuf           = o._vBuf; 
    _rawVBuf        = o._rawVBuf; 
    ABORT_IF( _pos != o._pos, "FIXME" );
    ABORT_IF( _length != o._length, "FIXME" );
    _videoIsDecoded = o._videoIsDecoded;
    _avContext      = o._avContext;
// FIXME    _properties     = o._properties;
    return (*this);
}

const Frame & 
Frame::copyAudio( const Frame & o )
{
    _aBuf           = o._aBuf;
    _rawABuf        = o._rawABuf;
    ABORT_IF( _pos != o._pos, "FIXME" );
    ABORT_IF( _length != o._length, "FIXME"<<PRINTV(_length)<<PRINTV(o._length) );

    _audioIsDecoded = o._audioIsDecoded;
    _audioContext   = o._audioContext;
// FIXME    _properties     = o._properties;

    return (*this);
}

PropertyNode *
Frame::getProperties()
{
    return _properties;
}

PropertyNode *
Frame::getMetaData()
{
    if ( _properties==0 ) decodeMetaData();
    return _properties;
}



void 
Frame::decodeMetaData()
{
    if ( _avContext ) {
        _avContext->decodeMetaData( *this );
    } 
//  FIXME
//     if ( _audioContext ) 
//         _audioContext->decodeMetaData( *this );
}


void 
Frame::decodeVideo( uint8_t * dest ) 
{
  _avContext->decodeVideo( *this, dest );
}

void 
Frame::decodeAudio( uint8_t * dest ) 
{
  _audioContext->decodeAudio( *this, dest );
}



void
Frame::addAlpha()
{
    // FIXME ...
    return;
}


void
Frame::flattenAlpha()
{
    // FIXME
    return;
}

/**
 * if \a o is <b>not locked</b>, this._videoBuffer will just be set 
 * to \a o. If \a o <b>is locked</b>, a new VideoBuffer containing a
 * deep copy of \a o. If \a o does not have videobuffer, it will be created
 * from the DVFrame.
 */
void
Frame::makeWriteableVBuf()
{
    if ( _vBuf ) {
        if ( _vBuf->isLocked() ) {
	  ABORT( "FIXME" );
// VideoBuffer * v = _vBuf;
// _vBuf = new VideoBuffer( *v );
        }
    } else {
      decodeVideo();
    }
    
    return;
}

/*
 *
 */
AudioBuffer * 
Frame::makeABuf()
{
    if ( _audioIsDecoded ) return _aBuf;

    if ( _audioContext ) {
        _audioContext->decodeAudio( *this );
    } else if ( _avContext ) {
        _avContext->decodeAudio( *this );
    }
    return _aBuf;
}



/**
 * this method will create a new or transform an existing VideoBuffer
 * to the requested format.
 */
VideoBuffer *
Frame::makeVBuf() 
{
    if ( _videoIsDecoded ) return _vBuf;

    _avContext->decodeVideo( *this );

    return _vBuf;
}


void 
Frame::dropRefs()
{
    if ( _rawABuf ) _rawABuf->dropRef();
    if ( _rawVBuf ) _rawVBuf->dropRef();
    if ( _aBuf ) _aBuf->dropRef();
    if ( _vBuf ) _vBuf->dropRef();
}

void 
Frame::dropVRefs()
{
    if ( _rawVBuf ) _rawVBuf->dropRef();
    if ( _vBuf ) _vBuf->dropRef();
}
void 
Frame::dropARefs()
{
    if ( _rawABuf ) _rawABuf->dropRef();
    if ( _aBuf ) _aBuf->dropRef();
}
