/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, Rolf.Dubitzky@gmx.net
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/**
 * \file server.cc implaments server.hh
 */
// config
#if defined( HAVE_CONFIG_H )
#    include "config.h"
#endif



#include "kdenserver.hh"

// std c
#include <errno.h>
#include <unistd.h>

// std c++
#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <string>
using std::string;
#include <sstream>
#include <algorithm>

//piave
#include "libpiave/avstream.hh"
#include "libpiave/mediaelements.hh"
#include "libpiave/avstream.hh"
#include "libpiave/streammgr.hh"
#include "libpiave/pluginmgr.hh"
#include "libpiave/scenebuilder.hh"

using namespace PIAVE;

// FIXME ... this must be moved to the SDL plugin
// SDL
#include <SDL/SDL.h>


// libxml
#include <libxml/parser.h>

// For translations

#ifdef ENABLE_NLS
#  include <libintl.h>
#  undef _
#  define _(String) dgettext (PACKAGE, String)
#  ifdef gettext_noop
#    define N_(String) gettext_noop (String)
#  else
#    define N_(String) (String)
#  endif
#else
#  if ! defined(textdomain)
#    define textdomain(String) (String)
#  endif
#  if ! defined(gettext)
#    define gettext(String) (String)
#  endif
#  if ! defined(dgettext)
#    define dgettext(Domain,Message) (Message)
#  endif
#  if ! defined(dcgettext)
#    define dcgettext(Domain,Message,Type) (Message)
#  endif
#  if ! defined(bindtextdomain)
#    define bindtextdomain(Domain,Directory) (Domain)
#  endif
#  if ! defined(_)
#    define _(String) (String)
#  endif
#  if ! defined(N_)
#    define N_(String) (String)
#  endif
#endif

/*
 * \class KdenLiveServer::CommandQueue
 *
 *
 */
void
KdenLiveServer::CommandQueue::kick()
{
    pthread_mutex_lock( &lock );
    pthread_cond_signal( &notempty );
    pthread_cond_signal( &notfull );
    pthread_mutex_unlock( &lock );
}

void
KdenLiveServer::CommandQueue::push( const VEMLCmnd * c )
{
    pthread_mutex_lock( &lock );
    /* Wait until buffer is not full */
    while ( dropQ.isFull() || stdQ.isFull() ) {
        WARN( "queue is full!! waiting for queue >not full<" );
        pthread_cond_wait( &notfull, &lock);
    }
    
    switch ( c->type() ) {
      case VEMLCmnd::stdQ:
          stdQ.push( c );
          break;
      case VEMLCmnd::dropQ:
          dropQ.remove( c );
          dropQ.push( c );
          break;
      default:
          WARN( "unknown commandtype "<<PRINTV(c->type()));
    }
    
    /* Signal that the buffer is now not empty */
    pthread_cond_signal( &notempty );
    pthread_mutex_unlock( &lock );
}

const VEMLCmnd *
KdenLiveServer::CommandQueue::pop( AccessMode a )
{
    const VEMLCmnd * cmnd = 0;

    // Wait until queue is not empty
    if ( stdQ.isEmpty() && dropQ.isEmpty() ) {
        if ( a == non_block ) {
            return cmnd;
        } else {
            pthread_mutex_lock( &lock );
//            INFO( "waiting for dropQueue >notempty<" );
            pthread_cond_wait( &notempty, &lock );
        }
    }

    // pop a cmnd from the queue
    if ( !stdQ.isEmpty() ) {
        cmnd = stdQ.pop();
    } else if ( !dropQ.isEmpty() ) {
        cmnd = dropQ.pop();
    }

    // unlock queues
    pthread_cond_signal( &notfull );
    pthread_mutex_unlock( &lock );
    
    return cmnd;
}


/*
 *  \class KdenLiveServer
 */

KdenLiveServer::KdenLiveServer( unsigned short port ) 
  : Server( port ),
    _screenOut( 0 ),
    _audioOut( 0 ),
    _theStoryBoard( new StoryBoard( "MasterBoard" ) ),

    _allVEMLCommands(),
    _seek_pos( 0 )
{
    VEMLCmnd::setAllCmnds( &_allVEMLCommands );

    /** createVideoXWindow */
    VEMLCmnd create( "createVideoXWindow", 
                     &KdenLiveServer::veml_createVideoXWindow, 
                     VEMLCmnd::stdQ );
    create.addAttr( "show", "true" );
    _allVEMLCommands[ "createVideoXWindow" ] = create;

    /** closeVideoXWindow */
    _allVEMLCommands[ "closeVideoXWindow" ] = 
        VEMLCmnd( "closeVideoXWindow", &KdenLiveServer::veml_closeVideoXWindow );

    /** seek */
    VEMLCmnd seek( "seek", &KdenLiveServer::veml_seek, VEMLCmnd::dropQ );
    seek.addAttr( "time", "-666.0" );
    _allVEMLCommands[ "seek" ] = seek;

    /** stop */
    _allVEMLCommands[ "quit" ] = VEMLCmnd( "quit", &KdenLiveServer::veml_quit );

    /** stop */
    _allVEMLCommands[ "stop" ] = VEMLCmnd( "stop", &KdenLiveServer::veml_stop );

    /** playIntervall */
    VEMLCmnd play( "play", &KdenLiveServer::veml_play );
    play.addAttr( "start",   "-1.0" );
    play.addAttr( "stop",    "-2.0" );
    play.addAttr( "speed",   "1.0"  );
    _allVEMLCommands[ "play" ] = play;

    /** getFileProperties */
    VEMLCmnd gfp( "getFileProperties", &KdenLiveServer::veml_getFileProperties );
    gfp.addAttr( "filename", "" );
    _allVEMLCommands[ "getFileProperties" ] = gfp;

    /** render */
    VEMLCmnd render( "render", &KdenLiveServer::veml_render );
    render.addAttr( "filename", "" );
    _allVEMLCommands[ "render" ] = render;

    /** getCapabilities */
    _allVEMLCommands[ "getCapabilities" ] = VEMLCmnd( "getCapabilities", &KdenLiveServer::veml_getCapabilities );

    /** ping */
    _allVEMLCommands[ "ping" ] = VEMLCmnd( "ping", &KdenLiveServer::veml_ping );

    /** setSceneList */
    VEMLCmnd ssl( "setSceneList", &KdenLiveServer::veml_setSceneList );
    _allVEMLCommands[ "setSceneList" ] = ssl;

}

KdenLiveServer::~KdenLiveServer()
{
    PDELETE( _theStoryBoard );
    PDELETE( _screenOut );
    PDELETE( _audioOut );
}

/*  print a list of all implemented VEL commands
 */
void
KdenLiveServer::printAllVEML( std::ostream & o ) const 
{
    o << _("Known VEML commands (maybe not all implemented):")<<endl;
    std::map< const std::string, VEMLCmnd >::const_iterator cIter;
    for ( cIter = _allVEMLCommands.begin();
          cIter != _allVEMLCommands.end();
          ++cIter )
    {
        cIter->second.toXML( o );
    }
}

/*
 *  handle local commands
 *
 */

void 
KdenLiveServer::inc_seek_pos()
{
    if ( !_theStoryBoard ) return;
    _seek_pos = _seek_pos + Global::renderFmt.getFrameLength();
    if ( _seek_pos >= _theStoryBoard->stop() ) {
        _seek_pos = _theStoryBoard->stop() - 1;
    }
    return;
}

void 
KdenLiveServer::dec_seek_pos()
{
    if ( !_theStoryBoard ) return;
    _seek_pos = _seek_pos - Global::renderFmt.getFrameLength();
    if ( _seek_pos < 0 ) {
        _seek_pos = 0;
    }
    return;
}

void 
KdenLiveServer::push_cmnd( const string & c )
{
    std::map< const string, VEMLCmnd >::iterator cmndIter
        = _allVEMLCommands.find( c );
    VEMLCmnd * cmnd = new VEMLCmnd( cmndIter->second );
    _q.push( cmnd );
    return;
}


/*
 *
 *   thread entries
 *
 */

void * 
KdenLiveServer::_listen_for_commands( void * s )
{
    KdenLiveServer * self = static_cast< KdenLiveServer * >( s );
    self->listenForCommands();
    return 0;
}

void * 
KdenLiveServer::_listen_for_SDL_events( void * s )
{
    KdenLiveServer * self = static_cast< KdenLiveServer * >( s );
    self->listenForSDLEvents();
    return 0;
}

void * 
KdenLiveServer::_process_commands( void * s )
{
    KdenLiveServer * self = static_cast< KdenLiveServer * >( s );
    self->processCommands();
    return 0;
}



/*
 *   producer
 */
void 
KdenLiveServer::listenForCommands()
{
    while ( getState() == state_process_commands ) {

        string message = "";
        int nBytesRead = receiveMessage( message );
        if ( nBytesRead <= 0 ) continue;

        /*
         *   try parsing command 
         */
        xmlDocPtr cmndDoc = xmlParseMemory( message.c_str(), message.size() );
        if ( !cmndDoc ) {
            WARN( "couldn't parse command" );
            continue;
        }
        xmlNodePtr messageRoot = xmlDocGetRootElement( cmndDoc );
        if ( !messageRoot ) {
            xmlFreeDoc( cmndDoc );
            continue;
        }

        /*
         *   try to find the commands in the known _commands
         */
        for ( xmlNodePtr xmlCmndNode = messageRoot->xmlChildrenNode;
              xmlCmndNode != 0;
              xmlCmndNode = xmlCmndNode->next )
        {
            if ( !xmlStrcmp(xmlCmndNode->name, (const xmlChar *)"text") ) continue;
            INFO( "Recieved VEML command "<< xmlCmndNode->name );
            std::map< const string, VEMLCmnd >::iterator cmndIter
                = _allVEMLCommands.find( (const char*) xmlCmndNode->name );
            if ( cmndIter == _allVEMLCommands.end() ) {
                WARN( "unknown command <"<<xmlCmndNode->name<<">" );
            } else {
                // found cmnd
                VEMLCmnd * cmnd = new VEMLCmnd( cmndIter->second );
                cmnd->copyFromXMLNode( cmndDoc, xmlCmndNode );
                if ( cmnd->type() == VEMLCmnd::priority ) {
                    INFO( "executing priority cmnd "<<cmnd->getName() );
                    PropertyNode * ret = (this->*(cmnd->call()))( cmnd );
                    if ( ret ) {
                        sendMessage( ret->toXML() );
                        PDELETE( ret );
                    }
                    PDELETE( cmnd );
                } else {
                    _q.push( cmnd );
                }
            }
        }
        xmlFreeDoc( cmndDoc );
    }
    
    _q.kick();
    
    INFO( "stop listen thread" );
    return;
}

/*
 *   Wait for SDL Events
 *
 */

void
KdenLiveServer::listenForSDLEvents()
{
    if ( !_listenForSDLEvents ) return;
/*
 *   wait till we have a video window
 *   FIXME !!  this depends on SDL output plugin beeing
 *      used, which is bullshit. either make eventhandling an 
 *      plugin property, or ... err.. do it another way.. but 
 *      this is BS.
 */
    while ( _screenOut == 0 ) {
        sleep( 1 );
    }
    sleep( 2 );

    INFO( "starting listenForSDLEvents " );
    SDL_Event event;
    while ( getState() == state_process_commands  )
    {
        SDL_WaitEvent( &event );
        switch ( event.type ) {
//           case SDL_KEYDOWN:
//               INFO( "SDL_Event SDL_KEYDOWN" );
//               break;
//           case SDL_MOUSEMOTION:
//               break;
//           case SDL_MOUSEBUTTONDOWN:
//               switch ( event.button.button ) {
//                 case 4: 
//                     dec_seek_pos(); 
//                     break;
//                 case 5: 
//                     inc_seek_pos(); 
//                     break;
//                 default: ;
//               }
//               push_cmnd( "seek" );
//               sprintf( Global::cbuf, "<playing time=\"%f\" />",_seek_pos.sec() );
//               sendMessage( Global::cbuf );
//               break;
//           case SDL_MOUSEBUTTONUP:
//               break;
          case SDL_QUIT:
              INFO( "SDL_Event SDL_QUIT" );
              setState( state_shutting_down );
              _q.kick();
              break;
          case SDL_VIDEORESIZE:
              INFO( "SDL_Event SDL_VIDEORESIZE" );
              break;
          case SDL_VIDEOEXPOSE: 
              INFO( "SDL_Event SDL_VIDEOEXPOSE" );
              push_cmnd( "seek" );
              break;
          default: ;
        }
    }

    return;
}


/*
 *     consumer
 */
void
KdenLiveServer::processCommands()
{	

    /*
     *   wait for message
     */
    INFO( "starting process thread" );
    
    while ( getState() == state_process_commands  )
    {
        const VEMLCmnd * cmnd;
        if ( _player.getState() == Player::playing ) {
            _player.playNextFrame();
            cmnd = _q.pop( CommandQueue::non_block );
        } else {
            cmnd = _q.pop( CommandQueue::block );
        }

        if ( cmnd==0 ) continue;

        // execute cmnd 
        PropertyNode * ret = (this->*(cmnd->call()))( cmnd );
        // return reply if necessary
        if ( ret ) {
            sendMessage( ret->toXML() );
            PDELETE( ret );
        }

        PDELETE( cmnd );
    }

    INFO( "stop process thread" );
    return;
}




void
KdenLiveServer::startProcessCommandsLoop()
{
    /*
     *     start producer and consumer threads
     */
    INFO( "starting main server loop" );
    _main_thread = pthread_self();
    pthread_create( &_listen_thread, NULL, _listen_for_commands, (void*)this );
    pthread_create( &_listen_SDL_thread, NULL, _listen_for_SDL_events, (void*)this );
    pthread_create( &_process_thread, NULL, _process_commands, (void*)this );

    /*  wait for the children to return */
    void * retval;
    pthread_join( _listen_thread, &retval);
    pthread_join( _listen_SDL_thread, &retval);
    pthread_join( _process_thread, &retval);

    return;
}



/*
 *
 *        V E M L   commands
 *
 *
 */


PropertyNode *
KdenLiveServer::veml_createVideoXWindow( const VEMLCmnd * cmd )
{
    if ( _screenOut ) return 0;

    PropertyNode * ret = 0;
    string visual( "xv" );

//  FIXME  don't rely on defaults, require values
    double w = 720/2;
    double h = 576/2;
    double a = 4.0/3.0;
    double fps = 25.0;
    cmd->getAttr( "visual", visual );
    cmd->getAttr( "fps", fps );
    cmd->getAttr( "width", w );
    cmd->getAttr( "height", h );
    cmd->getAttr( "aspect", a );
    
    // FIXME  get aspect of screen //
    Global::renderFmt.fps            = (int) fps;    
    Global::renderFmt.width   = (int) w*2; 
    Global::renderFmt.height  = (int) h*2;
//    Global::displayFmt.width  = (int)( w );  //   /scale
//    Global::display_height = (int)( w / a );  // not yet
//    Global::displayFmt.height = (int)( h );  // / scale


    /*
     *   try to get screen output plugin
     */
    /* define capabilities to look for */
    PropertyNode * screen_caps = new PropertyNode( "outstream" );
    PropertyNode * monitor = screen_caps->addDaughter( "monitor" );
    PropertyNode * video = monitor->addDaughter( "video" );
    video->addAttr( "type", visual );
    /* try to get plugin */
    Plugin * p = PluginMgr::find( screen_caps );
    if ( !p ) {
        cout << "WARNING: no output plugin to handle output to screen" << endl;
    } else {
        video->addAttr( "width" , w );
        video->addAttr( "height" , h );
        PluginBase * b = p->newElement( video );
        if ( !b ) {
            cout << "WARNING : instantiating plugin "<<p->getName()<<endl;
        } else {
            _screenOut = static_cast<OutAVStreamIFace*>( b );
            if ( _screenOut ==0 ) {
                cout << "WARNING : wrong plugin type " << endl;
            } else {
                /* setup reply */
                ret = new PropertyNode( "reply" );
                ret->addAttr( "command", "createVideoXWindow" );
                ret->addAttr( "status", "ok" );
                ret->addDaughter( _screenOut->getProperties() );
            }
        }
    }
    PDELETE( screen_caps );

    /*
     *   try to get audio output plugin
     */
    /* define capabilities to look for */
    PropertyNode * speaker_caps = new PropertyNode( "outstream" );
    monitor = speaker_caps->addDaughter( "monitor" );
    PropertyNode * audio = monitor->addDaughter( "audio" );
    audio->addAttr( "type", "oss" );
    /* try to get plugin */
    p = PluginMgr::find( speaker_caps );
    if ( !p ) {
        cout << "WARNING: no output plugin to handle audio" << endl;
    } else {
        PluginBase * b = p->newElement();
        if ( !b ) {
            cout << "WARNING : instantiating plugin "<<p->getName()<<endl;
        } else {
            //  FIXME why doesn't this work ?? 
//                    voutstream = dynamic_cast<OutAVStreamIFace*>( b );
            _audioOut = static_cast<OutAudioStreamIFace*>( b );
            if ( _audioOut ==0 ) {
                cout << "WARNING : wrong plugin type " << endl;
            }
        }
    }
    PDELETE( speaker_caps );

    return ret;
}

PropertyNode * 
KdenLiveServer::veml_closeVideoXWindow( const VEMLCmnd * cmd )
{
    PDELETE( _screenOut );
    PDELETE( _audioOut );
    return 0;
}


PropertyNode * 
KdenLiveServer::veml_seek( const VEMLCmnd * cmd )
{
    if (!_theStoryBoard) return 0;
    double time;
    cmd->getAttr( "time", time );
    if ( time > -0.0001 ) {
        _seek_pos = Time( time );
    }
    Frame frame = _theStoryBoard->getFrame( _seek_pos );
    _screenOut->putFrame( frame );
    return 0;
}

PropertyNode * 
KdenLiveServer::veml_quit( const VEMLCmnd * cmd )
{
    INFO( "veml_quit" );
    setState( state_shutting_down );
    return 0;
}

PropertyNode * 
KdenLiveServer::veml_stop( const VEMLCmnd * cmd )
{
    _player.switchToStop();

    PropertyNode * ret = new PropertyNode( "reply" );
    ret->addAttr( "command", "stop" );
    ret->addAttr( "status", "ok" );

    return ret;
}

PropertyNode * 
KdenLiveServer::veml_play( const VEMLCmnd * cmd )
{
    PropertyNode * ret = 0;
    if (!_theStoryBoard) return ret;
    double speed, start, stop;
    cmd->getAttr( "speed", speed );
    cmd->getAttr( "start", start );
    cmd->getAttr( "stop", stop );
    if ( speed == 0.0 ) {
        _player.switchToStop();
    } else {
        _player.setVideoOut( _screenOut );
        _player.setAudioOut( _audioOut );
        _player.setStoryBoard( _theStoryBoard );
        if ( start >= -0.1 ) {
            _player.setStart( start );
        } else {
            _player.setStart( _seek_pos );
        }
        _player.setStop( stop );
        _player.setSpeed( speed );
        _player.switchToPlay();

        // setup reply
        PropertyNode * ret = new PropertyNode( "reply" );
        ret->addAttr( "command", "play" );
        ret->addAttr( "status", "ok" );

    }
    return ret;
}

PropertyNode * 
KdenLiveServer::veml_getFileProperties( const VEMLCmnd * cmd )
{
    /* parse command */
    string filename;
    cmd->getAttr( "filename", filename );

    // setup reply
    PropertyNode * ret = new PropertyNode( "reply" );
    ret->addAttr( "command", "getFileProperties" );
    ret->addAttr( "filename", filename );

    InAVFStream * s1 = 0;
    try {
        s1 = PIAVE::StreamMgr::getAVInput( filename );

    } catch ( const PIAVE::file_error & e ) {
        /* setup error reply */
        ret->addAttr( "status", "error" );
        PropertyNode * errmsg = new PropertyNode( "errmsg" );
        errmsg->setContent( e.what() );
        ret->addDaughter( errmsg );
        return ret;

    } catch ( const std::exception & e ) {
        /* setup error reply */
        ret->addAttr( "status", "error" );
        PropertyNode * errmsg = new PropertyNode( "errmsg" );
        errmsg->setContent( e.what() );
        ret->addDaughter( errmsg );
        return ret;
    } 

    /* setup success reply */
    ret->addAttr( "status", "ok" );
    ret->addAttr( "duration", s1->length().sec() );
    ret->addDaughter( s1->getProperties() );
    return ret;
}


PropertyNode * 
KdenLiveServer::veml_render( const VEMLCmnd * cmd )
{
    /* parse command */
    string filename;
    bool ok = cmd->getAttr( "filename", filename );
    if ( !ok || filename=="" ) {
        WARN( "missing filename" );
        return 0;
    }

    if ( !_theStoryBoard ) {
        WARN( "no storyboard" );
        return 0;
    }

    /*
     *   try to get output plugin
     *   for output stream
     */
    OutAVStreamIFace * outstream = 0;
    PropertyNode * out_caps = new PropertyNode( "outstream" );
    PropertyNode * file = out_caps->addDaughter("file");
    PropertyNode * container = file->addDaughter("container");
    container->addAttr( "format", "rawdv" );
    // FIXME  define codec properties and look for a codec
    // check that codec are found and fit with outstream
    Plugin * p = PluginMgr::find( out_caps );
    if ( !p ) {
        cout << "WARNING: no output plugin to handle output to screen" << endl;
    } else {
        /* reuse 'file' to create element */
        // FIXME add all codec properties to the 'file' request
        file->addAttr( "filename", filename );
        PluginBase * b = p->newElement( file );
        if ( !b ) {
            cout << "WARNING : instantiating plugin "<<p->getName()<<endl;
        } else {
            //  FIXME why doesn't this work ?? 
//                    voutstream = dynamic_cast<OutAVStreamIFace*>( b );
            outstream = static_cast<OutAVStreamIFace*>( b );
        }
    }
    if ( outstream==0 ) {
        cout << "ERROR : no plugin can handle requested output format:" << endl;
        cout << out_caps->toXML() << endl;
        exit( 1 );
    }
    PDELETE( out_caps );
    

    /* 
     * main render loop 
     *
     */
    Frame frame;
    for ( Time time = _theStoryBoard->start();
          time < _theStoryBoard->stop();
          time += Global::renderFmt.getFrameLength() )
    {
        // FIXME: send messages every second
        if ( ((int)(time.sec()*10))==(((int)time.sec())*10) ) {
            sprintf( Global::cbuf, "<rendering time=\"%f\" />", time.sec() );
            sendMessage( Global::cbuf );
        }
        frame = _theStoryBoard->getFrame( time );
        if (outstream) outstream->putFrame( frame );
    }

    /* done.. 
     * setup reply 
     */
    PropertyNode * ret = new PropertyNode( "reply" );
    ret->addAttr( "command", "render" );
    ret->addAttr( "status", "ok" );
    return ret;
}

PropertyNode * 
KdenLiveServer::veml_ping( const VEMLCmnd * cmd )
{
    /* setup reply */
    PropertyNode * ret = new PropertyNode( "reply" );
    ret->addAttr( "command", "ping" );
    ret->addAttr( "status", "pong" );
    return ret;
}



PropertyNode * 
KdenLiveServer::veml_setSceneList( const VEMLCmnd * cmd )
{
    /* parse scenelist */
    INFO("setSceneList start");

    cmd->toXML( cout );

    const PropertyNode * scenelist = cmd->getDaughter( "scenelist" );
    if ( !scenelist ) {
        WARN( "invalid command" );
        return 0;
    }

    StorySeqNode * head = SceneBuilder::buildScenelist( scenelist );

    //
    INFO( "new HEAD >>> " );
    head->print( cout , 3 );
    head->setEmptyFrame( _theStoryBoard->getEmptyFrame() );
    _theStoryBoard->setHead( head );
    INFO( "new HEAD <<< " );


    /* setup reply */
    PropertyNode * ret = new PropertyNode( "reply" );
    ret->addAttr( "command", "setSceneList" );
    ret->addAttr( "status", "ok" );
    return ret;
}


PropertyNode * 
KdenLiveServer::veml_getCapabilities( const VEMLCmnd * cmd )
{
    /* setup reply */
    PropertyNode * p;
    PropertyNode * caps = new PropertyNode( "capabilities" );

    PropertyNode * renderer = caps->addDaughter( "renderer" );
    renderer->addAttr( "name", "PIAVE" );
    renderer->addAttr( "version", VERSION );
    p = renderer->addDaughter( "author" );
    p->addAttr( "name", "Rolf Dubitzky" );
    p->addAttr( "email", "rolf@dubitzky.de" );
    p = renderer->addDaughter( "about" );
    p->setContent( "blah... video render engine.. blah.. \n"
                   "blah... GPL...  blah... \n"
                   "cool.... blah... \n"
                   "home: www.dubitzky.de/piave\n"
                   "(c) 2003 Rolf Dubitzky\n" );

    PropertyNode * general = caps->addDaughter( "general" );
    general->addAttr( "subFrameEditing" , "no" );
    general->addAttr( "oneTrackEffects" , "yes" );
    general->addAttr( "twoTrackEffects" , "yes" );


    PluginMgr::getCapabilities( caps );
    return caps;
}

