/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more detailnamis.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**
 * \file operatorbase.cc implements operatorbase.hh
 */

// config
#if defined( HAVE_CONFIG_H )
#    include "config.h"
#endif


// stdc

// stdc++
#include <string>
using std::string;

// this header
#include "operatorbase.hh"

// PIAVE

using namespace PIAVE;

void
OpParDouble::setParameter( const PropertyNode * p )
{
    if ( getIsDynamic() ) {
        const PropertyNode * d = p->getDaughter( getName(), 0 );
        if ( d ) {
            const PropertyNode * kf;
            int i = 0;
            while ( 0 != (kf = d->getDaughter( "keyframe" , i++ )) ) {
                double t = 0;
                if ( kf->getAttr( "time", t ) ) {
                    double v;
                    std::stringstream s;
                    s << kf->getContent(); 
                    s >> v;
                    KeyFrame<double> * k = makeNewKeyFrame( t, v );
                } else {
                    WARN( "keyframe without time attribute" );
                }
            }
        }
    } else {
        const PropertyNode * d = p->getDaughter( getName(), 0 );
        if ( d ) {
            std::stringstream s;
            KeyFrame<double> * k = getKeyFrame( 0 );
            s << d->getContent(); 
            s >> k->value;
        }
    }    
}

double
OpParDouble::getValue( Time t ) const
{
    double v;
    if ( getIsDynamic() ) {
        ABORT_IF( _path.empty(), "property path is empty" );
        const_iterator keyFrameIt  = previous( pbegin(), pend(), t );
        Time    posa = keyFrameIt->pos;
        double  vala = keyFrameIt->value;
        ++keyFrameIt;
        if ( keyFrameIt==pend() ) {
            return vala;
        }
        Time    posb = keyFrameIt->pos;
        double  valb = keyFrameIt->value;
        double  r    = ( t - posa ) / ( posb - posa );
        v = ( vala + ( valb - vala ) * r );
    } else {
                // FIXME assert a single keyframe
        v = pbegin()->value;
    }
    return v;
}


/*
 *    OpParString
 *
 *
 */
void
OpParString::setParameter( const PropertyNode * p )
{
    if ( getIsDynamic() ) {
        WARN("????  strings cannot be dynamic");
    } else {
        const PropertyNode * d = p->getDaughter( getName(), 0 );
        if ( d ) {
            KeyFrame<string> * k = getKeyFrame( 0 );
            k->value = d->getContent();
        }
    }    
}

string
OpParString::getValue( Time t ) const
{
    string v;
    if ( getIsDynamic() ) {
        WARN("????  strings cannot be dynamic");
    } else {
        // FIXME assert a single keyframe
        v = pbegin()->value;
    }
    return v;
}

/*
 *    OpParColor
 *
 *
 */
void
OpParColor::setParameter( const PropertyNode * p )
{
    if ( getIsDynamic() ) {
    } else {
        
    }    
}

Color
OpParColor::getValue( Time t ) const
{
    Color v;
    if ( getIsDynamic() ) {
        ABORT_IF( _path.empty(), "property path is empty" );
        const_iterator keyFrameIt  = previous( pbegin(), pend(), t );
        Time    posa = keyFrameIt->pos;
        Color   vala = keyFrameIt->value;
        ++keyFrameIt;
        if ( keyFrameIt==pend() ) {
            return vala;
        }
        Time    posb = keyFrameIt->pos;
        Color   valb = keyFrameIt->value;
        double  r = ( t - posa ) / ( posb - posa );
        int c[4];
        for ( int j=0;j<4;++j) {
            c[j] = static_cast<int>( static_cast<int>(vala[j]) + 
                                     (( static_cast<int>(valb[j]) - 
                                        static_cast<int>(vala[j]) ) * r ) );
        }
        v = Color( c[0], c[1], c[2], c[3] );
    } else {
        // FIXME assert a single keyframe
        v = pbegin()->value;
    }
    return v;
}

/*
 *    OpParBox
 *
 *
 */
void
OpParBox::setParameter( const PropertyNode * p )
{
    if ( getIsDynamic() ) {
        const PropertyNode * d = p->getDaughter( getName(), 0 );
        if ( d ) {
            const PropertyNode * kf;
            int i = 0;
            while ( 0 != (kf = d->getDaughter( "keyframe" , i++ )) ) {
                double t = 0;
                if ( kf->getAttr( "time", t ) ) {
                    double x,y,w,h;
                    bool ok = true;
                    ok &= kf->getAttr( "x", x );
                    ok &= kf->getAttr( "y", y );
                    ok &= kf->getAttr( "width", w );
                    ok &= kf->getAttr( "height", h );
                    if ( ok ) {
                        makeNewKeyFrame( t, Box(x,y,w,h) );
                    } else {
                        WARN( "invalid box attributes" );
                    }
                } else {
                    WARN( "keyframe without time attribute" );
                }
            }
        }
    } else {
        const PropertyNode * d = p->getDaughter( getName(), 0 );
        if ( d ) {
            KeyFrame<Box> * k = getKeyFrame( 0 );
            double x,y,w,h;
            bool ok = true;
            ok &= d->getAttr( "x", x );
            ok &= d->getAttr( "y", y );
            ok &= d->getAttr( "width", w );
            ok &= d->getAttr( "height", h );
            if ( ok ) {
                k->value = Box(x,y,w,h);
            } else {
                WARN( "invalid box attributes" );
            }
        }
    }    
}

Box
OpParBox::getValue( Time t ) const
{
    Box v;
    if ( getIsDynamic() ) {
        ABORT_IF( _path.empty(), "property path is empty" );
        const_iterator keyFrameIt  = previous( pbegin(), pend(), t );
        Time    posa = keyFrameIt->pos;
        Box     vala = keyFrameIt->value;
        ++keyFrameIt;
        if ( keyFrameIt==pend() ) {
            return vala;
        }
        Time    posb = keyFrameIt->pos;
        Box     valb = keyFrameIt->value;
        double  r = ( t - posa ) / ( posb - posa );
        v = Box( vala.x() + ( valb.x() - vala.x() ) * r, 
                 vala.y() + ( valb.y() - vala.y() ) * r,
                 vala.w() + ( valb.w() - vala.w() ) * r, 
                 vala.h() + ( valb.h() - vala.h() ) * r );
    } else {
        // FIXME assert a single keyframe
        v = pbegin()->value;
    }
    return v;
}

/*
 *
 */
void 
Operator::setParameter( const PropertyNode * p )
{
    std::list< OpParameterBase* >::iterator i;
    for ( i =  _parameter.begin();
          i != _parameter.end();
          ++i )
    {
        (*i)->setParameter( p );
    }
}


/**
 *   \class ReplaceAudioOp
 */
void 
ReplaceAudioOp::renderFrame( Frame & dest, Frame & src_a, Frame & src_b, Time t )
{ 
    INFO("hi");
    src_a.copyAudio( src_b );  
    
    dest = src_a; 
};
