/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**
 * \file piave_base.cc implements piave_base.hh
 */

// config
#if defined( HAVE_CONFIG_H )
#    include "config.h"
#endif

// std c
#include <execinfo.h>
#include <stdio.h>
#include <stdlib.h>

// std++
#include <typeinfo>
#include <cmath>
#include <iostream>
#include <iomanip>
#include <algorithm>
#include <sstream>
using std::string;
using std::cout;
using std::cerr;
using std::endl;

// this header
#include "piave_base.hh"

// PIAVE
#include "avstream.hh"
#include "storybase.hh"
#include "mediaelements.hh"
#include "pluginmgr.hh"
#include "algorithm.hh"

using namespace PIAVE;


/*
 *
 *  Global
 *
 *
 */
bool Global::verbose = false;
bool Global::quiet   = false;

bool Global::force_decode = false;

VideoFmt Global::renderFmt;
//VideoFmt Global::displayFmt;

Algorithm * Global::algo = 0;

const Color Color::white             ( 255, 255, 255, 255 );
const Color Color::black             (   0,   0,   0, 255 );
const Color Color::red               ( 255,   0,   0, 255 );
const Color Color::green             (   0, 255,   0, 255 );
const Color Color::blue              (   0,   0, 255, 255 );
const Color Color::yellow            ( 255, 255,   0, 255 );
const Color Color::magenta           ( 255,   0, 255, 255 );
const Color Color::cyan              (   0, 255, 255, 255 );
const Color Color::transparent_black (   0,   0,   0,   0 );
const Color Color::transparent_white ( 255, 200, 255,   0 );

char Global::cbuf[1024];

/*
 *   \class Time
 *
 */
//     char * endptr( 0 );
//     double time_sec = strtod( s.c_str(), &endptr );
//     if ( s.c_str()==endptr ) {
//         WARN( "parse error in  "<<s );
//         WARN( "                \""<<endptr );
//         return false;
//     }

//    return static_cast<int>( trunc( ( Global::fps * s ) + 0.5 ) );

Time::Time( const std::string & s ) 
{
    std::stringstream ss;
    ss << s;
    ss >> _sec;
}

string
Time::toString() const
{
    int s = static_cast<int>(trunc( _sec ));
    double msec = _sec - s;
    int m = s/60;
    s = s%60;
    int h = m/60;
    m = m%60;
    std::stringstream formated;
    // I want torture stroustrup ...
    formated<<std::setw(2)<<std::setfill('0')<<h<<":"
            <<std::setw(2)<<std::setfill('0')<<m<<":"
            <<std::setw(6)
#if __GNUC__ >= 3
            <<std::fixed
#endif
            <<std::setprecision(3)<<std::setfill('0')<<(s+msec);
    return string( formated.str() );
}


timeval
Time::toTimeval() const
{
    timeval t;
    t.tv_sec = static_cast<int>( floor( _sec ) );
    t.tv_usec = static_cast<int>( floor( _sec * 1000 - t.tv_sec * 1000 ) );
    return t;
}


std::ostream& 
PIAVE::operator<<( std::ostream& o, const Time& t )
{
    o << t.toString();
    return o;
}

/*  
 *    \class  Global
 *
 *
 *            initialize the Global space
 *
 */
void
Global::Init()
{
    // install signal handler for segv
    signal( SIGSEGV, Global::fatalSignalHandler );
    algo = new Algorithm_C();
    return;
}


/*
 *  destroy stuff in Global space
 */
void
Global::Destroy()
{
    delete algo;
    return;
}


int
Global::hexCharToI( const char a, const char b )
{
    char buf[3];
    char *endptr = 0;
    buf[0] = a;
    buf[1] = b;
    buf[2] = '\000';
    int i = strtol( buf, &endptr, 16 );

    if ( (*endptr)!='\000' ) 
        return -1;

    return i;
}


/*
 *   help debugging
 */
void
Global::printBackTrace()
{
    void   * array[32];
    size_t   size    = backtrace( array, 32 );
    char  ** strings = backtrace_symbols( array, size );
   
    if ( ! Global::quiet ) {
        cerr << "--- backtrace ------ start ----" << endl;
        for ( size_t i = 0; i < size; i++ ) {
            cerr << strings[i] << endl;
        }
        cerr << "--- backtrace ------ stop -----" << endl;
    }
   
    free( strings );
    return;
}
   
   
void
Global::fatalSignalHandler( int sig )
{
    signal( sig, SIG_DFL );
    static int n = 0;
    switch( sig ) {
      case SIGSEGV:
          WARN( "PIAVE caught a segmentation violation "<<n++ );
          printBackTrace();
          break;
      default:
          WARN( "Caught signal: "<<sig );
          break;
    }
    // FIXME   call emergency_exit_handlers
    // (e.g. to clean open tcp connects)
    //    right noe they are just atexit()
    exit( -sig );
}

/**
 * \namespace PIAVE
 * the master namespace in which PIAVE lives
 *
 *
 */

void 
PIAVE::Init() 
{ 
    Global::Init(); 
};

/**
 * construct color from a string in format #123456 or #123456FF where
 * 12, 34, and 56 are the color values in either RGB or YUV space. A
 * fourth value may define the alpha channel.
 * \todo FIXME make leading '#' optional
 */
Color::Color( const string & c )
{
    int r,g,b;
    r=g=b=0;
    int a = 255;
    
    if ( c.size() >= 7 )
    {
        r = Global::hexCharToI( c[1], c[2] );
        g = Global::hexCharToI( c[3], c[4] );
        b = Global::hexCharToI( c[5], c[6] );
        if ( c.size() >= 9 )
        {
            a = Global::hexCharToI( c[7], c[8] );
        }
    } else {
        WARN( "Illegal color string: "<<c );
        r = -1;
    }

    if ( r<0 || 
         g<0 || 
         b<0 || 
         a<0 ) 
    {
        r=g=b=0;
        a = 255;
    }
    
    _col[0] = r;
    _col[1] = g;
    _col[2] = b;
    _col[3] = a;

}

/**
 * serialize a Color to a ostream
 * is used when writing properties to XML nodes
 */
string
Color::toString3() const
{
    std::ostringstream o;
    o << "#"
      << std::setbase(16) 
      << std::setfill('0')
      << std::setw(2)
      << (unsigned short)_col[0]
      << std::setw(2)
      << (unsigned short)_col[1]
      << std::setw(2)
      << (unsigned short)_col[2];
    return o.str();
}

string
Color::toString4() const
{
    std::ostringstream o;
    o << "#"
      << std::setbase(16) 
      << std::setfill('0')
      << std::setw(2)
      << (unsigned short)_col[0]
      << std::setw(2)
      << (unsigned short)_col[1]
      << std::setw(2)
      << (unsigned short)_col[2]
      << std::setw(2)
      << (unsigned short)_col[3];
    return o.str();
}

string
Color::toStringA() const
{
    std::ostringstream o;
    o << "#"
      << (unsigned short)_col[3];
    return o.str();
}

std::ostream& 
PIAVE::operator<<( std::ostream& o, const PIAVE::Color& c)
{
    o << (unsigned short)c.space() <<":"
      << (unsigned short)c[0] <<":"
      << (unsigned short)c[1] <<":"
      << (unsigned short)c[2] <<":"
      << (unsigned short)c[3];
    return o;
}


TimeInterval
TimeInterval::intersect( const TimeInterval & o ) const 
{
    Time b = std::max( start(), o.start() );
    Time e = std::min( stop(), o.stop() );
    if ( b<e ) {
        return TimeInterval( b,e-b );
    } else {
        return TimeInterval( e,0 );
    }
}

TimeInterval
TimeInterval::join( const TimeInterval & o ) const 
{
    Time b = std::min( start(), o.start() );
    Time e = std::max( stop(), o.stop() );
    if ( e-b <= ( length()+o.length()) ) {
        return TimeInterval( b,e-b );
    } else {
        return TimeInterval( b,0 );
    }
}

void
TimeInterval::print( std::ostream & o, int indent ) const
{
    string s( indent, ' ' );
    o << s << typeid(*this).name() 
      << " start: "<<start()
      << " length: "<<length()
      <<endl;
    return;
}


std::ostream& 
PIAVE::operator<<( std::ostream& o, const TimeInterval& t )
{
    o<<"start="<<t.start()<<" length="<<t.length();
    return o;
}

/* 
 *    \class Point
 *
 *
 */
// \todo implement antialiasing
int 
Point::lx() const 
{ 
    return static_cast<int>( _x * Global::renderFmt.width ); 
}

int 
Point::ly() const 
{ 
    return static_cast<int>( _y * Global::renderFmt.height ); 
}



/* 
 *    \class Box
 *
 *
 */
int 
Box::lx() const 
{ 
    return static_cast<int>( _x * Global::renderFmt.width  ); 
}

int 
Box::ly() const 
{ 
    return static_cast<int>( _y * Global::renderFmt.height ); 
}
int 
Box::lw() const 
{ 
    return static_cast<int>( _w * Global::renderFmt.width  ); 
}

int 
Box::lh() const 
{ 
    return static_cast<int>( _h * Global::renderFmt.height ); 
}

int 
Box::lx2() const 
{ 
    return static_cast<int>( x2() * Global::renderFmt.width  ); 
}

int 
Box::ly2() const 
{ 
    return static_cast<int>( y2() * Global::renderFmt.height ); 
}

std::ostream& 
PIAVE::operator<<( std::ostream& o, const Box& b )
{
    o<<"(x,y)=("<<b.x()<<","<<b.y()
     <<") (w,h)=("<<b.w()<<","<<b.h()<<")";
    return o;    
}
