/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**
 * \file piave_base.hh contains the declarations of all the very basic PIAVE
 * classes, elementary structures, and constants. It also contains ppmacros
 * used throughout PIAVE. It depends on no other piave files.
 *
 * Here a short idea of how I do things in piave:
 * 
 * I store the render setup in a single tree. This tree is made up of 
 * "MediaElements". There are two Kinds of MediaElements: Nodes and Leaves. 
 *  All kinds of audio and video data is a Leave. E.g. a DV file , or a JPEG 
 * image, or an audiofile.
 *  There are two kind of nodes: unary and binary nodes. A unary node consists of 
 * an UnaryOperator and a single MediaElement. A BinaryNode has a BinaryOp and 
 * two MediaElements. You can already guess that a UnaryOp can take a Frame and 
 * modify it, eg. deinterlace, chroma adjust, invert, draw a ColorFrame around, 
 * etc. and a binary node can take two frames and do somehting like transitions 
 * or picture in picture or whatever. You can easily verify, that you can build 
 * _any_ kind of structure with this scheme. No limit of "number of video 
 * tracks" or other stupid things. You can also stack many BinaryOps on top of 
 * each other and have many source Frames for a single output frame. The only 
 * place where this scheme reaches a limit is if you e.g. want to project three 
 * frames on the three dimensinal picture of a dice. Then you need all three 
 * source Frames in a single Operator. Of course it is a non-issue to implement 
 * a TrinaryOp (or N-Op or whatever).
 *  Every MediaElement has a local coordinate system (local time) which starts at 
 * 0 and a starting time relative to its parents local time. This way, you can 
 * rearrange nodes without bothering where in te tree they end up. This is very 
 * similar to local coordinates in Qt I guess? The length of a node is defined 
 * as the union of the length of the children.
 * 
 * 
 * time      0________________________100_______________________200_____ _ _ _
 * 
 * 
 * node1      __________________________________________________________ _ _ _
 * - unaryOp1 _________
 * - node2    _________________________________________________
 *   - binOp1                        _____
 *   - clip1  ____________________________
 *   - clip2                         __________________________
 * 
 * 
 * 
 * Clear? Questions? ;-)
 * 
 * 
 * Now PropertyPath: Every Operator needs a set of parameters which define its 
 * behaviour. There are static operators which have a single set of parameters 
 * and DynamicOperators. A dynamic operator owns a ProprtyPath, which is a 
 * vector of Properties. Example: binOp1 from above. In it's parents localtime 
 * it ranges from let's say 95 to 108. His own local time is from 0 to 13. The 
 * PropertyPath has at least one node (that would qualify as a static op). If it 
 * has two nodes, the first node must be at 0 and the second at 13. Let's say 
 * binOp1 is AlphaBlendOp, i.e. the operator which can blend clip1 and clip2 
 * with a certain alpha value, from 0 (opaque) to 1 (transparent). If you use a 
 * propery path with two propery-nodes, the first would have alpha=1 the second 
 * alpha=0. Then all intermediate values will be interpolated (linear only at 
 * the moment). This system is _very_ flexible. You can build all kinds of 
 * effects with a minimal set of building blocks. 
 * 
 * 
 * Example:  long sequence (clip1) , blend in a second scene (clip1) for a short 
 * time in the upper right. Make node (node1) with AlphaBlendOp. make four 
 * property nodes time=0;alpha=1 time=10;alpha=0.2 time=20;alpha=0.2 
 * time=30;alpha=1 . Make second node with clip1 and node1 and a PicInPicOp. 
 * Your Done. The small picture will fade in stay almost opaque for 10 Frames 
 * and fade out again.
 * 
 * 
 * This can be used to animate every kind of property an Operator might have, 
 * coordinates of the PicInPic, Color of a Frame, size of a scaling, speed of 
 * motion, direction e.g. kooridnates of FlyIn or ShiftIn transitions...
 * 
 * 
 * In the timeline these property-nodes must be visible. You must be able to 
 * insert new nodes, delete old nodes and popup a window to edit the values for 
 * a node.
 * 
 * 
 *
 */

#if ! defined ( PIAVE_BASE_HH )
#define PIAVE_BASE_HH

// std
#include <unistd.h>
#include <stdint.h>
#include <errno.h>

// std++
#include <iostream>
#include <string>

// For i18n

#ifdef ENABLE_NLS
#  include <libintl.h>
#  undef _
#  define _(String) dgettext (PACKAGE, String)
#  ifdef gettext_noop
#    define N_(String) gettext_noop (String)
#  else
#    define N_(String) (String)
#  endif
#else
#  define textdomain(String) (String)
#  define gettext(String) (String)
#  define dgettext(Domain,Message) (Message)
#  define dcgettext(Domain,Message,Type) (Message)
#  define bindtextdomain(Domain,Directory) (Domain)
#  define _(String) (String)
#  define N_(String) (String)
#endif


/*
 *    some convenience macros
 */
#define ABORT(S)      { std::cerr << __FILE__ << ":" << __LINE__ << " ERROR: " << S << std::endl; abort(); }
#define WARN(S)       { if ( ! PIAVE::Global::quiet ) \
                      std::cerr << __FILE__ << ":" << __LINE__ << " WARNING: " << S << std::endl; }
#define INFO(S)       { if ( PIAVE::Global::verbose && ! PIAVE::Global::quiet  ) \
                      std::cout << __FILE__ << ":" << __LINE__ << " INFO: " << S << std::endl; }
#define ABORT_IF(C,S) { if ( C ) { std::cerr << __FILE__ << ":" << __LINE__ \
                      << " ERROR: (" << #C <<"): "<< S << std::endl; abort(); }}
#define WARN_IF(C,S)  { if ( ! PIAVE::Global::quiet ) \
                      if ( C ) { std::cerr << __FILE__ << ":" << __LINE__ \
                      << " WARNING: (" << #C <<"): " << S << std::endl; }}
#define INFO_IF(C,S)  { if ( PIAVE::Global::verbose && ! PIAVE::Global::quiet  ) \
                      if ( C ) { std::cout << __FILE__ << ":" << __LINE__ \
                      << " INFO: (" << #C <<"): " << S << std::endl; }}

#define PRINTV(C) " "<<#C<<"="<<C

#define PDELETE(P) if ( P!=0 ) { delete P ; P = 0;}

/*
 *
 * \namespace PIAVE
 *
 */
namespace PIAVE {

    class FOURCC;
    class Algorithm;
    class Time;
    class VideoFmt;

    /**
     *  global setup
     *  this is where all unique instances of
     *  global symbols and global setup go
     *  statics only.. could be a namespace tho
     *  
     */
    class Global
    {
      public:
        /// be verbose 
        static bool verbose;

        /// don't even warn
        static bool quiet;

        /// for benchmarking
        static bool force_decode;

        /* 
         *    this is the format the render tree is working in.
         */
        static VideoFmt renderFmt;


        /*
         *   this should be initialized according
         *   to runtime CPU detection
         */
        static Algorithm * algo;

        // I hate C++ formated IO!!!!!!!!!!!!!!
        static char cbuf[1024];

        /** 
         * Init the PIAVE::Global space
         *
         * before calling Init() all configuration must be set
         * ( i.e. preview mode, video system etc. )
         * - allocate decode/encoder
         * - set target video system
         * \todo protect agains double calling, better: allow it
         */
        static void Init();

        /** 
         * Destroy the PIAVE::Global space
         */
        static void Destroy();


        /*
         *   a few small helper functions
         */

        /**
         * interpretes two characters as a hexadezimal value
         */
        static int hexCharToI( const char a, const char b );

        /**
         *  print backtrace to help debugging
         */
        static void printBackTrace();

        /**
         *  signal handler
         */
        static void fatalSignalHandler( int );
        
    };

    /**
     * right now just a dummy to make PIAVE::Global::Init look nicer
     * but later it might be more to do than just init the PIAVE::Global space
     */
    void Init();


    /*
     * \class Time 
     *
     * All time counting is based on seconds at the moment this
     * is not much more than a double with some restrictions, but there might
     * be more things to it in the end..
     *
     */
    class Time
    {
      public:
        Time( double s = 0.0 ) : _sec( s ) {}
        Time( const Time & o ) : _sec( o._sec ) {}
        explicit Time( const std::string & s );

        double sec() const { return _sec; }
        
        const Time operator=( const Time & o ) { _sec = o._sec; return (*this); }
        const Time operator+( const Time & o ) const { return Time( _sec + o._sec ); }
        const Time operator-( const Time & o ) const { return Time( _sec - o._sec ); }
        const double operator*( const Time & o ) const { return _sec * o._sec; }
        const double operator/( const Time & o ) const { return _sec / o._sec; }
        const Time operator+=( const Time & o ) { return _sec += o._sec; }
        const Time operator-=( const Time & o ) { return _sec -= o._sec; }

        bool operator==( const Time & o ) const { 
            return (_sec-0.000001<o._sec)&&(_sec+0.000001>o._sec); }
        bool operator!=( const Time & o ) const { 
            return (_sec-0.000001>o._sec)||(_sec+0.000001<o._sec); }
        bool operator>( const Time & o ) const { return _sec>o._sec; }
        bool operator<( const Time & o ) const { return _sec<o._sec; }
        bool operator>=( const Time & o ) const { return ! (_sec<o._sec); }
        bool operator<=( const Time & o ) const { return ! (_sec>o._sec); }

        /** convert time value to string in format 
         *  hh:mm:ss.usec
         */
        std::string toString() const;

        /** convert time to timval (sys/time.h)
         *  with microsecond precision 
         */
        timeval toTimeval() const;

      private:
        double _sec;
    };
    std::ostream& operator<<( std::ostream&, const Time& );


    /*
     *  \class TimeInterval
     *  base for all story elements
     */
    class TimeInterval
    {
      public:
        TimeInterval( Time start=0, Time length=0 ) : 
          _start( start ), _length( length ) {} 
        TimeInterval( const TimeInterval & o ) :
          _start( o.start() ), _length( o.length() ) {};

        Time start()  const { return _start; } 
        /// stop() points to frame <B>AFTER</B> the interval, like STL end() !!
        Time stop()   const { return _start+_length; }
        Time length() const { return _length; }
        

        Time setStart ( const Time start )   { return _start=start; }
        Time setStop  ( const Time stop )    { if (stop>=_start) _length=stop-_start; return _length; }
        Time setLength( const Time length )  { return _length=length; }

        const TimeInterval & setTime( const TimeInterval & i ) 
            { _start=i.start(); _length=i.length(); return (*this); }
        const TimeInterval & operator=( const TimeInterval & o )
            { setTime( o ); return (*this); }

        /// @return true if t is inside this
        bool contains ( const Time t ) const 
            { return (t>=_start && t<_start+_length); }
        /// @return empty interval if t and this don't overlap
        TimeInterval intersect( const TimeInterval & t ) const ;
        /// @return empty interval if t and this don't overlap
        TimeInterval join( const TimeInterval & t ) const ;
        /// compares start and length of interavls
        bool operator==( const TimeInterval & o ) const {
            return _start==o._start && _length==o._length; };
        /// @return ! operator-
        bool operator!=( const TimeInterval & o ) const {
            return ! operator==( o ); }

        /**
         */
        virtual void print( std::ostream & o, int indent ) const;

      private:
        Time _start;
        Time _length;
    };
    std::ostream& operator<<( std::ostream&, const TimeInterval& );

    
    /*
     *  helper algorithm to find the previos property node in a
     *  property path.
     *
     */
    template<class I, class T>
    I
    previous( I first, I last, const T& val)
    {
        I o;
        while ( first != last ) {
            if ( (*first) == val ) return first;
            o = first;
            ++first;
            if ( first == last ) return o;
            if ( (*first) > val ) return o;
        }
            
        return first;
    }

    /**
     * \class Point 
     *  err... well... a point
     *
     * coordinates in constrctor are relative \in [0.0,1.0]
     * lx,ly,... are in the local coordinate system defined
     * be the project render size. 
     *
     */
    class Point
    {
      public:
        Point( double x, double y ) : _x( x ), _y( y ) {}
        Point( const Point & p ) : _x( p._x ), _y( p._y ) {}

//         const Point & set( const double x, const double y ) 
//             {_x=x;_y=y;return (*this);}
//         const Point & operator=( const Point & p ) 
//             {_x=x;_y=y;return (*this);}

//         bool operator==( const Point & p ) 
//             {return (p._x==_x&&p._y==y);}
//         bool operator!=( const Point & p ) 
//             {return (p._x!=_x&&p._y==y);}

//         double setx( const double x ) {return _x=x;}
//         double sety( const double y ) {return _y=y;}

        // @{
        /// relative coordinates [0.0,1.0]
        double x() const { return _x; }
        double y() const { return _y; }
        // @}

        // @{
        /// absolute coordinates wrt, render size
        int lx() const;
        int ly() const;
        // @}

      private:
        double _x;
        double _y;
    };

    /**
     * it's, yes, right, a box! ! .
     *   
     * coordinates in constrctor are relative \in [0.0,1.0]
     * lx,ly,... are in the local coordinate system defined
     * be the project render size. 
     *
     */
    class Box
    {
      public:
        Box( double x = 0.1 , double y = 0.1 , double w = 0.8 , double h = 0.8 ) :
          _x(x),_y(y),_w(w),_h(h) {}
        /// upper left and lower right
        Box( Point ul, Point lr ) : 
          _x(ul.x()),_y(ul.y()),_w(lr.x()-ul.x()),_h(lr.y()-ul.y()) {}

        // @{
        /// relative coordinates [0.0,1.0]
        double x() const { return _x; }
        double y() const { return _y; }
        double w() const { return _w; }
        double h() const { return _h; }
        double x2() const { return _x+_w; }
        double y2() const { return _y+_h; }
        // @}

        // @{
        /// absolute coordinates wrt, render size
        // \todo implement antialiasing
        int lx() const;
        int ly() const;
        int lw() const;
        int lh() const;
        int lx2() const;
        int ly2() const;
        // @}

        /// upper left
        Point ul() const { return Point( _x,_y ); }
        /// lower right
        Point lr() const { return Point( x2(),y2() ); }

      private:
        double _x,_y,_w,_h;

    };
    std::ostream& operator<<( std::ostream&, const Box& );


    /**
     * Color
     *
     */
    class Color 
    {
      public:
        enum Space {
            NONE,
            RGB,
            YUV
        };

        Color( uint8_t a=0, uint8_t b=0, uint8_t c=0, uint8_t d=0 ) {
            _col[0] = a; _col[1] = b; _col[2] = c; _col[3] = d; }
        Color( const std::string & c );

        uint8_t &       operator [] ( uint i )       { return _col[i]; }
        const uint8_t & operator [] ( uint i ) const { return _col[i]; }
        Space space() const { return _space; }
        void  setSpace( Space s ) { _space=s; }

        std::string     toString3() const;
        std::string     toString4() const;
        std::string     toStringA() const;

      private:
        Space _space; // FIXME !! doubles Frame::Colorspace ??
        uint8_t _col[4]; // FIXME use r g b ??? can I save array accesses?

      public:
        static const Color white;
        static const Color black;
        static const Color red;
        static const Color green;
        static const Color blue;
        static const Color yellow;
        static const Color magenta;
        static const Color cyan;
        static const Color transparent_black;
        static const Color transparent_white;
    };
    std::ostream& operator<<( std::ostream&, const Color& );

};

#endif
