/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**
 * \file player.cc
 */

// config
#if defined( HAVE_CONFIG_H )
#    include "config.h"
#endif



// std c
#include <unistd.h>
#include <stdio.h>

// this header
#include "player.hh"

// std c++
#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <string>
using std::string;


// PIAVE
#include "frame.hh"
#include "server.hh"

using namespace PIAVE;


/*
 *  \class Player
 *
 */
Player::Player()
  : 
  _state( not_ready ),
  _theStory( 0 ),
  _videoOut( 0 ),
  _audioOut( 0 ),
  _speed( 1.0 ),
  _start( 0 ),
  _stop( -1 ),
  _pos( 0 )
{
    _firstFlash.tv_sec = 0;
    _firstFlash.tv_usec = 0;
    _lastFlash.tv_sec = 0;
    _lastFlash.tv_usec = 0;
}

Player::~Player()
{
}

void
Player::switchToPlay()
{
    switch ( _state )
    {
      case paused:
      case not_ready:
      case stopped: 
          if ( _theStory ) {
              _state = playing;
              if ( Server::Inst ) {
                  Server::Inst->sendMessage( "<render status=\"playing\"/>" );
              }
              _lastFlash.tv_sec = 0;
              _lastFlash.tv_usec = 0;
              gettimeofday( &_firstFlash, NULL );
              _pos = _start;
              playNextFrame();
          }
          break;
      case playing: return;
      default: WARN( "unknown state" );
    }
}

void
Player::switchToStop()
{
    switch ( _state )
    {
      case stopped: return;
      case paused:
      case not_ready:
      case playing:
          _pos = 0;
          _state = stopped;
          if ( Server::Inst ) {
              Server::Inst->sendMessage( "<render status=\"stopped\"/>" );
          }
          break;
      default: WARN( "unknown state" );
    }
}

void
Player::switchToPause()
{
    switch ( _state )
    {
      case paused: return;
      case not_ready:
      case stopped: 
      case playing:
          _state = paused;
          if ( Server::Inst ) {
              Server::Inst->sendMessage( "<render status=\"paused\"/>" );
          }
          break;
      default: WARN( "unknown state" );
    }
}

bool
Player::diffTime( timespec *r,  const timeval * x, timeval * y )
{
    /* Perform the carry for the later subtraction by updating Y. */
    if (x->tv_usec < y->tv_usec) {
        int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
        y->tv_usec -= 1000000 * nsec;
        y->tv_sec += nsec;
    }
    if (x->tv_usec - y->tv_usec > 1000000) {
        int nsec = (x->tv_usec - y->tv_usec) / 1000000;
        y->tv_usec += 1000000 * nsec;
        y->tv_sec -= nsec;
    }
    
    /* Compute the time remaining to wait.
       `tv_usec' is certainly positive. */
    r->tv_sec = x->tv_sec - y->tv_sec;
    r->tv_nsec = ( x->tv_usec - y->tv_usec ) * 1000;
    
    /* Return 1 if result is negative. */
    return (x->tv_sec < y->tv_sec) ;
}

void
Player::addTime( timeval *r, const timeval * x, const timeval * y )
{
    /* Perform the carry for the later subtraction by updating Y. */
    r->tv_sec = x->tv_sec + y->tv_sec;
    long int u = x->tv_usec + y->tv_usec; 
    r->tv_usec = u % 1000000;
    r->tv_sec += u / 1000000;
    return;
}

void
Player::playNextFrame()
{
    if ( _state!=playing ) return;
    if ( !_theStory ) return;

    static bool first = true;
    static int v = 10;

    /* get the Frame */
    Frame frame = _theStory->getFrame( _pos );

    /* wait till it is time to display it */
    if ( _lastFlash.tv_sec == 0 ) {
        /* first call, nothing to wait for, just initialization */
        gettimeofday( &_lastFlash, NULL );
    } else {
        /* get this frames duration */
        timeval thisFrameDur = frame.getLength().toTimeval();
        // FIXME do it right with seconds 
        thisFrameDur.tv_usec = static_cast<int>( (double)thisFrameDur.tv_usec / _speed ); 
//         if ( first ) 
//         {
//             thisFrameDur.tv_usec = static_cast<int>( (double)thisFrameDur.tv_usec / 2 ); 
//             first = false;
//         } else {
// //            thisFrameDur.tv_usec = static_cast<int>( (double)thisFrameDur.tv_usec * 0.8 ); 
//         }            

        /* where are we now? */
        timeval now;
        gettimeofday( &now, NULL );

        /* calculate the time of next flash, and the remaining time */
        timeval nextFlash;
        addTime( &nextFlash, &_lastFlash, &thisFrameDur );
        timespec remaining;
        diffTime( &remaining, &nextFlash, &now );
        _lastFlash = nextFlash;

        /* wait */
        nanosleep( &remaining, NULL );
    }
    
    /* in case we have a server, give feedback about which frame
     * beeing displayed */
    if ( Server::Inst ) {
        sprintf( Global::cbuf, "<playing time=\"%f\" />", 
                 getPos().sec() );
        Server::Inst->sendMessage( Global::cbuf );
    }

    /* display the frame */
    if ( v=0 )
    {
        INFO( "playing "<<_pos<<PRINTV(_start)
              <<PRINTV(_stop)<<PRINTV( _theStory->stop() ) );
        v=10;
    } else v--;
    if ( _videoOut ) _videoOut->putFrame( frame );
    if ( _audioOut ) _audioOut->putFrame( frame );

    /* calculate the position of the next frame to display */

    _pos += frame.getLength();
    if ( frame.getLength() !=  Global::renderFmt.getFrameLength() )
    {
//        INFO( PRINTV( frame.getLength() )<<PRINTV( Global::renderFmt.getFrameLength() ) );
//        WARN( "FIXME" );
    }

    /* check on stop conditions */
    if ( _stop >= _start ) 
    { // stop is actually set
        if ( _pos > _stop ) {
            switchToStop();
        }
    } else { // stop is unset -> play to the end
        if ( _pos >= _theStory->stop() ) {
            switchToStop();
        }
    }
//     INFO( "playing "<<_pos<<PRINTV(_start)
//           <<PRINTV(_stop)<<PRINTV( _theStory->stop() ) );

    return;
}



