/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**
 * \file player.hh implements a simple main loop and all infrastructure to try to
 * run this main loop in the requested speed.
 */

#if ! defined( PLAYER_HH )
#define PLAYER_HH

// stdc
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/time.h>

// stdc++
#include <string>

// PIAVE
#include "piave_base.hh"
#include "storybase.hh"


namespace PIAVE {

    /*
     *  \class Player implements a simple main loop and all
     * infrastructure to try to * run this main loop in the requested
     * speed.
     *
     * The Player does neither own tho storyboard, nor the out streams
     * it just uses the streams provided by the system. The only purpose is
     * to handle the timing of the play speed.
     * 
     * \todo  sound will only be played when speed == 1.0
     *
     */

    class Player
    {
      public:
        enum State { not_ready, stopped, paused, playing };
        
      public:
        Player();
        ~Player();

        /// get properties
        double       getSpeed()       const { return _speed; }
        Time         getPos()         const { return _pos; }
        State        getState()       const { return _state; }

        /// set properties
        void setStoryBoard( StoryBoard * s ) { _theStory = s; }
        void setVideoOut( OutAVStreamIFace * vo ) { _videoOut = vo; }
        void setAudioOut( OutAudioStreamIFace * ao ) { _audioOut = ao; }
        void setSpeed( double s ) { _speed = s; }
        void setStart( Time s ) { _start = s; }
        void setStop( Time s ) { _stop = s; }

        /// set state of the player
        void switchToPlay();
        void switchToStop();
        void switchToPause();

        /* will play the next frame. this call will block as long as necessary
         * with respect to the last call to keep playing at the correct speed.
         */
        void playNextFrame();

      private:
        /*   r = x - y    
         * @return true if positive, flase if negative
         */
        bool diffTime( timespec * r, const timeval * x,  timeval * y );

        /// r = x + y;
        void addTime( timeval * r, const timeval * x, const timeval * y );


      private:
        /// the state the player is in
        State         _state;

        ///  this will be displayed
        StoryBoard   * _theStory;

        /// this is where the video goes
        OutAVStreamIFace * _videoOut;

        /// this is where audio goes;
        OutAudioStreamIFace * _audioOut;

        /// 0.0 == stopped; 1.0 normal; <1.0  slomo, >1,0 fast fwd
        double        _speed;

        /// start playing at this position
        Time          _start;

        /// stop playing at this position
        Time          _stop;

        /// the current playing position in the timeline
        Time          _pos;

        /** the time of displaying the first of all frames was
         *  displayed. Used just for statistical reasons. */
        timeval _firstFlash;

        /** the time when the last frame was displayed. Important to
         *  determin how long to wait till the next frame will be
         *  displayed */
        timeval _lastFlash;
    };

};

#endif
