/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**
 * \file plugin.hh
 */

#if ! defined( PLUGIN_HH )
#define PLUGIN_HH

// STD C++
#include <iostream>
#include <string>

// PIAVE
#include "piave_base.hh"

#define PLUGIN_IFACE_VERSION 1

namespace PIAVE { 

    class PropertyNode;

    class Plugin;
    
    typedef Plugin * (*PluginMaker)();

    /** \struct PluginInfo
     *  defines the entry symbol of a piave plugin library
     */
    struct PluginInfo {
        const char * name;
        int iface_version;
        PluginMaker maker;
    };

    /** this is a convenient macro to simplify the definition of the
     *  entry symbol it only works for one plugin per library. In case
     *  of more plugins in pne library, you have to define the
     *  pluginInfo array by hand.
     */
#define PIAVE_PLUGIN( N )  \
        Plugin * new_ ## N () { return new  N (); }; \
        PluginInfo pluginInfo[] = { { #N , PLUGIN_IFACE_VERSION , &new_ ##N } , \
                                    { 0  , 0                    , 0 } \
                                  };;

    /** \enum PluginType_e
     *  enumerates the various categories of plugins.
     *  at some point it is necessary to typecast the plugins into the
     *  class of their catagory. Even though this usually clear by the
     *  requested capabilities of which category a certain PluginBase is, the
     *  category asserts that nothing can go wrong.
     */
    enum PluginType_e {
        /* meta types for queries */
        PLUGIN_NO_TYPE = 0, 
        PLUGIN_ANY_TYPE, 
        /* effect plugins */
        PLUGIN_VIDEO_OP,
        PLUGIN_AUDIO_OP,
        /* iostreams and container */
        PLUGIN_INSTREAM,
        PLUGIN_OUTSTREAM,
        /* codecs */
        PLUGIN_VIDEO_CODEC,
        PLUGIN_AUDIO_CODEC,
        /* media elements */
        PLUGIN_MEDIA_ELEMENT
    };


    /**
     *  \class PluginBase
     *  all classes that want to be plugins should inherit this class.
     */
    class PluginBase
    {
      public:
        PluginBase() {};
        virtual ~PluginBase() {}
        virtual PropertyNode * getProperties( PropertyNode *p  = 0 ) const = 0;

    };

    /**
     * \class Plugin 
     * baseclass for all Plugins
     *
     */
    class Plugin 
    {
      public:
        Plugin( const std::string & name, PluginType_e t );
        virtual ~Plugin();

        std::string     getName() const { return _name; }
        int             getPriority()const { return _priority; }
        PluginType_e    getType() const { return _type; }
        PropertyNode *  getCapabilities() const;
        bool            isEnabled() const { return _enabled; }

        void setName( const std::string & n ) { _name = n; }
        void setPriority( int p ) { _priority = p; }
        void setCapabilities( PropertyNode * caps ) { _caps = caps; };
        void setEnabled( bool enabled = true ) { _enabled = enabled; }

        /// this class serves as a factory for the effect/codec/whatever
        virtual PluginBase    * newElement( const PropertyNode * p = 0 ) = 0;
        

      private:
        /** the name deosn't actually mean much, but helps for
         *  debugging and overview 
         */
        std::string _name;

        /** in case more than one plugin fit the requirements of a
         *  query the plugin with highest priority is selected. 
         */
        int _priority; 

        /** in principle the requested properties in a
         *  PluginMgr::find() call already determine the plugin type,
         *  but this aditional, slightly redundant parameter makes the
         *  typecast a little safer.
         */
        PluginType_e _type;

        /** the capabilites of the plugin. When somebody requests a
         * plugin with certain capabilities from the PluginMgr , it is
         * important, that the plugin defines the correct
         * capabilities, so it can be identified. 
         *
         * \todo document the minimum capabilities a
         * screen/sound/whatever plugin needs to define
         */
        PropertyNode * _caps; 

        /** a plugin might be disabled if it has turned out, that is
         *  does not work in the current environmen. IEEE1394 might
         *  not be available or aRts may not be running etc...
         */
        bool _enabled;

    };

}

#endif
