/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**
 * \file pluginmgr.cc implments the plugin manager
 */

// config
#if defined( HAVE_CONFIG_H )
#    include "config.h"
#endif


// std c
#include <dlfcn.h>
#include <sys/types.h>
#include <dirent.h>

// std c++
#include <iostream>
#include <string>
using std::string;
using std::cout;
using std::endl;

// this
#include "pluginmgr.hh"
#include "property.hh"

// std c
#include <dlfcn.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>


// PIAVE
#include "piave_base.hh"

using namespace PIAVE;

PIAVE::PluginMgr * PIAVE::PluginMgr::_inst = 0;





int
PluginMgr::openDLL( const std::string & fileName )
{
    int n = 0;
    void * lib = 0;
    if( !(lib = dlopen ( fileName.c_str() , RTLD_LAZY | RTLD_GLOBAL))) {
        WARN( "failed to open dll: "<< fileName 
              << " with errors:" << endl << dlerror() );
        return 1;
    } else {
        INFO( "opened dll: " << fileName );

        PluginInfo * pluginInfo = 
            static_cast<PluginInfo *>( dlsym(lib, "pluginInfo") );

        if ( pluginInfo == 0 ) {
            WARN( "no pluginInfo symbol - "<<fileName<<" - no PIAVE plugin ??" );
            return 0;
        }

        const char * pluginName = 0;
        while ( ( pluginName = pluginInfo->name ) ) {
            INFO( "   Register plugin "<< PRINTV( pluginName ) );
            PluginMap::const_iterator plugIter = Inst()->_plugins.find( pluginName );
            if ( plugIter ==  Inst()->_plugins.end() ) {
                Inst()->_plugins[ pluginName ] = (*pluginInfo->maker)();
                n++;
            } else {
                INFO( "plugin \""<<pluginName<<"\" already registered." );
            }
            pluginInfo++;
        }
    }
    return n;
}



int 
PluginMgr::openDLLsFromDir( const string & dirName, bool recurse )
{
    int n = 0;
    DIR * dir = opendir( dirName.c_str() );
    if (dir) {
        struct dirent * dirEntry;
        while ( (dirEntry = readdir( dir ) ) != NULL) {
            string filename( dirEntry->d_name );
            if ( filename == "." || filename == ".." ) continue;

            // a [otential plugin ?
            filename = dirName + "/" + filename;
            if ( filename.find( ".so" ) != string::npos ) {
                n += Inst()->openDLL(  filename );
            }

            // a directory ?
            if ( recurse ) {
                // a subdir
                struct stat dirStat;
                if( !stat( filename.c_str(), &dirStat ) ) {
                    if ( S_ISDIR( dirStat.st_mode ) ) {
                        n += Inst()->openDLLsFromDir( filename, recurse );
                    }
                } else {
                    WARN( "problem accessing "<<PRINTV(filename));
                }
            }
        }
    }
    return n;
}


int 
PluginMgr::openDLLsFromDefaultPath()
{
    int n = 0;

    /* first look in cwd */
    n += Inst()->openDLLsFromDir( "./plugins" );

    /* then in the path defined at compile time */
    n += Inst()->openDLLsFromDir( PIAVE_PLUGINDIR );

    return n;
}

void
PluginMgr::listPlugins( std::ostream & o )
{
    o << "=====[ BEGIN LOADED PLUGINS ]============" << endl;
    PluginMap::const_iterator iter;
    for ( iter =   Inst()->_plugins.begin(); 
          iter !=  Inst()->_plugins.end();
          iter++ )
    {
        Plugin * p = iter->second;
        o << "PluginID:   " << iter->first << endl
          << "PluginName: " << p->getName() <<endl
          << "PluginType: " << p->getType() <<endl;
        o << "Capabilities:" << endl;
        const PropertyNode * pn = p->getCapabilities();
        if ( pn ) {
            o << pn->toXML() << endl;
        } else {
            o << "none" << endl;
        }
        PDELETE( pn );
        o << "----------------------------------" << endl;
    }
    o << "=====[ END LOADED PLUGINS ]============" << endl;
}


Plugin * 
PluginMgr::find( PropertyNode * caps, PluginType_e type )
{
    Plugin * match = 0;

    PluginMap::const_iterator iter;
    for ( iter =   Inst()->_plugins.begin(); 
          iter !=  Inst()->_plugins.end();
          iter++ )
    {
        Plugin * p = iter->second;
        if ( !p->isEnabled() ) continue;
        if ( type != PLUGIN_ANY_TYPE && p->getType() != type ) continue;
        const PropertyNode * plugin_caps = p->getCapabilities();
        if( plugin_caps ) {
            if ( plugin_caps->provides( caps ) ) {
                if ( match ) {
                    if ( p->getPriority() > match->getPriority() ) {
                        match = p;
                    }
                } else {
                    match = p;
                }
            }
            PDELETE( plugin_caps );
        }
    }
    if ( !match ) {
        WARN( "plugin not found" );
    }
    return match;
}


Plugin * 
PluginMgr::find( std::string name, PluginType_e type )
{
    Plugin * match = 0;

    PropertyNode * caps = new PropertyNode( "capabilities" );
    caps->addAttr( "name", name );

    match = find( caps, type );

    PDELETE( caps );
    return match;
}


void   
PluginMgr::getCapabilities( PropertyNode * c )
{
    /* the categorization into the groubs "effects", "iostreams", and
     * "codecs" is basically for kdenlive 
     */
    PropertyNode * effects = c->addDaughter( "effects" );
    PropertyNode * iostreams = c->addDaughter( "iostreams" );
    PropertyNode * codecs = c->addDaughter( "codecs" );
    PluginMap::const_iterator iter;
    for ( iter =   Inst()->_plugins.begin(); 
          iter !=  Inst()->_plugins.end();
          iter++ )
    {
        Plugin * p = iter->second;
        switch ( p->getType() ) {
          case PLUGIN_VIDEO_OP:
          case PLUGIN_AUDIO_OP:
          case PLUGIN_MEDIA_ELEMENT:
              effects->addDaughter( p->getCapabilities() );
              break;
          case PLUGIN_INSTREAM:
          case PLUGIN_OUTSTREAM:
              iostreams->addDaughter( p->getCapabilities() );
              break;
          case PLUGIN_VIDEO_CODEC:
          case PLUGIN_AUDIO_CODEC:
              codecs->addDaughter( p->getCapabilities() );
              break;
          case PLUGIN_ANY_TYPE:
          case PLUGIN_NO_TYPE: 
          default: 
              WARN( "unknown plugin type in plugin: "<<p->getName() );
              break;
        }
    }
}
