/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**
 * \file pluginmgr.cc declares the pluginmanager
 */

#if ! defined( PLUGINMGR_HH )
#define PLUGINMGR_HH

// STD C++
#include <iostream>
#include <map>
#include <string>

// PIAVE
#include "exception.hh"
#include "plugin.hh"

namespace PIAVE 
{
    
    /**
     *  \class PluginMgr 
     *
     *  The PluginMgr is a singleton factory for and a register of all
     *  known plugins.
     *
     *  It also serves as a loader. It tries to open all shared
     *  library files from the default or specified directories and
     *  searches for the symbol "piavePluginInfo". If found, it is
     *  assumed the library contains one or more plugins which will be
     *  added to the register.
     */
    typedef std::map< std::string, Plugin * > PluginMap;
    class PluginMgr
    {
      private:
        ///  Hide constructor, since this is a singleton.
        PluginMgr() {};
        static PluginMgr * Inst() { if (_inst==0) {_inst=new PluginMgr();} return _inst;};

      public:

        static const PluginMap & getPluginMap() { return Inst()->_plugins; }

        static int    openDLL( const std::string & fileName );
        static int    openDLLsFromDir( const std::string & dirName, bool recurse = true );
        static int    openDLLsFromDefaultPath();

        static void   listPlugins( std::ostream & o );
        static void   getCapabilities( PropertyNode * c );

        /** find a enabled plugin that fits the required caps.  in
         *  case more than one plugin fits the caps, the plugin with
         *  the highes priority is returned.
         *
         *  It is _not_ checked if the plugin will work. Some plugins
         *  might require e.g. a special sound server to run and will 
         *  throw an exception when trying to instantiate.
         */
        static Plugin * find( PropertyNode * caps, PluginType_e type = PLUGIN_ANY_TYPE );

        /**  find plugin by name */
        static Plugin * find( std::string name, PluginType_e type = PLUGIN_ANY_TYPE );
        
      private:
        static PluginMgr * _inst;

        PluginMap    _plugins;
    };
    
}

#endif
