/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

// config
#if defined( HAVE_CONFIG_H )
#    include "config.h"
#endif

// std c
#include <stdio.h>

// std c++
#include <sstream>

// this 
#include "property.hh"

// piave
#include "piave_base.hh"


using namespace PIAVE;
using std::string;
using std::cout;
using std::cerr;
using std::endl;

PropertyNode::PropertyNode( const string & n ) : 
  _name( n ),
  _content( 0 ),
  _attrMap( 0 ),
  _daughters( 0 )
{}


PropertyNode::PropertyNode( const PropertyNode & o ):
  _name( o._name ),
  _content( 0 ),
  _attrMap( 0 ),
  _daughters( 0 )
{
    if ( o._content ) {
        _content = new string( *o._content );
    }

    if ( o._attrMap ) {
        _attrMap = new AttrMap( *o._attrMap );
    }

    if ( o._daughters ) {
        PropertyNodeList::const_iterator iter;
        for ( iter =  o._daughters->begin();
              iter != o._daughters->end();
              ++iter )
        {
           addDaughter( new PropertyNode( *(*iter) ) );
        }
    }
    return;
}


PropertyNode::~PropertyNode()
{
    PDELETE( _content );
    PDELETE( _attrMap );
    if ( _daughters ) {
        PropertyNodeList::const_iterator iter;
        for ( iter =  _daughters->begin();
              iter != _daughters->end();
              ++iter )
        {
           delete (*iter);
        }
        PDELETE( _daughters );
    }
}


void
PropertyNode::toXML( std::ostream & o, int indent ) const
{
    string s( indent, ' ' );
    o << s << "<"<<_name;
    if ( _attrMap ) {
        AttrMap::const_iterator iter;
        for ( iter =  _attrMap->begin();
              iter != _attrMap->end();
              ++iter )
        {
            o << " " << iter->first
              << "=\"" << iter->second << "\"";
        }
    }

    if ( _daughters ) {
        o << ">" <<endl;
        PropertyNodeList::const_iterator iter;
        for ( iter =  _daughters->begin();
              iter != _daughters->end();
              ++iter )
        {
            (*iter)->toXML( o, indent + 4 );
        }
    }

    if ( _content ) {
        o << ">" << (*_content);
    }

    if ( _content ) {
        o << "</"<<_name<<">"<<endl;
    } else if ( _daughters ) {
        o << s << "</"<<_name<<">"<<endl;
    } else {
        o << " />" <<endl;
    }

    return;
}

std::string 
PropertyNode::toXML() const 
{
    std::ostringstream o;
    toXML( o );
    return o.str();
}

void
PropertyNode::operator=( const PropertyNode & o ) 
{
    _name = o._name;
    PDELETE( _attrMap );
    if ( o._attrMap ) {
        _attrMap = new AttrMap( *o._attrMap );
    }

    // daughter commands are not copied

    return;
}

void 
PropertyNode::addAttr( const std::string & n, const std::string & v )
{
    if (!_attrMap) _attrMap = new AttrMap();
    (*_attrMap)[ n ] = v;
}

void 
PropertyNode::addAttr( const std::string & n, const double & v )
{
    sprintf( Global::cbuf, "%f", v );
    addAttr( n, Global::cbuf );
}

void 
PropertyNode::addAttr( const std::string & n, const int & v )
{
    sprintf( Global::cbuf, "%d", v );
    addAttr( n, Global::cbuf );
}


void 
PropertyNode::addDaughter( PropertyNode * c )
{
    if ( !_daughters ) _daughters = new PropertyNodeList();
    _daughters->push_back( c );
}

void 
PropertyNode::setName( const std::string & c )
{
    _name = c;
}

void 
PropertyNode::setContent( const std::string & c )
{
    if ( !_content ) 
        _content = new string( c );
    else
        (*_content) = c;
}

PropertyNode *
PropertyNode::getDaughter( const std::string & n, int w, bool recurse )
{
    if ( !_daughters ) return 0;
    PropertyNodeList::const_iterator iter;
    for ( iter =  _daughters->begin();
          iter != _daughters->end();
          ++iter )
    {
        if ( (*iter)->getName() == n ) {
            if ( w == 0 ) {
                return (*iter);
            }
            w--;
        }
    }

    if ( recurse ) {
        PropertyNodeList::const_iterator iter;
        for ( iter =  _daughters->begin();
              iter != _daughters->end();
              ++iter )
        {
            PropertyNode * p = (*iter)->getDaughter( n , w, recurse );
            if (p) return p;
        }
    }

    return 0;
}

const PropertyNode *
PropertyNode::getDaughter( const std::string & n, int w, bool recurse ) const
{
    PropertyNode * self = const_cast<PropertyNode *>( this );
    return self->getDaughter( n, w, recurse );
}

const PropertyNode *
PropertyNode::getDaughter( unsigned int w ) const
{
    if ( _daughters &&  _daughters->size()>w ) {
        return (*_daughters)[w];
    }
    return 0;
}

PropertyNode *
PropertyNode::getDaughter( unsigned int w )
{
    if ( _daughters &&  _daughters->size()>w ) {
        return (*_daughters)[w];
    }
    return 0;
}

PropertyNode* 
PropertyNode::addDaughter( const std::string & n )
{
   PropertyNode * d = new PropertyNode( n );
   addDaughter( d );
   return d;
}


bool
PropertyNode::getAttr( const std::string & a, string & v ) const
{
    if ( !_attrMap ) {
        return false;
    }

    AttrMap::const_iterator attrIter = _attrMap->find( a );
    if ( attrIter == _attrMap->end() ) return false;
    v = attrIter->second;
 
    return true;
}

bool
PropertyNode::getAttr( const std::string & a, double & v ) const
{
    string s;
    if ( !getAttr( a, s ) ) return false;
    char * endptr( 0 );
    v = strtod( s.c_str(), &endptr );
    if ( s.c_str()==endptr ) {
        WARN( "parse error in  "<<s );
        WARN( "                \""<<endptr );
        return false;
    }
    return true;
}

bool
PropertyNode::getAttrRec( const std::string & a, string & v ) const
{
    if ( getAttr( a, v ) ) return true;
    if ( _daughters ) {
        PropertyNodeList::const_iterator iter;
        for ( iter =  _daughters->begin();
              iter != _daughters->end();
              ++iter )
        {
            
            if ( *iter && (*iter)->getAttrRec( a, v ) ) return true;
        }
    }
    return false;
}

bool
PropertyNode::getAttrRec( const std::string & a, double & v ) const
{
    if ( getAttr( a, v ) ) return true;
    if ( _daughters ) {
        PropertyNodeList::const_iterator iter;
        for ( iter =  _daughters->begin();
              iter != _daughters->end();
              ++iter )
        {
            if ( *iter && (*iter)->getAttrRec( a, v ) ) return true;
        }
    }
    return false;
}

bool 
PropertyNode::provides( PropertyNode * caps, bool recurse ) const
{
    bool match = false;

    // name has to match
    if ( _name == caps->getName() ) {
        match = true;

        // attribs have to match
        if ( caps->_attrMap ) {
            if ( _attrMap ) {
                AttrMap::const_iterator iter;
                for ( iter =  caps->_attrMap->begin();
                      ( iter != caps->_attrMap->end() ) && match;
                      ++iter )
                {
                    string val;
                    bool exist = getAttr( iter->first, val );
                    if ( !exist ) {
                        match = false;
                    } else if ( val != iter->second ) {
                        match = false;
                    }
                }
            } else {
                match = false;
            }
        }

        // daughters have to match
        if ( match && caps->_daughters ) {
            if ( _daughters ) {
                // check all daughters matche
                PropertyNodeList::const_iterator citer;
                for ( citer =  caps->_daughters->begin();
                      citer != caps->_daughters->end() && match;
                      ++citer )
                {
                    // check our daughters, does anyone match ?
                    bool found_a_matching_daughter = false;
                    PropertyNodeList::const_iterator diter;
                    for ( diter =  _daughters->begin();
                          diter != _daughters->end()  && !found_a_matching_daughter;
                          ++diter )
                    {
                        found_a_matching_daughter = (*diter)->provides( *citer, false );
                    }
                    match = found_a_matching_daughter;
                }
            } else {
                match = false;
            }
        }
    }


    if ( !match && _daughters && recurse ) {
        PropertyNodeList::const_iterator iter;
        for ( iter =  _daughters->begin();
              iter != _daughters->end();
              ++iter )
        {
            match = (*iter)->provides( caps, true );
        }
    }

    return match;
}
