/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, Rolf.Dubitzky@gmx.net
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/**
 * \file 
 */

// config
#if defined( HAVE_CONFIG_H )
#    include "config.h"
#endif

// this header
#include "scenebuilder.hh"

// std c++
#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <string>
using std::string;
#include <sstream>
#include <algorithm>


// PIAVE
#include "piave_base.hh"
#include "storybase.hh"
#include "mediaelements.hh"
#include "property.hh"
#include "streammgr.hh"
#include "pluginmgr.hh"
#include "operatorbase.hh"

using namespace PIAVE;

StorySeqNode * 
SceneBuilder::buildScenelist( const PropertyNode * scenelist )
{
    // FIXME  this is just a temporary hack
    //        will be replaced by some kind of factory ...
    INFO( "start building render tree" );
    StorySeqNode * head = new StorySeqNode();
    int i = 0;
    for ( const PropertyNode * scene = scenelist->getDaughter( "scene" );
          scene != 0;
          scene = scenelist->getDaughter( "scene", ++i ) ) 
    {
        INFO( "building scene" );
        double duration;
        if ( !scene->getAttr( "duration", duration ) ) {
            // FIXME in case of missing duration try to get it from chunk
            //  and make WARN -> INFO
            WARN( "missing duration" );
        }

        const PropertyNode * dummy = scene->getDaughter( 1 );
        if ( dummy ) {
            WARN( "scene must have exactly one element, " 
                  "all further elements will be ignored" );
        }
        
        const PropertyNode * sceneContent = scene->getDaughter( 0 );
        MediaElement * element = buildMediaNode( sceneContent );
        element->setStart( head->length() );
        element->setLength( duration );
        head->addElement( element );
    }
    return head;
}



MediaElement * 
SceneBuilder::buildMediaNode( const PropertyNode * mediaNode )
{
    MediaElement * node = 0;

    /* What kind of node? */

    /* VEML 0.1  "video" -> AVChunk */
    if ( mediaNode->getName() == "input" ||
         mediaNode->getName() == "video" ||
         mediaNode->getName() == "audio" ) 
    {
        node = buildInputNode( mediaNode );
    }

    /* VEML 0.2 */
    if ( mediaNode->getName() == "par" ) 
    {
        node = buildParNode( mediaNode );
    }

    //  and replaceed by:
    if ( node == 0 ) 
    {
        /* try plugins to handle this node */
        node = buildPluginNode( mediaNode );
    }

    if ( node == 0 ) 
    {
        WARN( "Don't know how to handle node: "<<mediaNode->getName() );
    }

    return node;   
}

MediaElement * 
SceneBuilder::buildInputNode( const PropertyNode * inputNode )
{
    INFO( "building input" );
    string src;
    double inpoint = 0;
    double outpoint = -1;
    bool found = false;

    // "file" is equivalent to "src"
    found = inputNode->getAttr( "file", src );
    if ( !found ) 
    {
        found = inputNode->getAttr( "src", src );
    }
    inputNode->getAttr( "inpoint", inpoint );
    inputNode->getAttr( "outpoint", outpoint );
    InAVFStream * s1 = PIAVE::StreamMgr::getAVInput( src );
    AVChunk * avChunk = 0;
    if ( s1 ) {
        avChunk = new PIAVE::AVChunk( s1 );
        avChunk->setAVFStreamOffset( inpoint );
        if ( outpoint>inpoint ) {
            avChunk->setLength( outpoint-inpoint );
        }
        avChunk->fixMe( 1 );
    }
    INFO(PRINTV( outpoint)<<PRINTV(inpoint )<<PRINTV(avChunk->length()));
    return avChunk;
}


MediaElement * 
SceneBuilder::buildPluginNode( const PropertyNode * pluginNode )
{
    MediaElement * mediaElement = 0;
    
    /* this is supposed to be an plugin, do we know it? */
    Plugin * pluginHandler = PluginMgr::find( pluginNode->getName() );
    if ( pluginHandler ) 
    {
        PluginBase * mediaElement = pluginHandler->newElement( pluginNode );
        // FIXME
        return (MediaElement*) mediaElement;

    } else {

        WARN( "Unknown VEML node: \""<< pluginNode->getName()<<"\". Missing plugin?" );
    }

    return mediaElement;
}


MediaElement * 
SceneBuilder::buildEffectNode( const PropertyNode * effectNode )
{
    INFO( "buildEffectNode start" );
    double nTracks = 0;
    effectNode->getAttr( "tracks", nTracks );
    if ( nTracks == 0 ) {
        return 0;
    }

    /*
     *   find the video operator
     */
    PluginBase * videoOp = 0;
    const PropertyNode * videoNode = effectNode->getDaughter( "video" );
    string videoOpName = "";
    if ( videoNode ) {
        videoNode->getAttr( "effect" , videoOpName );
    }

    Plugin * pluginHandler = 0;
    if ( videoOpName == "" ) {
        WARN( "No effect name specified for video op" );
    } else {
        pluginHandler = PluginMgr::find( videoOpName );
        if ( pluginHandler ) {
            INFO( "this schould be the only one!" );
            videoOp = pluginHandler->newElement( videoNode );
            INFO( "this schould be the only dd one!" );
        } else {
            WARN( "Unknown video effecet: \""<<videoOpName<<"\". Missing plugin?" );
        }
    }

    /*
     *   find the audio operator
     */
    PluginBase * audioOp = 0;
    const PropertyNode * audioNode = effectNode->getDaughter( "audio" );
    string audioOpName = "";
    if ( audioNode ) { 
        audioNode->getAttr( "effect" , audioOpName );
    }
    if ( audioOpName == "" ) {
        WARN( "No effect name specified for audio op" );
    } else {
        pluginHandler = PluginMgr::find( audioOpName );
        if ( pluginHandler ) {
            audioOp = pluginHandler->newElement( audioNode );
        } else {
            WARN( "Unknown audio effecet: \""<<audioOpName<<"\". Missing plugin?" );
        }
    }

    /*  
     *   get inputs  
     */
    const PropertyNode * inputs = effectNode->getDaughter( "inputs" );
    if ( !inputs ) 
    {
        WARN( "No inputs for effect specified!" );
        return 0;
    }

    MediaElement * effectElement = 0;
    if ( nTracks == 1 ) {
        INFO( "creating UNode" );
        MediaElement * A = buildMediaNode( inputs->getDaughter( 0 ) );
        // FIXME check for input has exactly one daugher
        // FIXME check caps for num tracks audio/video etc..
        // FIXME make the following casts typesafe
        UnaryOp * v = (UnaryOp*) videoOp;
        effectElement = new StoryUNode( A, v , (UnaryOp*) audioOp );
    } else if ( nTracks == 2 ) {
        INFO( "creating BNode" );
        MediaElement * A = buildMediaNode( inputs->getDaughter( 0 ) );
        MediaElement * B = buildMediaNode( inputs->getDaughter( 1 ) );
        // FIXME check for input has exactly one daugher
        // FIXME check caps for num tracks audio/video etc..
        // FIXME make the following casts typesafe
        effectElement = new StoryBNode( A, B, (BinaryOp*) videoOp, (BinaryOp*)audioOp );
    } else {
        // FIXME search plugins for effects with >2 tracks
        WARN( "a maximum of two tracks is supported per node" );
    }

    
    effectElement->fixMe(1);

    INFO( "buildEffectNode start" );
    return effectElement;
}


MediaElement * 
SceneBuilder::buildParNode( const PropertyNode * parNode )
{
    INFO( "buildParNode" );
    string src;
    double inpoint = 0;
    double outpoint = -1;
    bool found = false;


    // FIXME only audio/video nodes allowed, must extend to 
    //       containers/effects

    int nVideo=0, nAudio=0;
    MediaElement *video[2], *audio[2];

    for (int i=0; i<parNode->getNDaughters(); i++ )
    {
        const PropertyNode * d = parNode->getDaughter( i );
        if ( d->getName()== "video" && nVideo<2 )
        {
            video[ nVideo++ ] =  buildInputNode( d );
        }
        if ( d->getName()== "audio" && nAudio<2 )
        {
            audio[ nAudio++ ] =  buildInputNode( d );
        }
        // FIXME: support effects here
    }

    if ( nVideo == nAudio == 0 )
    {
        WARN( "<par> cotainer is missing any input" );
        return 0;
    }

    if ( nVideo+nAudio == 1 )
    {  /* this is not really a <par> container, sice it has only one input */
        if ( nVideo==1 ) return video[0];
        if ( nAudio==1 ) return audio[0];
    }


    if ( nVideo+nAudio != 2 )
    {
        ABORT("???");
    }

    MediaElement * parElement = 0;
    if ( nVideo==2 )
    {
        // FIXME: support video effects
        parElement = new StoryBNode( video[0], video[1] );
    } else if ( nAudio==2 ) {
        // FIXME: support audio effects
        parElement = new StoryBNode( audio[0], audio[1] );
    } else {
        INFO("hi");
        parElement = new StoryBNode( video[0], audio[0], 0, new ReplaceAudioOp() );
    }

    return parElement;
}

