/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, Rolf.Dubitzky@gmx.net
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**
 * \file server.cc implaments server.hh
 */

// config
#if defined( HAVE_CONFIG_H )
#    include "config.h"
#endif


// this
#include "server.hh"

// std c
#include <errno.h>

// std c++
#include <sstream>
using std::string;
using std::cout;
using std::cerr;
using std::endl;

using namespace PIAVE;

string Server::whitespace = " \n\r\f\t";
Server * Server::Inst = 0;

Server::Server( unsigned short port ) :
  _state( state_not_initialized ),
  m_port( port ),
  m_listenSocket( -1),
  m_commSocket( -1 ),
  m_messageBuffer(""),
  m_totalBytesRead(0)
{
    WARN_IF( Inst != 0 , "Only one server is supported" );
    Inst = this;
    atexit( PIAVE::Server::cleanup_at_exit );
}

Server::~Server()
{
    closeCommSocket();
    closeListenSocket();
    Inst = 0;
}

void
Server::cleanup_at_exit()
{
    WARN( "atexit handler:" );
    if ( Inst ) {
        Inst->closeCommSocket();
        Inst->closeListenSocket();
    }
    return;
}

int 
Server::execute()
{
    if ( ! ( _state == state_not_initialized ||
             _state == state_not_connected ) ) {
        WARN( "state error" );
        return EXIT_FAILURE;
    }

    while( _state != state_shutting_down &&
           _state != state_error_connecting )
    {
        switch ( _state ) 
        {
          case state_not_initialized:
              initListenSocket();
              if ( m_listenSocket == -1 ) {
                  setState( state_error_connecting );
              } else {
                  setState( state_not_connected );
              }
              break;
            
          case state_not_connected: {
              // Wait for an incoming connection 
              sockaddr_in incomingAddress;
              socklen_t length = sizeof( incomingAddress );
              int socket = accept( m_listenSocket, (sockaddr *)&incomingAddress, &length );
              if ( socket==-1 ) {
                  if ( errno == EAGAIN ) {
                      // don't change state, just try again later
                      sleep( 1 );
                  } else {
                      setState( state_error_connecting );
                  }
              } else {
                  INFO( "Got connection on socket " << socket );
                  closeListenSocket();
                  m_commSocket = socket;
                  setState( state_connected );
              }
              break;
          }

          case state_connected:
          case state_process_commands:
              /* process commands. This includes both sending, recieving and acknowledging. */
              setState( state_process_commands );
              startProcessCommandsLoop();
              break;
            
          default:
              WARN( "state error" );
        }

    } 

    if ( _state == state_error_connecting ) {
        WARN("Error accepting connection");
    }

    closeCommSocket();

    setState( state_shutting_down );
    return EXIT_SUCCESS;
}

void Server::initListenSocket()
{
    INFO( "initListenSocket" );
    
    m_socketAddress.sin_family = AF_INET;         // host byte order
    m_socketAddress.sin_port = htons(m_port);     // short, network byte order
    m_socketAddress.sin_addr.s_addr = htonl(INADDR_ANY);
    memset(&(m_socketAddress.sin_zero), '\0', 8); // zero the rest of the struct
	
    m_listenSocket = socket(AF_INET, SOCK_STREAM, 0);
    if(m_listenSocket==-1) {
        WARN("Error initialising server");
        closeListenSocket();
        return;
    }

    if(bind(m_listenSocket, (sockaddr *)&m_socketAddress, sizeof(m_socketAddress)) == -1) {
        WARN("Error initialising server");
        closeListenSocket();
        return;
    }

    if(listen(m_listenSocket, 5) == -1) {
        WARN("Error initialising server");
        closeListenSocket();
        return;
    }

    fcntl(m_listenSocket, F_SETFL, O_NONBLOCK);
    INFO( "Listening on port " << m_port );
}

void Server::closeListenSocket()
{
    if ( m_listenSocket != -1 ) {
        INFO( "Stopp listening for connections" );
        close( m_listenSocket );
        m_listenSocket = -1;	
    }
}

void 
Server::closeCommSocket()
{
    if ( m_commSocket != -1 ) {
        INFO( "Close communication socket" );
        close( m_commSocket );
        m_commSocket=-1;
    }
}

void 
Server::sendMessage( const string & str )
{
    string mstr = str;
    int idx = 0;
    int length = mstr.length();
    while( idx >= 0 ) {
        idx = mstr.find( "\n\n", idx );
        if ( idx>=0 ) {
            mstr.replace( idx, 2, "\n", 1 );
        }
    }
    mstr = mstr + string( "\n\n" );
//     INFO( "<<<<<<<<<< sending message start" );
//     INFO( mstr );
//     INFO( "<<<<<<<<<< sending message stop" );
    send( m_commSocket, mstr.c_str(), mstr.length(), 0 );
}

int
Server::receiveMessage( string & str )
{
    static char cbuff[5024];

    int nBytesRead = recv( m_commSocket, cbuff, 5024, 0);
    
    if ( nBytesRead == -1 ) {
        WARN( "error on recv "<<errno );
    } else if ( nBytesRead == 0 ) {
        INFO( "Connection closed by client" );
        setState( state_shutting_down );
    }
    cbuff[ nBytesRead ] = 0;

    // This somewhat subverts the recieveMessage interface, since numBytes will not mean
    // quite what it did before - sometimes bytes will be recieved and held onto until later.
    // Perhaps this code should be moved to kdenserver? --jmw
    m_messageBuffer += cbuff;

    // We search for the *last occurance of newlines in the buffer, we want to return every complete
    // command that we find, not just the first one.
    int endMessage = m_messageBuffer.rfind("\n\n");

    if(endMessage != string::npos)
    {
    	str = "<message>";
    	str += m_messageBuffer.substr(0, endMessage);
    	str += "</message>\n";
	m_messageBuffer = m_messageBuffer.substr(endMessage, m_messageBuffer.length()-endMessage);
	// return nBytesRead which is length of entire message.
	int temp = nBytesRead - endMessage;
	nBytesRead = m_totalBytesRead + endMessage;
	m_totalBytesRead = temp;
    } else {
        str = "";
	m_totalBytesRead += nBytesRead;
	nBytesRead = 0;
    }


//     INFO( ">>>>>>>>> receive message start" );
//     INFO( str );
//     INFO( ">>>>>>>>> receive message stop" );

    return nBytesRead;
}


