/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 *
 * This class is based in the nullcutter.cc example from
 * kdenlive by Jason Wood
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#if ! defined ( SERVER_HH )
#define SERVER_HH

// std
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>

// std++
#include <iostream>
#include <string>

// piave
#include "libpiave/piave_base.hh"
#include "libpiave/storybase.hh"


namespace PIAVE {

    /**
     * \class Server 
     * listens on a tcp port for incomming commands.
     *
     * This base class handles the necessary tcpsetup required for
     * the renderer to connect to a client
     * 
     * This class is implemented as a "state mashine".
     * 
     *       state_not_initialized                              <-\
     *         ->  state_not_connected,                   <---\   |
     *               ->  state_connected,           <----\  --/   |
     *                     ->  state_process_commands, --/        |
     *                            ->  state_shutting_down      ---/
     *         ->   state_error_connecting,
     *
     *
     * @author Jason Wood, Rolf Dubitzky
     *
     *
     */
    class Server
    {
      public:
        enum State { 
            state_not_initialized, 
            state_not_connected, 
            state_error_connecting,
            state_connected, 
            state_process_commands, 
            state_shutting_down 
        };

      public: 
        Server( unsigned short port );
        virtual ~Server();
        /** Starts the execution of the cutter. This is the main loop;
         *  by the time this function is called, the server is already
         *  listening for connections. */
        int execute();
        
        /** Sends a message down the communication socket. Convenience
         * function. */
        void sendMessage( const std::string & str );
        int receiveMessage( std::string & str );

      protected:
        /** Process any commands that have been recieved, send any
         * acknowledgements that are required. */
        virtual void startProcessCommandsLoop() = 0;
        void setState( State state ) { _state = state; }
        State getState() const { return _state; };

      private: // Private methods
        /** Initialise a server and start listening for connections */
        void initListenSocket();
        /** Close down the socket that listens for connections. */
        void closeListenSocket();
        /** Close down the socket that communicates with the client.*/
        void closeCommSocket();
        
      private: 
        /** This server is a little state mashine */
        State _state;


        /** The port that the cutter listens in on */
        unsigned short m_port;
        /** The socket used for listening */
        int m_listenSocket;
        /** The socket used for communication */
        int m_commSocket;
        /** The internet address that this socket is bound to. */
        sockaddr_in m_socketAddress;
        /** The command buffer, stores partial commands until the rest of them is picked up. */
        std::string m_commandBuffer;
	
	/** The total bytes read in the recieveMessage method. This resets to 0 whenever a full message
	 * has been recieved. */
	int m_totalBytesRead;

	/** A string buffer which stores recieved messages until they are complete. */
	std::string m_messageBuffer;

        static std::string whitespace;
      public:
        static Server * Inst;
        static void cleanup_at_exit();
    };

};

#endif
