/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**
 * \file storybase.cc implements storybase.hh
 */

// config
#if defined( HAVE_CONFIG_H )
#    include "config.h"
#endif


// stdc
#include <math.h>

// stdc++
#include <iostream>

// this header
#include "storybase.hh"

// PIAVE
#include "avstream.hh"
#include "operatorbase.hh"
#include "bufferpool.hh"
#include "mediaelements.hh"
#include "pluginmgr.hh"

using namespace PIAVE;
using std::string;
using std::cout;
using std::endl;

/**
 * class StoryElement
 *
 *
 */

StoryElement::StoryElement( const string &name, Time start, Time length ) 
  : TimeInterval( start, length ), 
    _name( name ),
    _comment()
{}

/**
 * class MediaElement
 *
 *
 */

MediaElement::MediaElement( const string &name, Time start, Time length ) 
  : StoryElement( name, start, length ),
    PluginBase()
{
};


/**
 * \class StoryBNode
 *
 *
 */

StoryBNode::StoryBNode( MediaElement * A,
                        MediaElement * B,
                        BinaryOp     * videoOp,
                        BinaryOp     * audioOp )
  : MediaElement( "node" ),
    _A( A ),
    _B( B ),
    _videoOp( videoOp ),
    _audioOp( audioOp )
{

    if ( _videoOp && A && B && _videoOp->start() == 0 && _videoOp->length() == 0 ) {
        _videoOp->setTime( A->intersect( *B ) );
    }

    if ( _audioOp && A && B && _audioOp->start() == 0 && _audioOp->length() == 0 ) {
        _audioOp->setTime( A->intersect( *B ) );
    }

    int unfixed = fixMe( 5 );
    ABORT_IF( unfixed>0, name() );
    
    return;
}


Frame
StoryBNode::getFrame( Time t )
{
    Frame f;
    Time localTime = t - start();

    if ( _A->contains( localTime ) &&
         _B->contains( localTime ) )
    {
        Frame src_a = _A->getFrame( localTime );
        Frame src_b = _B->getFrame( localTime );
        if ( _videoOp && _videoOp->contains( localTime ) ) 
            _videoOp->renderFrame( f, src_a, src_b, localTime );
        if ( _audioOp && _audioOp->contains( localTime ) )
            _audioOp->renderFrame( f, src_a, src_b, localTime );
    } else if ( _A->contains( localTime ) ) {
        f = _A->getFrame( localTime );
    } else if ( _B->contains( localTime ) ) {
        f = _B->getFrame( localTime );
    } else {
        ABORT( "time out of bounds!" );
    }
    return f;
}
 

int 
StoryBNode::fixMe( int fix_lvl )
{
    int unfixed = 0;

    if ( _A==0 || _B==0 ) 
    {
        WARN( "media input missing "<<PRINTV(_A)<<PRINTV(_B) );
        return 1;
    }

    /*
     * check that timing informations are sane and consistent 
     */

    /* start */
    Time ABstart = std::min(_A->start(),_B->start());
    if ( ABstart != 0 )
    {
        WARN( "no input started at 0: min(_A->start(),_B->start())="<<ABstart );
        if ( fix_lvl>10 ) 
        {
            _A->setStart( 0 );
            WARN( "fixed _A._start to 0" );
        } else {
            unfixed++;
        }
    }

    /* durations */
    TimeInterval ABjoin = _A->join( *_B );
    if ( ABjoin.length() == 0 ) 
    {
        WARN( "A and B don't overlap !" );
        if ( fix_lvl>10 ) {
            if ( _A->stop() < _B->start() ) 
            {
                _B->setStart( _A->stop() );
                WARN( "fixed _B._start" );
            } else if ( _B->stop() < _A->start() ){
                _A->setStart( _B->stop() );
                WARN( "fixed _A._start" );
            } else {
                WARN( "bogus!" );
                unfixed++;
            }
        } else {
            unfixed++;
        }
    }

    if ( ABjoin.length() != length() )
    {
        WARN( "a.join(B) doesn't cover me exactly! length()="<<stop()
              <<" ABjoin.length="<<ABjoin.length() );
        if ( fix_lvl>0 ) {
            setLength( ABjoin.length() );
            WARN( "fixed my length" );
        } else {
            unfixed++;
        }
    }
       
    /* operator intersections */
    if ( _videoOp )
    {
        TimeInterval ti = _videoOp->intersect( *_A ).intersect( *_B );
        if ( _videoOp->length() > 0 && ti != *_videoOp ) {
            WARN( "op is not completely contained in A intersect B" );
            if ( fix_lvl > 0 ) 
            {
                _videoOp->setTime( ti );
                WARN( "fixed _op bounds" );
            } else {
                unfixed++;
            }
        }
    }

    if ( _audioOp )
    {
        TimeInterval ti = _audioOp->intersect( *_A ).intersect( *_B );
        if ( _audioOp->length() > 0 && ti != *_audioOp ) {
            WARN( "op is not completely contained in A intersect B" );
            if ( fix_lvl > 0 ) 
            {
                _audioOp->setTime( ti );
                WARN( "fixed _op bounds" );
            } else {
                unfixed++;
            }
        }
    }

    unfixed += _A->fixMe( fix_lvl );
    unfixed += _B->fixMe( fix_lvl );
    if ( _videoOp ) unfixed += _videoOp->fixMe( fix_lvl );
    if ( _audioOp ) unfixed += _audioOp->fixMe( fix_lvl );

    return unfixed;
}


void 
StoryBNode::print( std::ostream & o, int indent ) const
{
    string s( indent, ' ' );
    o << s << "StoryBNode:"<<endl;
    o << s << "  inA:"<<endl;
    _A->print( o, indent+4 );
    o << s << "  inB:"<<endl;
    _B->print( o, indent+4 );
    o << s << "  VideoOp"<<endl;
    if ( _videoOp )
    {
        _videoOp->print( o, indent+4 );
    } else {
        o << s << "no videoop" << endl;
    }
    o << s << "  AudioOp"<<endl;
    if ( _audioOp )
    {
        _audioOp->print( o, indent+4 );
    } else {
        o << s << "no audioop" << endl;
    }
    return;
}


/**
 * \class StoryUNode
 *
 *
 */
StoryUNode::StoryUNode( MediaElement * A,
                        UnaryOp      * videoOp,
                        UnaryOp      * audioOp)
  : MediaElement( "node" ),
    _A( A ),
    _videoOp( videoOp ),
    _audioOp( audioOp )
{

    if ( 0==_videoOp ) _videoOp = new NullUOp();
    if ( A && _videoOp && _videoOp->start() == 0 && _videoOp->length() == 0 ) {
        _videoOp->setLength( A->length() );
    }

    if ( 0==_audioOp ) _audioOp = new NullUOp();
    if ( A && _audioOp && _audioOp->start() == 0 && _audioOp->length() == 0 ) {
        _audioOp->setLength( A->length() );
    }

    int unfixed = fixMe( 5 );
    ABORT_IF( unfixed>0, name() );
    
    return;
}



Frame
StoryUNode::getFrame( Time t )
{
    Time localTime = t - start();

    Frame f = _A->getFrame( localTime );

    if ( _videoOp->contains( localTime ) ) {
        _videoOp->renderFrame( f, localTime );
    } else {
        INFO( "Op not in range"<<PRINTV(localTime)<<" "<<(*_videoOp) );
    }
    if ( _audioOp->contains( localTime ) ) {
        _audioOp->renderFrame( f, localTime );
    } else {
        INFO( "Op not in range" );
    }

    return f;
}
 

int 
StoryUNode::fixMe( int fix_lvl )
{
    int unfixed = 0;

    if ( _A == 0 ) {
        WARN( "_A == 0" );
        unfixed++;
    }
    if ( _videoOp == 0 ) {
        WARN( "_video == 0" );
        unfixed++;
    }
    if ( _audioOp == 0 ) {
        WARN( "_audio == 0" );
        unfixed++;
    }

    if ( _A->length() != length() )
    {
        INFO( "_A doesn't cover unary exactly!"
              <<PRINTV(name())<<PRINTV(length())<<PRINTV(_A->length()) );
        if ( fix_lvl>0 ) {
            setLength( _A->length() );
            INFO( "fixed my length" );
        } else {
            unfixed++;
        }
    }

    // allow operators which do not match node boundaries???
    if ( _audioOp->length() != length() )
    {
        WARN( "audioOp doesn't cover unary exactly!"
              <<PRINTV(name())<<PRINTV(length())<<PRINTV(_audioOp->length()) );
        if ( fix_lvl>0 ) {
            _audioOp->setStart( 0 );
            _audioOp->setLength( length() );
            WARN( "fixed op length" );
        } else {
            unfixed++;
        }
    }

    if ( _videoOp->length() != length() )
    {
        WARN( "videoOp doesn't cover unary exactly!"
              <<PRINTV(name())<<PRINTV(length())<<PRINTV(_videoOp->length()) );
        if ( fix_lvl>0 ) {
            _videoOp->setStart( 0 );
            _videoOp->setLength( length() );
            WARN( "fixed op length videoOp="<<_videoOp->name()<<" "<<(*_videoOp) );
        } else {
            unfixed++;
        }
    }

    if ( _A ) unfixed += _A->fixMe( fix_lvl );
    if ( _videoOp ) unfixed += _videoOp->fixMe( fix_lvl );
    if ( _audioOp ) unfixed += _audioOp->fixMe( fix_lvl );

    return unfixed;
}

void 
StoryUNode::print( std::ostream & o, int indent ) const
{
    string s( indent, ' ' );
    o << s << "StoryUNode:"<<endl;
    o << s << "  in:"<<endl;
    _A->print( o, indent+4 );
    o << s << "  VideoOp"<<endl;
    _videoOp->print( o, indent+4 );
    o << s << "  AudioOp"<<endl;
    _audioOp->print( o, indent+4 );
    return;
}



/**
 *
 *  class StorySeqNode
 *
 */


StorySeqNode::StorySeqNode()
  : MediaElement( "mnode" ),
    _elements(),
    _emptyFrame( 0 )
{
    return;
}


StorySeqNode::~StorySeqNode()
{
    std::vector< MediaElement* >::const_iterator eIter;
    for ( eIter = _elements.begin();
          eIter != _elements.end();
          ++eIter
        )
    {
        delete (*eIter);
    }
    return;
}


void
StorySeqNode::addElement( MediaElement * m )
{
    if ( m!=0 ) {
        bool ok = true;

        // check for overlaps
        std::vector< MediaElement* >::const_iterator eIter;
        for ( eIter = _elements.begin();
              eIter != _elements.end();
              ++eIter
            )
        {
            TimeInterval ti = (*eIter)->intersect( *m );
            if ( ti.length() != 0 ) {
                (*eIter)->print( cout , 0 );
                m->print( cout , 0 );
                WARN( "StorySeqNodes must not intersect!" );
                ok = false;
            }
        }
        
        // is no overlaps: append
        if ( ok ) {
            _elements.push_back( m );
        }
        
        // find new length
        Time max_start = -100000;
        Time l = 0;
        for ( eIter = _elements.begin();
              eIter != _elements.end();
              ++eIter
            )
        {
            if ( (*eIter)->start() > max_start ) {
                max_start = (*eIter)->start();
                l = (*eIter)->length();
            }
        }
        setLength( max_start + l );
    }
}
 
Frame
StorySeqNode::getFrame( Time t )
{
    Frame f;
    Time localTime = t - start();

    // search element containing localtime
    std::vector< MediaElement* >::iterator eIter;
    for ( eIter = _elements.begin();
          eIter != _elements.end();
          ++eIter
        )
    {
        if ( (*eIter)->contains( localTime ) ) {
            return (*eIter)->getFrame( localTime );
        }
    }

    /* no element contains localtime 
     * this happens for gaps in the list
     */
    if ( _emptyFrame ) {
        return _emptyFrame->getFrame( 0 );
    }
    
    return f;
}
 

int 
StorySeqNode::fixMe( int fix_lvl )
{
    int unfixed = 0;
    // implement
    return unfixed;
}


void 
StorySeqNode::print( std::ostream & o, int indent ) const
{
    string s( indent, ' ' );
    o << s << "StorySeqNode:  start "<<start()<<"  length "<<length()<<endl;
    std::vector< MediaElement* >::const_iterator eIter;
    for ( eIter = _elements.begin();
          eIter != _elements.end();
          ++eIter
        )
    {
        (*eIter)->print( o, indent + 4 );
    }
    
    return;
}






/**
 *
 *  class StoryBoard
 *
 */
 
StoryBoard::StoryBoard( const string name ) 
  : MediaElement( name ),
    _head( 0 ),
    _emptyFrame( 0 ),
    _nMeasuredFrames( 0 ),
    _sumMean( 0 ),
    _sumSquares( 0 ),
    _fastestFrame( 0 ),
    _slowestFrame( 0 )
{
    Plugin * pluginHandler = PluginMgr::find( "stillcolor" );
    if ( pluginHandler ) 
    {
        _emptyFrame = (MediaElement*)pluginHandler->newElement();
    } else {
        WARN( "plugin \"stillcolor\" is missing! This may cause a seg.fault." );
    }
}

StoryBoard::~StoryBoard()
{
    if(_head) {
        INFO("Deleting HEAD");
        PDELETE( _head );
    }
}

void
StoryBoard::setHead( MediaElement * s )
{
    if(_head) {
        INFO( "Deleting HEAD");
	PDELETE( _head );
    }
    _head = s;
    if ( _head == 0 ) return;
    _head->setStart( 0 );
    setLength( _head->length() );
    INFO( "setting HEAD with length="<<length() );
    int unfixed = _head->fixMe( 5 );
    ABORT_IF( unfixed>0, name() );
    return;
}


Frame
StoryBoard::getFrame( Time t )
{
    if ( !_head ) {    
        Frame f;
        if ( _emptyFrame ) {
            f = _emptyFrame->getFrame( t );
        } 
        return f;
    }

    clock_t start = clock();

    Frame f = _head->getFrame( t );

    clock_t stop = clock();

    if ( start != clock_t(-1) && stop != clock_t(-1) ) {
        _nMeasuredFrames++;
        double elapsed = double( stop - start );
        if ( elapsed>_slowestFrame ) 
            _slowestFrame = elapsed;
        if ( elapsed<_fastestFrame || _fastestFrame==0 ) 
            _fastestFrame = elapsed;
        _sumMean += elapsed;
        _sumSquares += elapsed*elapsed;
    }        

    return f;
}

PropertyNode * 
StoryBoard::getProperties( PropertyNode * p ) const
{
    // FIXME
    return p;
}

void 
StoryBoard::print( std::ostream & o, int indent ) const
{
    string s( indent, ' ' );
    o << s << "StoryBoard:"<<endl;
    o << s << "  head:"<<endl;
    if ( _head ) {
        _head->print( o, indent+4 );
    } else {
        o << "    none"<<endl;
    }
    return;
}


void
StoryBoard::printStats()
{
    double mean = _sumMean/_nMeasuredFrames;
    double rms = sqrt( _sumSquares/_nMeasuredFrames - mean );
    double unit = CLOCKS_PER_SEC/1000;

//     cout << "StoryBoard statistics ("
//          << (Global::preview?"half":"full")
//          << " width) :"<<endl;
    // FIXME print render aize
    cout << "| sum             : " 
         << _sumMean/unit  << " msec"<<endl;
    cout << "| measured frames : " << _nMeasuredFrames << endl;
    cout << "| frames per sec  : " << 1000.0/(mean/unit) << endl;
    cout << "| mean            : " 
         << mean/unit << " msec"<<endl;
    cout << "| RMS             : " 
         << rms/unit << " msec"<<endl;
    cout << "| fastest         : " 
         << _fastestFrame/unit  << " msec"<<endl;
    cout << "| slowest         : " 
         << _slowestFrame/unit  << " msec"<<endl;

}
