/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/** 
 * \file storybase.hh contains basic classes for the StoryBoard 
 * \see StoryElement MediaElement UnaryOp BinaryOp
 *
 */

#if ! defined ( STORYBASE_HH )
#define STORYBASE_HH

// std
#include <iostream>
#include <string>

// piave
#include "piave_base.hh"
#include "avstream.hh"
#include "frame.hh"
#include "property.hh"

namespace PIAVE {

    class UnaryOp;
    class BinaryOp;

    /**
     * The StoryElement is the abstract base class for all objects that can be element of a StoryBoard
     */
    class StoryElement : public TimeInterval
    {
      public:
        /// default constructor, all params have default values
        StoryElement( const std::string &name = std::string(), Time start = 0, Time length = 0 ); 
        /// does nothing yet
        virtual ~StoryElement() {};

        /// get name
        const std::string & name() const { return _name; } 
        /// set name 
        const std::string & setName( const std::string &name ) { 
            _name = name; return _name; } 
        /// get comment
        const std::string & comment() const { return _comment; } 
        /// set comment 
        const std::string & setComment( const std::string &comment ) { 
            _comment = comment; return _comment; } 

        /**
         * check subtree for internal inconsistencies and fix according to \a fix_lvl
         * \param fix_lvl
         *    -   0 : don't fix, just warn
         *    -  >0 : fix not so serious problems
         *             allowed to happen during editing 
         *    - >10 : fix serious problems
         *             should never happen, bug if otherwise
         *             ! might leak memory or corrupt track !
         * \return number of unfixed internal inconsistencies
         */
        virtual int fixMe( int fix_lvl=0 ) { 
            WARN( "ABSTRACT! OVERRIDE ME! StoryElement::fixMe "<<name() ); return 0; };

      private:
        /// the \a _name is supposed to be a short std::string describing the element 
        std::string _name;
        /// the \a _comment can hold a more detailed description / text
        std::string _comment;
    };
    

    /**
     * abstract base class for all elemets that <b>contain</b> data
     * as oposed to operators which only <b>modify</b> data
     */
    class MediaElement : public StoryElement , public PluginBase
    {
      public:
        /// default constructor, all params have default values
        MediaElement( const std::string &name = std::string(), 
                      Time start = 0, Time length = 0 );
        /// does nothing yet
        virtual ~MediaElement() {};

        /**
         * update framebuffer and return it
         * 
         * \param t : local node time
         * \return a pointer to a Frame buffer
         */
        virtual Frame getFrame( Time t ) = 0;
    };

    /**
     * \brief StoryBNode manages two child elements connected via 
     *        a BinaryOp for video and 
     *        a BinaryOp for audio.
     *  There are exactly three principle types of BNodes:
     *   - video-video nodes:
     *        can have an audio and an video op
     *   - audio-audio nodes:
     *        no video op, only audio op
     *   - video-audio nodes:
     *        no video, only audio op
     *
     */
    class StoryBNode : public MediaElement
    {
      public:
        StoryBNode( MediaElement * A,
                    MediaElement * B,
                    BinaryOp     * videoOp = 0,
                    BinaryOp     * audioOp = 0 );

        // override
        virtual Frame getFrame( Time t );
        virtual PropertyNode * getProperties( PropertyNode *p  = 0 ) const { return p; }
        virtual int fixMe( int fix_lvl=0 );

        virtual void print( std::ostream & o, int indent ) const;

      private:
        MediaElement    * _A;
        MediaElement    * _B;
        BinaryOp        * _videoOp;
        BinaryOp        * _audioOp;

    };

    /**
     * \brief StoryUNode manages one single child element
     *        an UnaryOp for video and 
     *        an UnaryOp for audio will be rendered to
     *        the child.
     */
    class StoryUNode : public MediaElement
    {
      public:
        StoryUNode( MediaElement * A,
                    UnaryOp      * videoOp = 0,
                    UnaryOp      * audioOp = 0 );

        //
        virtual Frame getFrame( Time t );
        virtual PropertyNode * getProperties( PropertyNode *p  = 0 ) const { return p; }
        virtual int fixMe( int fix_lvl=0 );

        virtual void print( std::ostream & o, int indent ) const;

      private:
        MediaElement  * _A;
        UnaryOp       * _videoOp;
        UnaryOp       * _audioOp;
    };


    /**
     * \class StorySeqNode manages any number of nodes which are in
     * sequential order and have no operator connecting them.  The
     * contained nodes may not overlap. Gaps between to contained
     * nodes are filled with _emptyFrame
     */
    class StorySeqNode : public MediaElement
    {
      public:
        StorySeqNode( );
        virtual ~StorySeqNode( );

        //
        void setEmptyFrame( MediaElement * m ) { _emptyFrame=m; }
        void addElement( MediaElement * m );

        //
        virtual Frame getFrame( Time t );
        virtual PropertyNode * getProperties( PropertyNode *p  = 0 ) const { return p; }
        virtual int fixMe( int fix_lvl=0 );
        virtual void print( std::ostream & o, int indent ) const;

      private:
        std::vector< MediaElement* >  _elements;
        MediaElement * _emptyFrame;
        
    };


    /**
     *  a small wrapper around the root of the render tree
     */
    class StoryBoard : public MediaElement
    {
      public:
        StoryBoard( const std::string name );
	virtual ~StoryBoard();

        MediaElement * getEmptyFrame() const { return _emptyFrame; }

        virtual Frame getFrame( Time t );
        
        virtual PropertyNode * getProperties( PropertyNode * p  = 0 ) const;

        /// \todo implement me
        virtual int fixMe( int fix_lvl=0 ) { return 0; };

        virtual void print( std::ostream & o, int indent ) const;

        void setHead( MediaElement * s );
        
        void printStats();

      private:
        MediaElement * _head;
        MediaElement * _emptyFrame;
        // stats
        int _nMeasuredFrames;
        double _sumMean;
        double _sumSquares;
        double _fastestFrame;
        double _slowestFrame;
     };

};

#endif
