/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**
 * \file avstream.hh declares the in and out streaming interface for PIAVE
 */

#if ! defined( STREAMMGR_HH )
#define STREAMMGR_HH

// stdc
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>

// stdc++
#include <string>
#include <map>
#include <vector>

// PIAVE
#include "piave_base.hh"

namespace PIAVE {

    class PropertyNode;
    class Stream;
    class InAVStreamIFace;
    class InAudioStreamIFace;
    class InAVFStream;
    class InAudioFStream;
    class OutAVStreamIFace;
    class OutAudioStreamIFace;

    /**
     *  Class StreamMgr
     *  Factory for all piave in- and out- filestreams. If you want to open a stream, 
     *  use StreamManaget::OpenXxxStream( filename ). He will check a list, of
     *  open / existing streames and files.
     *   Streams are identified by their name, i.e. fileNmae for files, or generic
     *  names for X shared mem or Xv visuals
     *
     *  ToDo : 
     *     teach about Output streams
     *     file type detection
     */

    class StreamMgr
    {
      private:
        StreamMgr();

      public:
        static StreamMgr * Inst() {
            if (!_singleton) _singleton = new StreamMgr();
            return _singleton;
        }

        /* get video input, with properties described in p */        
        static InAVStreamIFace * getAVInput( const PropertyNode * p );
        /* get audio input, with properties described in p */        
        static InAudioStreamIFace * getAudioInput( const PropertyNode * p );

        /* shortcut for above to get input direct from file
         * the StreamMgr will try to find a plugin that can handle
         * the format of <filename>.
         */
        static InAVFStream * getAVInput( const std::string& fileName );
        static InAudioFStream * getAudioInput( const std::string& fileName );

        /* in case we don't know if it's audio or video */
        static Stream * getInput( const std::string& fileName );

        /* get av output, with properties described in p */        
        static OutAVStreamIFace * getOutAVStream( const PropertyNode * p );
        static OutAudioStreamIFace * getOutAudioStream( const PropertyNode * p );

        /* try to expand a given filename with all existing path entries
         * \return true if file was found, false otherwise
         */
        static bool expandFileName( const::string fileName );
        
        /* add a directory to the list of dirs to search for files */
        static void addPathEntry( const::string path ) { Inst()->_pathList.push_back( path ); }
        
      private:
        /* try to find the appropriate stream plugin */
        InAVFStream * openInAVFStream( const std::string& fileName );
        
      private:
        static StreamMgr * _singleton;

        /* files are opened just once. _One_ stream per file.
         * this stream may be used in many AVChunks
         */
        std::map< std::string, InAVFStream* > _inAVStreams;

        /* a list of paths to search for files */
        std::vector< string > _pathList;
    };

};

#endif
