/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/** 
 * \file frame.hh
 * Code for handling video and audio frames
 *
 * Classes in this file represent audio and video frames and all operations 
 * related to allocation and conversion of raw data buffers.
 *
 *
 * based on:
 *
 * frame.h -- utilities for process digital video frames
 * Copyright (C) 2000 Arne Schirmacher <arne@schirmacher.de>
 */

#if ! defined ( VIDEOFORMAT_HH )
#define VIDEOFORMAT_HH

// std c
#include <stdint.h>

// std c++
#include <string>
#include <vector>

namespace PIAVE {

    struct FOURCC {
        explicit FOURCC( uint32_t c = 0 ) : fourcc( c ) {};
        explicit FOURCC( const char c[5] ) : 
          fourcc( (c[0]) + (c[1]<<8) + (c[2]<<16) + (c[3]<<24) ) {}

         bool operator==( const FOURCC & o ) const { 
            return o.fourcc==fourcc; }
         bool operator!=( const FOURCC & o ) const { 
            return !(o.fourcc==fourcc); }
        std::string toString() const;

        uint32_t fourcc;
        static const FOURCC NONE;
        static const FOURCC RGB;
        static const FOURCC YUY2;
    };
    std::ostream& operator<<( std::ostream&, const FOURCC& );


    class PixFmt {
      public:

        enum alpha_e { flat, alpha };

        PixFmt( FOURCC fourcc, alpha_e alpha = flat ) :
          _fourcc( fourcc ), _alpha( alpha ) {};

        alpha_e getAlpha() const { return _alpha; }
        FOURCC getFOURCC() const { return _fourcc; }

        bool operator == ( const PixFmt & o ) const {
            return (_fourcc==o._fourcc && _alpha==o._alpha); }

        PixFmt & operator = ( const PixFmt & o ) {
            _fourcc=o._fourcc; _alpha==o._alpha; return (*this); }

      private:
        FOURCC     _fourcc;
        alpha_e    _alpha;

      public:
        static const PixFmt NONE;
        static const PixFmt RGB;
        static const PixFmt RGB_A;
        static const PixFmt YUY2;
        static const PixFmt YUY2_A;

    };
    std::ostream& operator<<( std::ostream&, const PixFmt& );


    class VideoFmt
    {
      public:
        VideoFmt() : width( 0 ), height( 0 ), 
                     aspect( -1 ), fps( -1 ), fourcc( FOURCC::NONE ) {}
        double         getFrameLength() const { return 1.0/fps; }
        unsigned int   width;
        unsigned int   height;
        double         aspect;
        double         fps;      //   <0 :  means non-fixed
        FOURCC         fourcc;
    };

}

            
#endif
