/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/** 
 * \file filter.cc implements filter.hh
 */

// config
#if defined( HAVE_CONFIG_H )
#    include "config.h"
#endif

// std c++
#include <iostream>
#include <sstream>
using std::string;
using std::cout;
using std::cerr;
using std::endl;

#include "image.hh"
#include "libpiave/property.hh"
#include "libpiave/bufferpool.hh"
#include "libpiave/algorithm.hh"

using namespace PIAVE;

/*
 *   Plugin handler for invert filter
 *
 *
 */
Plugin * new_ColorStillPlugin() { return new ColorStillPlugin(); };
PluginInfo pluginInfo[] = { 
    { "ColorStillPlugin", PLUGIN_IFACE_VERSION , &new_ColorStillPlugin } ,
    { 0, 0, 0 }
};


ColorStillPlugin::ColorStillPlugin()
 : Plugin( "stillcolor", PLUGIN_MEDIA_ELEMENT ) 
{
    PropertyNode * caps = new PropertyNode( "capabilities" );
    caps->addAttr( "name", getName() );

    PropertyNode *p, *pp;
    p = caps->addDaughter( "parameter" );
    p->addAttr( "name", "yuvcolor" );
    p->addAttr( "type", "yuvcolor" );
    p = caps->addDaughter( "parameter" );
    p->addAttr( "name", "rgbcolor" );
    p->addAttr( "type", "rgbcolor" );
    p = caps->addDaughter( "parameter" );
    p->addAttr( "name", "alpha" );
    p->addAttr( "type", "alphachannel" );

    // FIXME comunicate default values

    p = caps->addDaughter( "about" );
    p->setContent( "A simple color with alpha channel.\n" );

    setCapabilities( caps );
};

PluginBase * 
ColorStillPlugin::newElement( const PropertyNode * p )
{
    ColorStill * cs = new ColorStill( "ColorStill" );

    /* if we didn't get properties, we're done */
    if (p==0) 
        return cs;

    /* get and set properties */
    string color, alpha;
    bool gotit = false;
    gotit = p->getAttr( "yuvcolor" , color );
    if ( !gotit ) 
    {
        gotit = p->getAttr( "rgbcolor" , color );
        if ( gotit ) 
        {
            cs->_color = Global::algo->RGB2YCbCr( Color( color ) );
        }
    } else {
        cs->_color = Color( color );
    }
        
    gotit = p->getAttr( "alpha" , alpha );
    if ( gotit ) 
    {
        cs->setUseAlpha( true );
        cs->_color[3] = Global::hexCharToI( alpha[1], alpha[2] );
    }

    /* done */
    return cs;
}


/**
 *
 *  Class ColorStill
 *
 */
ColorStill::ColorStill( const string& name )
  : MediaElement( name ),
    _initialized( false ),
    _useAlpha( false ),
    _vb(0)
{
    _color = Global::algo->RGB2YCbCr( Color::black );
}

ColorStill::~ColorStill()
{
    PDELETE( _vb );
}

void
ColorStill::_init()
{
    /// FIXME \todo this is completely broken with YUV/RGB !!
    _initialized = true;
    _vb = new VideoBuffer( Global::renderFmt.width, Global::renderFmt.height, true );
    if ( useAlpha() ) _vb->addAlpha();
    // this frame should not be overwritten
    _vb->lock();

    unsigned char * p = _vb->getPixels();
    unsigned char * end = p + _vb->getSizeBytes();
    while ( p<end ) {
        *p++ = _color[0];   // Y
        *p++ = _color[2];   // V
        *p++ = _color[0];   // Y
        *p++ = _color[1];   // U
    }

    if ( useAlpha() ) {
        ABORT( "FIXME" );
    }

    return;
}

Frame
ColorStill::getFrame( Time t )
{
    if ( !_initialized ) _init();
    Frame f;
    f.setVBuf( _vb );
    f.setVideoIsDecoded( true );
    f.setPos( t );
    f.setLength( Global::renderFmt.getFrameLength() );
    return f;
}

PropertyNode * 
ColorStill::getProperties( PropertyNode * p ) const
{
    if (!p) 
        p = new PropertyNode( "ColorStill" );

    p->addAttr( "yuvcolor", _color.toString3() );
    p->addAttr( "alpha", _color.toStringA() );

    return p;
}

void 
ColorStill::print( std::ostream & o, int indent )
{
    string s( indent, ' ' );
    o << s << "ColorStill:"<<endl;
    o << s << "  color: "<< _color <<endl;
    return;
}

#if 0


/**
 * class StillImage
 *
 *
 *
 */

StillImage::StillImage( const string& filename, int x_off, int y_off )
  : MediaElement( filename ),
    _x_off( x_off ), _y_off( y_off )
{
    _init();
}

StillImage::~StillImage()
{
    /// \todo delete some stuff
}


void
StillImage::_init()
{

//     // read to local buffer
//     _pixbuf = gdk_pixbuf_new_from_file( name().c_str() );
//     WARN_IF( !_pixbuf , "problem reading \""<<name()<<"\"" )
//     if ( _pixbuf ) {
//         INFO( "construct StillImage:" );
//         INFO( "  n_channels="<<gdk_pixbuf_get_n_channels( _pixbuf ) );
//         INFO( "  has_alpha="<<gdk_pixbuf_get_has_alpha( _pixbuf ) );
//         INFO( "  bits_per_sample="<<gdk_pixbuf_get_bits_per_sample( _pixbuf ) );
//         INFO( "  width="<<gdk_pixbuf_get_width( _pixbuf ) );
//         INFO( "  height="<<gdk_pixbuf_get_height( _pixbuf ) );
//         INFO( "  rowstride="<<gdk_pixbuf_get_rowstride( _pixbuf ) );
//     }
//     gboolean has_alpha = gdk_pixbuf_get_has_alpha( _pixbuf );
    _vb = new VideoBuffer( Global::renderFmt.width, Global::renderFmt.height );
    // this frame should not be overwritten
    _vb->lock();

    // FIXME !! RD allow for negative offset
    ABORT_IF( _x_off < 0 || _y_off < 0 , "can't handle! FIXME" );

    // copy to the framebuffer
//    double scale = (Global::preview?0.5:1.0);

//     int rw = std::min( Global::render_width, 
//                  unsigned( scale * gdk_pixbuf_get_width( _pixbuf ) ) );
//     int rh = std::min( Global::render_height, 
//                  unsigned( scale * gdk_pixbuf_get_height( _pixbuf ) ) );

    ABORT( "FIXME" );
//     gdk_pixbuf_scale( _pixbuf, _vb->getPixBuf(),
//                       _x_off,_y_off,
//                       rw, rh,
//                       _x_off,_y_off,
//                       scale,scale,
//                       GDK_INTERP_NEAREST );

    return;
}

Frame
StillImage::getFrame( Time t )
{
    Frame f;
    f.setVBuf( _vb );
    return f;
}

void 
StillImage::print( std::ostream & o, int indent )
{
    string s( indent, ' ' );
    o << s << "StillImage:"<<endl;
    o << s << "  filename: "<< name() <<endl;
    return;
}

#endif
