/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/** 
 * \file textmaster.cc implements textmaster.hh
 */

// config
#if defined( HAVE_CONFIG_H )
#    include "config.h"
#endif


// this header
#include "textmaster.hh"

// std c++
#include <string>
using std::string;

// PIAVE
#include "libpiave/piave_base.hh"
#include "libpiave/videoformat.hh"

using namespace PIAVE;

PIAVE_PLUGIN( TextMasterPlugin );

/*
 * \class TextMasterPlugin is the plugin handler for the TextMaster Effect
 *
 */
TextMasterPlugin::TextMasterPlugin()
  : Plugin( "textmaster", PLUGIN_VIDEO_OP )
{
    PropertyNode * caps = new PropertyNode( "capabilities" );
    caps->addAttr( "name", getName() );

    PropertyNode * p;
    p = caps->addDaughter( "input" );
    p->addAttr( "input", "A" );
    p->addAttr( "video", "yes" );

    p = caps->addDaughter( "about" );
    p->setContent( "This effect adds text to a video stream.\n" );

    // FIXME autogenerate this from parameterlist
    p = caps->addDaughter( "parameter" );
    p->addAttr( "name", "text" );
    p->addAttr( "type", "string" );
    p->setContent( "The text to render ontop the video image."
                   "Multiple lines of text may be separated by \'\\n\' .\n" );

    setCapabilities( caps );
}

PluginBase *
TextMasterPlugin::newElement( const PropertyNode * p )
{
    INFO("new textmaster");

    TextMaster * textMaster =  new TextMaster();
    textMaster->setParameter( p );

    return textMaster;
}


/*
 *    \class FreeType  why can't people just code C++ ???? .. because it sucks... I know...
 */
void
FreeType::init()
{
    INFO( "FreeType::init()" );
    if ( is_initialized ) return;
    FT_Error error = FT_Init_FreeType( &library );
    if ( error ) {
        WARN( "Error initializing FreeType" );
        return;
    }
    is_initialized = true;
}

bool FreeType::is_initialized = false;
FT_Library FreeType::library;



/*
 * \class TextMaster  yeah yeah!
 *
 */
TextMaster::TextMaster() : 
  UnaryOp( "TextMaster" ),
  _glyphs_initialized( false ),
  _text( "dtext", "PIAVE is a video editor!" ),
  _fontName( "font", "/usr/X11R6/lib/X11/fonts/TTF/luximb.ttf" ),
  _horizAlign( hcenter ),
  _vertAlign( vcenter ),
  _size( "size", 0.05 ),
  _shadow_offset( 0.02, 0.01 ),
  _position( "position" , Box() ),
  _color( "color", Color::white )
{
    addParameter( &_text );
    addParameter( &_fontName );
    _size.setIsDynamic();
    addParameter( &_size );
    _position.setIsDynamic();
    addParameter( &_position );
    _color.setIsDynamic();
    addParameter( &_color );

    FreeType::init();
};

void
TextMaster::renderFrame( Frame & f, Time t )
{
    f.makeWriteableVBuf();
    Time localTime = t - start();

    // FIXME OPTIMIZE: better cache the bitmaps, not the glyphs in
    // case of static operator ...  even better, remember the last
    // values and don't rerender in case nothing changed.
    initGlyphs();
    const Color color = Color::white;// _color.interpolateVal( localTime );
    const Box b =_position.getValue( localTime );


    std::vector< FT_Vector >::const_iterator gp = _glyphPos.begin();
    for( std::vector< FT_Glyph >::const_iterator g = _glyphs.begin();
         g != _glyphs.end();
         ++g )
    {
        FT_Vector pen;
        FT_Glyph  image = (*g); // FIXME why copy this???

        // 
        pen.x = b.lx() + gp->x;
        pen.y = b.ly() + gp->y;
        ++gp;
        FT_Error error = FT_Glyph_To_Bitmap( &image, ft_render_mode_normal,
                                             &pen, 0 );

        if (!error) {
            FT_BitmapGlyph  bit = (FT_BitmapGlyph)image;
            FT_Vector pos;
            pos.x = pen.x + bit->left;
            pos.y = pen.y - bit->top;
            renderGlyphBitmapToFrame( f, bit->bitmap,
                                      pos, color );	   
            FT_Done_Glyph( image );
        } else {
            WARN( "FT_Glyph_To_Bitmap "<<PRINTV(error) );
        }
    }

    return;
}

PropertyNode * 
TextMaster::getProperties( PropertyNode * p ) const
{
    WARN( "xFIXME" );
    // gather info from (dynamic) values
    return 0;
}

int  
TextMaster::fixMe( int fix_lvl )
{
    INFO("fixTextmaster");
    return 0;
}

void 
TextMaster::initGlyphs()
{
    // FIXME, broken in case of dynamic size 
    if ( _glyphs_initialized ) return;
    _glyphs_initialized = true;

    // relative start
    FT_Vector pen;
    pen.x = 0;
    pen.y = 0;

    //FIXME replace by FTC_Lookup_Face 
    FT_Error error = FT_New_Face( FreeType::library,
                                  _fontName.getValue().c_str(),
                                  0,
                                  &_face );
    if ( error ) {
        WARN( "Error New Face FreeType "<<PRINTV(error) );
        return;
    }
    
    error = FT_Set_Char_Size(
        _face,     //    handle to face object 
        0,         //    0 = same as other
        static_cast<int>(_size.getValue()*Global::renderFmt.height*64),   
                   //    char_height in 1/64th of points */
        0,         //    0 = same as before = default = 72
        0 );       //    0 = same as other
    if ( error ) {
        WARN( "Error Char Size FreeType "<<PRINTV(error) );
        return;
    }
    
    FT_Glyph  g;
    FT_UInt   prev_glyph_idx = 0;
    FT_Bool   use_kerning = FT_HAS_KERNING( _face );

    const string & text = _text.getValue();
    for ( std::string::const_iterator c = text.begin(); 
          c != text.end(); ++c ) {

        // GET glyph idx for the characte
        FT_UInt glyph_idx 
            = FT_Get_Char_Index( _face, *c );

        // retrieve kerning distance and move pen position
        if ( use_kerning && prev_glyph_idx && glyph_idx ) {
            FT_Vector  delta;
            FT_Get_Kerning( _face, prev_glyph_idx, glyph_idx,
                            ft_kerning_default, &delta );
            // this is in pixels ....
            // \todo  make pixel independent
            pen.x += delta.x >> 6;
        }

        // store current pen position
        _glyphPos.push_back( pen );

        // load glyph image into the slot. DO NOT RENDER IT !!
        error = FT_Load_Glyph( _face, glyph_idx, FT_LOAD_DEFAULT );
        WARN_IF( error, "on FT_Load_Glyph "<<PRINTV( error ) );

        // extract glyph image and store it in our table
        error = FT_Get_Glyph( _face->glyph, &g );
        _glyphs.push_back( g );
        WARN_IF( error, "on FT_Get_Glyph "<<PRINTV( error ) );

        // increment pen position
        pen.x += _face->glyph->advance.x >> 6;

        // record current glyph index
        prev_glyph_idx = glyph_idx;

    }

}

void 
TextMaster::renderGlyphBitmapToFrame( Frame & f, const FT_Bitmap& glyph_bm, 
                                      const FT_Vector& pos, const Color color )
{
    // FIXME !!
    //    this only works for non-alpha targets
    // FIXME
    //    clip text at frame bounds
    int row_stride = f.getVBuf()->getLineStride();

    uint8_t * row_origin;
    uint8_t * glyph_pixel;

    // FIXME optimize: this is bullshit: render both to same bitmap and cache it
    if ( _shadow_offset.lx() != 0 || _shadow_offset.ly() != 0 ) {
        row_origin = f.getVBuf()->getPixels();
        row_origin += pos.x*3 + (pos.y+_shadow_offset.ly())*row_stride + _shadow_offset.lx();
        glyph_pixel = glyph_bm.buffer;
        // loop vertical
        for ( int row=0; row < glyph_bm.rows; row++) {
            uint8_t * target_pixel = row_origin;
            // loop horizontal
            for (int i=0; i < glyph_bm.width; i++) {
                int alpha = *(glyph_pixel++) / 2;
                if ( 0==alpha ) {
                    target_pixel+=3;
                    continue;
                }
                int aalpha = 255 - alpha;
                *(target_pixel+0) = (uint8_t)((aalpha*(*(target_pixel+0))) >> 8);
                *(target_pixel+1) = (uint8_t)((aalpha*(*(target_pixel+1))) >> 8);
                *(target_pixel+2) = (uint8_t)((aalpha*(*(target_pixel+2))) >> 8);
                target_pixel+=3;
            }
            row_origin += row_stride;
        }    
    }

    row_origin = f.getVBuf()->getPixels();
    row_origin += pos.x*3 + pos.y*row_stride;
    glyph_pixel = glyph_bm.buffer;
    // loop vertical
    for ( int row=0; row < glyph_bm.rows; row++) {
        uint8_t * target_pixel = row_origin;
        // loop horizontal
        for (int i=0; i < glyph_bm.width; i++) {
            int alpha = *(glyph_pixel++);
            if ( 0==alpha ) {
                target_pixel+=3;
                continue;
            }
            int aalpha = 255 - alpha;
            *(target_pixel+0) = (uint8_t)((aalpha*(*(target_pixel+0))+alpha*(color[0])) >> 8);
            *(target_pixel+1) = (uint8_t)((aalpha*(*(target_pixel+1))+alpha*(color[1])) >> 8);
            *(target_pixel+2) = (uint8_t)((aalpha*(*(target_pixel+2))+alpha*(color[2])) >> 8);
            target_pixel+=3;
        }
        row_origin += row_stride;
    }    

}
