/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**
 * \file rawdvstream.cc implements rawdvstream.hh
 */

// config
#if defined( HAVE_CONFIG_H )
#    include "config.h"
#endif

// std++
#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <string>
using std::string;


// PIAVE
#include "libpiave/piave_base.hh"
#include "libpiave/frame.hh"
#include "libpiave/property.hh"
#include "libpiave/bufferpool.hh"
#include "libpiave/plugin.hh"
#include "libpiave/pluginmgr.hh"


// this header
#include "avistream.hh"

using namespace PIAVE;


/*
 *   Plugin info
 */
Plugin * new_InAviDVStreamPlugin() { return new InAviStreamPlugin(); };
Plugin * new_OutAviDVStreamPlugin() { return new OutAviStreamPlugin(); };
PluginInfo pluginInfo[] = { 
    { "InAviDVStreamPlugin", PLUGIN_IFACE_VERSION , &new_InAviDVStreamPlugin } ,
    { "OutAviDVStreamPlugin", PLUGIN_IFACE_VERSION , &new_OutAviDVStreamPlugin } ,
    { 0, 0, 0 }
};



/*
 *   Plugin for instream
 */
InAviStreamPlugin::InAviStreamPlugin() :
  Plugin( "InAviStream", PLUGIN_INSTREAM )
{
    PropertyNode * caps = new PropertyNode( "instream" );
    caps->addAttr( "name", getName() );

    PropertyNode * file = caps->addDaughter( "file" );

    PropertyNode * container = file->addDaughter( "container" );
    container->addAttr( "format", "avi" );
    container->addAttr( "extension", ".avi" );

    PropertyNode * about = caps->addDaughter( "about" );
    about->setContent( "Can parse AVI containers.\n" );

    setCapabilities( caps );
}

PluginBase * 
InAviStreamPlugin::newElement( const PropertyNode * p )
{
    if ( p ) {
        string filename;
        bool ok = p->getAttr( "filename", filename );
        if ( ok ) {
            return new InAviStream( filename );;
        }
    }

    WARN( "missing filename" );
    return 0;
}

/*
 *   Plugin for outstream
 */
OutAviStreamPlugin::OutAviStreamPlugin() :
  Plugin( "OutAviStream", PLUGIN_OUTSTREAM ) 
{
    PropertyNode * caps = new PropertyNode( "outstream" );
    caps->addAttr( "name", getName() );

    PropertyNode * file = caps->addDaughter( "file" );

    PropertyNode * container = file->addDaughter( "container" );
    container->addAttr( "format", "rawdv" );
    container->addAttr( "extension", ".dv" );

    PropertyNode * about = caps->addDaughter( "about" );
    about->setContent( "Can write raw DV files.\n" );

    setCapabilities( caps );
}

PluginBase * 
OutAviStreamPlugin::newElement( const PropertyNode * p )
{
    if ( !p ) {
        WARN( "missing filename" );
        return 0;
    }
    string filename;
    p->getAttr( "filename", filename );
    return new OutAviStream( filename );
}

/*
 *  AVI Container and other avi stuff
 *
 */
AviContainer::AviContainer() :
  file_list( -1 ),
  riff_list( -1 ),
  hdrl_list( -1 ),
  avih_chunk( -1 ),
  movi_list( -1 ),
  junk_chunk( -1 ),
  idx1_chunk( -1 ),

  index_type( -1 ),
  current_ix00( -1 ),

  odml_list( -1 ),
  dmlh_chunk( -1 ),
  isUpdateIdx1( false )

{
    for (int i = 0; i < 2; ++i) {
        indx[i] = new AviSuperIndex();
        memset( indx[i], 0, sizeof(AviSuperIndex) );
        ix[i] = new AviStdIndex;
        memset( ix[i], 0, sizeof(AviStdIndex) );

        indx_chunk[i] = -1;
        ix_chunk[i] = -1;
        strl_list[i] = -1;
        strh_chunk[i] = -1;
        strf_chunk[i] = -1;
    }

    idx1 = new AviSimpleIndex;
    memset( idx1, 0, sizeof(AviSimpleIndex) );

    memset( dmlh, 62, sizeof(int) );

}

AviContainer::~AviContainer()
{
    for(int i=0; i<2; ++i) {
        PDELETE( indx[i] );
	PDELETE( ix[i] );
    }
    PDELETE(idx1);
}

AviDirEntry::AviDirEntry() :
  type( FOURCC::NONE ),
  name( FOURCC::NONE ),
  length( -1 ),
  offset( -1 ),
  parentList( -1 ),
  writtenFlag( 0 )
{}


AviDirEntry::AviDirEntry( const FOURCC& t, const FOURCC& n, int l, int o, int p) : 
  type( t ), 
  name( n ), 
  length( l ), 
  offset( o ), 
  parentList( p ), 
  writtenFlag( 0 )
{}


int
AviContainer::newDirEntry( const FOURCC& type, const FOURCC& name, off_t length, int idx )
{
    /* Put all parameters in an RIFFDirEntry object. The offset is
       currently unknown. */
//    INFO("AviContainer::newDirEntry");
    AviDirEntry thisEntry( type, name, length, 0, idx );

    /* If the new chunk is in a list, then get the offset and size
       of that list. The offset of this chunk is the end of the list
       (parent_offset + parent_length) plus the size of the chunk
       header. */
    if ( idx != RIFF_NO_PARENT_IDX ) {
        AviDirEntry parent = dir[ idx ];
        thisEntry.offset = parent.offset + parent.length + RIFF_HEADERSIZE;
    }

    /* The list which this new chunk is a member of has now increased in
       size. Get that directory entry and bump up its length by the size
       of the chunk. Since that list may also be contained in another
       list, walk up to the top of the tree. */
    while ( idx != RIFF_NO_PARENT_IDX ) {
        AviDirEntry parent = dir[ idx ];
        parent.length += RIFF_HEADERSIZE + length;
        parent.writtenFlag = false;
        dir[ idx ] = parent;
        idx = parent.parentList;
    }

    dir.push_back( thisEntry );

    return dir.size() - 1;
}


int
AviContainer::getDirEntry( const FOURCC& type, int n )
{
    int i, j = 0;
    int count = dir.size();

    for (i = 0; i < count; ++i)
        if (dir[i].type == type)
		{
			if (j == n)
				return i;
			j++;
		}

    return -1;    
}


/*
 *
 *  class InAviStream 
 *
 */
InAviStream::InAviStream( const string& fileName ) throw(exception)
  : InAVFStream( fileName ),
    _dvDecoder( 0 ),
    _hasAviAudio( false ),
    _frameSize( 0 ),
    _container( new AviContainer() )
{
    INFO( "Init AVI container - begin" );

    /* parse the RIFF header */
    parseHeader();

    /*  init the decoder  */
    PropertyNode * dvDecoderProp = new PropertyNode( "decoder" );
    PropertyNode * f = dvDecoderProp->addDaughter("format");

    /* walk through AVI streams and see which formats we can find
     * plugins for.
     */
    Plugin * dvDecoderPlugin = 0;
	AviStreamHeader	aviStreamHeader;
	FOURCC strh("strh");
    int i = -1, n = 0;
	while ( (i = _container->getDirEntry( strh, n++ )) != -1  &&
            dvDecoderPlugin == 0 )
	{
		readChunk( i, (void*)&aviStreamHeader );
//         INFO( "found AVI chunk "<<PRINTV(aviStreamHeader.fccHandler)
//               <<PRINTV(aviStreamHeader.fccHandler.fourcc));

        if ( aviStreamHeader.fccHandler.fourcc != 0 ) {
            f->addAttr("fourcc", aviStreamHeader.fccHandler.toString() );
            dvDecoderPlugin = 
                PluginMgr::find( dvDecoderProp, PLUGIN_VIDEO_CODEC );
//             if ( dvDecoderPlugin ) {
//                 INFO( "found plugin for stream "<<aviStreamHeader.fccHandler );
//             } else {
//                 WARN( "NO plugin for stream "<<aviStreamHeader.fccHandler );
//             }        
        }        
    }

    delete dvDecoderProp;

    if ( !dvDecoderPlugin ) {
        INFO( "Couldn't get native piave decoder for AVI file" );
        throw file_type_error( "no codec found" );
    }

    /*
     *    ok, we have a readable AVI file
     */
    readIndex();

    /*
     *    DV type 1 has audio decoded in the video stream
     */
    if ( findStreamWithType( FOURCC( "auds" ) ) != -1 ) {
        INFO( "Found AVI audio stream" );
        _hasAviAudio = true;
    } else {
        INFO( "AVI has no audio (or audio is interleaved in video, DV1)" );
    }

    /* doesn't really matter, ...  */
    if ( _container->getDirEntry( FOURCC("dmlh") ) != -1 ) {
        INFO( "AVI file is an OpenDML file, good." );
    }

//     INFO( PRINTV(_container->mainHdr.dwSuggestedBufferSize) );
//     INFO( PRINTV(_container->mainHdr.dwWidth) );

    /* FIXME : all the rest of the constructor is only working for DV,
     * but right now I * support DV only anyway
     */
    _dvDecoder = (AVDecoderIFace*) dvDecoderPlugin->newElement();

    /* read a frame and try to init the decoder. Allocate enough for
     * PAl or NTSC, just in case the file is not nicely written
     */
    Frame frame = getFrame( 0 );
    _dvDecoder->init( frame.getRawVBuf()->buf() );
    frame.dropRefs();

    /* set the correct properties */
    setVideoFmt( _dvDecoder->getVideoFmt() );
    setLength( _container->mainHdr.dwTotalFrames * getVideoFmt().getFrameLength() );

    return;
}

InAviStream::~InAviStream()
{
    PDELETE(_container);
    PDELETE(_dvDecoder);
}

Frame
InAviStream::getFrame( Time t )
{
    Frame f;

    // FIXME : alle the following code relies on constant frames per second
    //         this should be done via the AVI index.

//    if ( t>=length() ) t = length() - getVideoFmt().getFrameLength();
    if ( t<0 ) t = 0.0;

    int   offset_frames  = static_cast<int>( t.sec() * getVideoFmt().fps );
    if ( offset_frames<0 ) offset_frames=0;
    off_t offset_bytes = -1;
    int   size = -1;

    switch ( _container->index_type ) {
      case AVI_LARGE_INDEX: {
          /* find relevant index in indx0 */
          int i = 0;
          for ( i = 0; 
                offset_frames >= _container->indx[0]->aIndex[i].dwDuration; 
                ++i ) 
          {
              WARN( "something is screwed up, trying to reed beyond end of chunk" );
              offset_frames -= _container->indx[0]->aIndex[i].dwDuration;
          }

          if ( i != _container->current_ix00 ) {
//               INFO( "read frame with size "
//                     <<PRINTV(_container->indx[0]->aIndex[i].dwSize));
              int o = _container->indx[0]->aIndex[i].qwOffset + RIFF_HEADERSIZE;
              int s = _container->indx[0]->aIndex[i].dwSize - RIFF_HEADERSIZE;
              lseek( _fd, o, SEEK_SET);
              read( _fd, _container->ix[0], s );
//               INFO( "hhh"<<PRINTV(o)<<PRINTV(s));
//               INFO( PRINTV(offset_frames));
              _container->current_ix00 = i;
          }

          if ( offset_frames < _container->ix[0]->nEntriesInUse) {
              offset_bytes = _container->ix[0]->qwBaseOffset + 
                  _container->ix[0]->aIndex[offset_frames].dwOffset;
              size = _container->ix[0]->aIndex[offset_frames].dwSize;
          } else {
              WARN( "something is screwed up, trying to reed beyond end of chunk" );
              return f;
          }
          break;
      }
      case AVI_SMALL_INDEX: {
          // the following doen't seem to make much sense ... ????
          int index = -1;
          int frameNumIndex = 0;
          FOURCC chunkID1 = FOURCC("00dc");
          FOURCC chunkID2 = FOURCC("00db");
          for (int i = 0; i < _container->idx1->nEntriesInUse; ++i) {
              if (_container->idx1->aIndex[i].dwChunkId == chunkID1 ||
                  _container->idx1->aIndex[i].dwChunkId == chunkID2 ) {
                  if (frameNumIndex == offset_frames) {
                      index = i;
                      break;
                  }
                  ++frameNumIndex;
              }
          }
//          INFO(PRINTV(index)<<PRINTV(frameNumIndex)<<PRINTV(offset_frames));
          
          if (index != -1) {
              // compatibility check for broken dvgrab dv2 format
              // new, correct dv2 format
              offset_bytes = _container->idx1->aIndex[index].dwOffset + RIFF_HEADERSIZE + _container->dir[_container->movi_list].offset;
              size = _container->idx1->aIndex[index].dwSize;
          } else {
              WARN( "something is screwed up, trying to reed beyond end of chunk" );
              return f;
          }
          break;
      }
    }

//    INFO( PRINTV(offset_bytes)<<PRINTV(size) );

    ssize_t ret;
    ret = lseek( _fd, offset_bytes, SEEK_SET);
    if ( ret == -1 ) {
        WARN( "could not find frame "<<t<<" in file "<<getFileName() );
        return f;
    }

    Buffer * buffer = BufferPool::getNewBuffer( size );
    ret = read( _fd, buffer->buf(), size);
    if ( ret != size ) {
        WARN( "error reading "<<getFileName() );
        WARN( PRINTV( ret )<<PRINTV( size )<<PRINTV( t.sec() )<<PRINTV( offset_bytes ) );
        WARN( "FIXME when does this happen? what to do?" );
        return f;
    }


    f.setRawVBuf( buffer ); 
    f.setAVContext( this );
    f.setPos( t );
    f.setLength( 1.0 / _dvDecoder->getFPS() );
    _deliveredFrames++;

    return f;
}


void 
InAviStream::decodeVideo( Frame & f, uint8_t * dest )
{
    /* don't do it here, we have a decoder plugin */
    _dvDecoder->decodeVideo( f, dest );
}

void 
InAviStream::decodeAudio( Frame & f, uint8_t * dest )
{
    /* don't do it here, we have a decoder plugin */
    _dvDecoder->decodeAudio( f, dest );
}




PropertyNode * 
InAviStream::getProperties(PropertyNode * p) const
{
    if ( !_dvDecoder ) {
        WARN( "arghh!" );
        return 0;
    }

    // container format
    PropertyNode * stream = new PropertyNode( "stream" );
    stream->addAttr( "filename", getFileName() );
    stream->addAttr( "duration", length().sec() );

    PropertyNode * container = stream->addDaughter( "container" );
    container->addAttr( "format", "AVI" );

    // audio and video format
    _dvDecoder->getProperties( container );

    return stream;
}


void
InAviStream::parseHeader()
{
    RIFFChunk riffchunk;
    FOURCC RIFF( "RIFF" );
    off_t filesize;
    int   root_idx =
        _container->newDirEntry( FOURCC("FILE"), FOURCC("FILE"), 0, RIFF_NO_PARENT_IDX );

    off_t pos = lseek( _fd, 0, SEEK_SET );
    /* calculate file size from RIFF header instead from physical file. */

    while ( (read( _fd, &riffchunk, sizeof(RIFFChunk)) > 0) && 
            riffchunk.type == RIFF ) {

//        INFO( PRINTV(pos)<<PRINTV(riffchunk.type)<<PRINTV(riffchunk.type.fourcc)<<PRINTV(riffchunk.length));
        
        filesize += riffchunk.length + RIFF_HEADERSIZE;
//        INFO( "Found RIFF Chunk with "<<PRINTV(riffchunk.length) );
        // FIXME ; handle file exceptions
        lseek( _fd, pos, SEEK_SET );
        parseList( root_idx );
        pos = lseek( _fd, 0, SEEK_CUR);
        ABORT_IF( pos == -1, "Fuck" );
    }

    // FIXME : this code looks like it schould maybe go into AviContainer
    _container->avih_chunk =  _container->getDirEntry( FOURCC("avih") );
    if ( _container->avih_chunk != -1 ) {
        readChunk( _container->avih_chunk, (void*)&(_container->mainHdr) );
    } else {
        throw file_type_error( "not an AVI file" );
    }
}

void
InAviStream::parseList( int parent_idx )
{
//    INFO( "InAviStream::parseList" );

    /* Read in the chunk header (type and length). */
    FOURCC      type;
    int         length;

    read(_fd, &type, sizeof(type));
    read(_fd, &length, sizeof(length));
    if (length & 1)
        length++;

    /* The contents of the list starts here. Obtain its offset. The list
       name (4 bytes) is already part of the contents). */
    off_t pos = lseek( _fd, 0, SEEK_CUR);
    ABORT_IF( pos == -1, "Fuck" );
    FOURCC      name;
    ABORT_IF( read( _fd, &name, sizeof(name)) <0, "Fuck" );

//    INFO( PRINTV(pos)<<PRINTV(type)<<PRINTV(type.fourcc)<<PRINTV(name)<<PRINTV(length));

    /* if we encounter a movi list, do not read it. It takes too much time
       and we don't need it anyway. */
    if ( name != FOURCC("movi") ) {
//        INFO( "found !movi list" );

        /* Add an entry for this list. */
        int new_idx = 
            _container->newDirEntry( type, name, sizeof(name), parent_idx );

        /* Read in any chunks contained in this list. This list is the
           parent for all chunks it contains. */
        int listEnd = pos + length;
        while ( pos < listEnd ) {
            parseChunk( new_idx );
            pos = lseek( _fd, 0, SEEK_CUR);
            ABORT_IF(pos == (off_t)-1, "Fuck");
        }
    } else {
        /* Add an entry for this list. */
//        INFO( "found movi list" );
        _container->movi_list = 
            _container->newDirEntry( type, name, length, parent_idx );
        pos = lseek( _fd, length - 4, SEEK_CUR);
        ABORT_IF( pos == (off_t)-1, "Fuck" );
    }
}

void
InAviStream::parseChunk( int parent_idx )
{
//    INFO("InAviStream::parseChunk");
    FOURCC      type;
    int         length;

    /* Check whether it is a LIST. If so, let ParseList deal with it */
    read( _fd, &type, sizeof(type));
//    INFO( "parsing chunk "<<PRINTV(type.fourcc)<<PRINTV(type) );
    if (type == FOURCC("LIST")) {
        ABORT_IF(lseek( _fd, -sizeof(type), SEEK_CUR) == (off_t)-1, "Fuck" );
        parseList( parent_idx );
    } else {
        /* it is a normal chunk, create a new directory entry for it */
        ABORT_IF( read( _fd, &length, sizeof(length) ) <0, "Fuck" );
        if (length & 1) length++;
        _container->newDirEntry( type, FOURCC::NONE, length, parent_idx );
        ABORT_IF( lseek( _fd, length, SEEK_CUR) == (off_t)-1, "Fuck" );
    }

}

void
InAviStream::readChunk( int idx , void * data )
{
//    INFO("InAviStream::readChunk");

    const AviDirEntry& entry = _container->dir[ idx ];
    ABORT_IF(lseek(_fd, entry.offset, SEEK_SET) == (off_t)-1, "Fuck");
    ABORT_IF(read(_fd, data, entry.length )<0, "Fuck");

}


void
InAviStream::readIndex()
{
//    INFO( "InAviStream::readIndex" );

    /*
     *
     */
    _container->indx_chunk[0] = _container->getDirEntry( FOURCC("indx") );
    if ( _container->indx_chunk[0] != -1 ) {
        readChunk( _container->indx_chunk[0], (void*)_container->indx[0]);
        _container->index_type = AVI_LARGE_INDEX;
        INFO( "found AVI with large index" );
		
        /* recalc number of frames from each index */
        _container->mainHdr.dwTotalFrames = 0;
        for (int i = 0; 
             i < _container->indx[0]->nEntriesInUse; 
             i++ )
        {
            _container->mainHdr.dwTotalFrames 
                += _container->indx[0]->aIndex[i].dwDuration;
        }

    } else {
        _container->idx1_chunk = _container->getDirEntry( FOURCC("idx1") );
        if ( _container->idx1_chunk != -1) {
            readChunk( _container->idx1_chunk, (void*)_container->idx1);
            _container->idx1->nEntriesInUse = 
                _container->dir[_container->idx1_chunk].length / 16;
            _container->index_type = AVI_SMALL_INDEX;
            INFO( "found AVI with small index" );
            
            /* recalc number of frames from the simple index */
            int frameNumIndex = 0;
            FOURCC chunkID1 = FOURCC("00dc");
            FOURCC chunkID2 = FOURCC("00db");
            for (int i = 0; i < _container->idx1->nEntriesInUse; ++i) {
                if (_container->idx1->aIndex[i].dwChunkId == chunkID1 || 
                    _container->idx1->aIndex[i].dwChunkId == chunkID2 ) {
                    ++frameNumIndex;
                }
            }
            _container->mainHdr.dwTotalFrames = frameNumIndex;
        }
    }
//    INFO( PRINTV( _container->mainHdr.dwTotalFrames ));

    return;
}

int 
InAviStream::findStreamWithType( FOURCC type )
{
	AviStreamHeader	aviStreamHeader;
	FOURCC strh( "strh" );
	
    int i = -1, n = 0;
	while ( (i = _container->getDirEntry( strh, n++ )) != -1 )
	{
		readChunk( i, (void*)&aviStreamHeader );
		if ( aviStreamHeader.fccType == type ) {
			return i;
        }
	}
    return -1;
}

int
InAviStream::findStreamWithHandler( FOURCC format )
{
    ABORT("InAviStream::verifyStreamFormat" );
    
    /*
     *   
     */
	AviStreamHeader	aviStreamHeader;
	FOURCC strh("strh");
    int i = -1, n = 0;
	while ( (i = _container->getDirEntry( strh, n++ )) != -1 )
	{
		readChunk( i, (void*)&aviStreamHeader );
//        INFO( "found AVI chunck "<<PRINTV(aviStreamHeader.fccHandler));
		if (aviStreamHeader.fccHandler == format)
			return i;
	}

    /*
     *   
     */
	BITMAPINFOHEADER bitmapInfoHeader;
	FOURCC strf("strf");
	n = 0;
	while ( (i = _container->getDirEntry( strf, n++ )) != -1 )
	{
		readChunk( i, (void*)&bitmapInfoHeader );
//        INFO( "found AVI chunck biCompr: "<<FOURCC(bitmapInfoHeader.biCompression));
		if ( FOURCC(bitmapInfoHeader.biCompression) == format)
			return i;
	}

    return -1;
}



/*
 *  class OutAviStream 
 */
OutAviStream::OutAviStream( const string& fileName )
  : OutAVFStream( fileName ),
    _dvEncoder( 0 ),
    _frameSize( 0 )
{
    INFO( "opened avi dv file \""<<fileName<<"\" for writing." );

    /*  init the encoder  */
//    _dvEncoder = new LibDVEncoder();
//    _dvEncoder->init();
//    _frameSize = _dvEncoder->getFrameSize();
    
    return;
}

OutAviStream::~OutAviStream()
{
//	PDELETE( _dvEncoder );
}


void
OutAviStream::putFrame( Frame & f )
{

    if ( f.getVideoIsDecoded() ) {
        WARN( "FIXME use decoder to encode");
    } else if ( f.getRawVBuf() ) {
        WARN( "FIXME check format of RawVBuf and see if we match" );
        write( _fd, f.getRawVBuf(), 
               _frameSize );
    } else {
        WARN( "no video data" );
    }

    
    return;
}

PropertyNode * 
OutAviStream::getProperties(PropertyNode * p) const
{
    ABORT( "FIXME" );
    return 0;
}
