/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**
 * \file ossstream.cc
 */

// config
#if defined( HAVE_CONFIG_H )
#    include "config.h"
#endif


// std c
#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>

// std c++
#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <string>
using std::string;

// this header
#include "ossstream.hh"

// PIAVE
#include "libpiave/frame.hh"
#include "libpiave/property.hh"

using namespace PIAVE;

PIAVE_PLUGIN( OssPlugin );


/*
 *    wrapper plugin
 */
OssPlugin::OssPlugin() :
  Plugin( "OSSStream", PLUGIN_OUTSTREAM )
{
    PropertyNode * caps = new PropertyNode( "outstream" );
    caps->addAttr( "name", getName() );

    PropertyNode * monitor = caps->addDaughter( "monitor" );
    PropertyNode * audio = monitor->addDaughter( "audio" );
    audio->addAttr("type","oss");

    PropertyNode * about = caps->addDaughter( "about" );
    about->setContent( "Simple OSS layer for audio output.\n" );

    setCapabilities( caps );

}


PluginBase * 
OssPlugin::newElement( const PropertyNode * p )
{
    return new OutOSSStream();
}






/*
 *  \class OutOSSstream
 *
 *
 */
OutOSSStream::OutOSSStream() : 
  OutAudioStreamIFace(),
  _fd( -1 ),
  _deviceName( "/dev/dsp" )
{
}

OutOSSStream::~OutOSSStream()
{
    if ( _fd != -1 ) {
        close( _fd );
        _fd = -1;
    }
}


void
OutOSSStream::init( Frame & f ) throw(exception)
{
    if ( getState() == state_error ) return;

    int format = AFMT_S16_NE;
    int frequency_request = -1;
    int channels_request = -1;


    InAVStreamIFace * aContext = f.getAudioContext();
    if ( !aContext ) {
        aContext = f.getAVContext();
        if ( !aContext ) {
            INFO( "no audio/av context" );
            return;
        }
    }

    /* get audio properties */
    double frequency = -1.0;
    double channels = -1.0;
    PropertyNode * av_props = aContext->getProperties();
    PropertyNode * audio = 0;
    PropertyNode * p = 0;
    int d = 0;
    string type;
    while( ( p = av_props->getDaughter( "codec", d++, true ) ) && !audio ) {
        if ( p->getAttr( "type", type ) ) {
            if ( type == "audio" ) {
                audio = p;
            }
        }
    }
    if ( !audio ) {
        WARN( "Couldn't get properties of audio codec" );
        setState( state_error );
        return;
    }
    audio->getAttr( "frequency", frequency );
    audio->getAttr( "channels", channels );
    frequency_request = static_cast<int>( frequency );
    channels_request = static_cast<int>( channels );
    INFO( PRINTV( frequency )<<PRINTV( channels ) );
    PDELETE( av_props );


    /* sanity checks */
    if ( frequency_request < 0  ||
         channels_request < 0 )
    {
        WARN( "couldn't get all parameters"
              <<PRINTV(frequency_request)<<PRINTV(channels_request) );
        setState( state_error );
        return;
    }


    /* open OSS devoce */
    int frequency_reply = frequency_request;
    int channels_reply = channels_request;

    /* open audio device */
    _fd = open( _deviceName.c_str(), O_WRONLY | O_NONBLOCK | O_NDELAY );
    if ( _fd == -1 ) {
        WARN( "couldn't open "<<_deviceName<<" in non blocking mode. Trying blocking mode." );
        _fd = open( _deviceName.c_str(), O_WRONLY );
        if ( _fd == -1 ) {
            WARN( "couldn't open device: " << _deviceName );
            setState( state_error );
            return;
        }
    }
    
    int fragment = 0x0004000B;
    /* set sample format -- try for 16bit */
    if ( ioctl( _fd, SNDCTL_DSP_SETFRAGMENT, &fragment) == -1 ) {
        WARN( "error on SNDCTL_DSP_SETFRAGMENT" );
        setState( state_error );
        return;
    }        
    if ( ioctl( _fd, SOUND_PCM_WRITE_BITS, &format) == -1 ) {
        WARN( "error on SOUND_PCM_WRITE_BITS" );
        setState( state_error );
        return;
    }        
    if ( format != AFMT_S16_NE ) {
        WARN( "format not supported "<<PRINTV( format ) );
        setState( state_error );
        return;
    }        

    /* set stereo mode */
    if ( ioctl( _fd, SOUND_PCM_WRITE_CHANNELS, &channels_reply) == -1 ) {
        WARN( "error on SNDCTL_DSP_CHANNELS" );
        setState( state_error );
        return;
    }        
    if ( channels_reply != channels_request ) {
        WARN( "couldn't get all channels  "
              << " we request=" << channels_request
              << " -- we got="<< channels_reply );
        setState( state_error );
        return;
    }        

    /* set sampling rate */
    if ( ioctl( _fd, SOUND_PCM_WRITE_RATE, &frequency_reply) == -1 ) {
        WARN( "error on SOUND_PCM_WRITE_RATE" );
        setState( state_error );
        return;
    }        
    if ( frequency_reply != frequency_request ) {
        WARN( "couldn't get right sampling rate "
              << " we request=" << frequency_request
              << " -- we got=" << frequency_reply );
        setState( state_error );
        return;
    }

    setState( state_open );

    // FIXME set frequency and channels as properties of this stream
    _nChannels = channels_reply;
    _frequency = frequency_reply;

    return;
}


void 
OutOSSStream::putFrame( Frame & f )
{
    if ( getState() != state_open ) {
        if ( getState() != state_error ) {
            init( f );
            if ( getState() == state_error ) {
                return;
            }
        } else {
            return;
        }
    }

    AudioBuffer * a = f.makeABuf();
    if ( !a ) return;

    if ( a->getNChannels() != _nChannels ) {
        WARN( "audio input nChannels != my nchannels "
              <<PRINTV( a->getNChannels() ) << PRINTV( _nChannels ) );
        return;
    }

    /*   either the channels are in sepearete buffers, or interleaved
     *   in a single buffer. If they are in seperate buffers, we are
     *   going to interleave them into a single buffer
     */
    int16_t * buffer = (int16_t*)a->interleave()->buf();

    /* Send the audio to the device */
    int total = a->getNSamples() * _nChannels * sizeof( int16_t );
    int written = 0;
    while ( written < total) {
        int result = write( _fd, buffer + written, total - written);
        if ( result <= 0 ) {
            WARN( "writing to "<<_deviceName<<PRINTV( result ) );
            return;
        }
        written += result;
    }

    /* play it _now_ */
    if ( ioctl( _fd, SNDCTL_DSP_POST, NULL ) == -1 ) {
        WARN( "while playing sound" );
        return;
    }

    /* done with the audio */
    f.dropARefs();

    return;

}

PropertyNode * 
OutOSSStream::getProperties( PropertyNode * p ) const
{
    WARN( "FIXME" );
    return 0;
}
