/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * \file frame.cc implements frame.hh
 *
 * based on:
 *
 * frame.cc -- utilities for process digital video frames
 *
 */

// config
#if defined( HAVE_CONFIG_H )
#    include "config.h"
#endif

// C includes
#include <pthread.h>

// std
#include <string>
using std::string;
#include <iostream>
using std::cout;
using std::cerr;
using std::endl;

// this header
#include "libdv_codec.hh"

// PIAVE
#include "libpiave/piave_base.hh"
#include "libpiave/frame.hh"
#include "libpiave/property.hh"
#include "libpiave/bufferpool.hh"

using namespace PIAVE;

/*
 * 
 *
 */
const ssize_t         LibDV::PAL_FRAMESIZE;
const ssize_t         LibDV::NTSC_FRAMESIZE;
const double          LibDV::PAL_FPS;
const unsigned int    LibDV::PAL_WIDTH;
const unsigned int    LibDV::PAL_HEIGHT;
const double          LibDV::NTSC_FPS;
const unsigned int    LibDV::NTSC_WIDTH;
const unsigned int    LibDV::NTSC_HEIGHT;
const double          LibDV::ASPECT_4x3;
const double          LibDV::ASPECT_16x9;

uint8_t LibDV::tmp_buffer[ 720 * 576 * 3 ];


/*
 *    plugin wrapper for raw DV codec
 */
LibDVDecoderPlugin::LibDVDecoderPlugin() :
    Plugin( "libdv_decoder", PLUGIN_VIDEO_CODEC )
{
    PropertyNode * caps = new PropertyNode( "decoder" );
    caps->addAttr( "name", getName() );

    /* stolen from mplayer */
    PropertyNode * format;
    format = caps->addDaughter( "format" );
    format->addAttr( "fourcc", "DVSD" );
    format = caps->addDaughter( "format" );
    format->addAttr( "fourcc", "dvsd" );
    format = caps->addDaughter( "format" );
    format->addAttr( "fourcc", "DVCS" );
    format = caps->addDaughter( "format" );
    format->addAttr( "fourcc", "dvcs" );
    format = caps->addDaughter( "format" );
    format->addAttr( "fourcc", "dvc" );

    PropertyNode * p = 0;
    PropertyNode * par = caps->addDaughter( "parameter" );
    par->addAttr( "name", "quality" );
    par->addAttr( "type", "list" );
    p = par->addDaughter( "listitem" );
    p->addAttr("name","DC - very fast");
    p->addAttr("value","1");
    p = par->addDaughter( "listitem" );
    p->addAttr("name","AC1 - medium");
    p->addAttr("value","2");
    p = par->addDaughter( "listitem" );
    p->addAttr("name","AC2 - best");
    p->addAttr("value","3");
    

    PropertyNode * about = caps->addDaughter( "about" );
    about->setContent(_("This codec uses libdv to decode DV video.\n"));
    
    setCapabilities( caps );
}

PluginBase * 
LibDVDecoderPlugin::newElement( const PropertyNode * p )
{
    // FIXME maybe implement quality as an option ....
    return new LibDVDecoder();
}

LibDVEncoderPlugin::LibDVEncoderPlugin() :
    Plugin( "libdv_encoder", PLUGIN_VIDEO_CODEC )
{
    PropertyNode * caps = new PropertyNode( "encoder" );
    caps->addAttr( "name", getName() );

    // FIXME  create capabilities
    // describe PAL/NTSC for video
    // describe audio options
//    WARN( "FIXME" );

    PropertyNode * about = caps->addDaughter( "about" );
    about->setContent(_("This codec uses libdv to encode DV video.\n"));
    
    setCapabilities( caps );
   
}



PluginBase * 
LibDVEncoderPlugin::newElement( const PropertyNode * p )
{
    // FIXME maybe implement quality as an option ....
    return new LibDVEncoder();
}

/*
 *    \class LibDVDecoder
 *
 *
 */
LibDVDecoder::LibDVDecoder() :
  _dv_decoder( 0 ),
  _dv_video_system( e_dv_system_none ),
  _frameSize( 0 )
{
    memset( _pixels, 0, sizeof( _pixels ) );
    memset( _pitches, 0, sizeof( _pitches ) );

}


void 
LibDVDecoder::decodeMetaData( Frame & f, PropertyNode * p )
{
    if ( p == 0 ) return;
    if ( !f.getRawVBuf()->buf() ) return;
    if ( !isInitialized() ) {
        init( f.getRawVBuf()->buf() );
    }

    int err;

    /* try to parse DV headers  */
    err = dv_parse_header( _dv_decoder, f.getRawVBuf()->buf() );
    if ( err<0 ) {
        WARN( "couldn't parse header" );
        return;
    }
	dv_parse_packs( _dv_decoder, f.getRawVBuf()->buf() );


    /*
     *  now get properties  
     */

    /* for this works in most cases, but not always.  when a piece of
     * tape has been overwritten, the beginning of the old (partly
     * overwritten) recording after th piece which has been
     * overwritten is not detected 
     */
    bool isNewRec = _dv_decoder->audio->aaux_asc.pc2.rec_st == 0;
    p->addAttr( "isNewRecording", isNewRec?"t":"f" );

    /* My Sony DCR-TRV30E returns for frame_changed:
     *     1   = normal video; a new frame
     *     0   = for all frames in a still shot except the first frame
     *    -1  = when playing beyond recorded tape
     */
    int frameChanged = dv_frame_changed( _dv_decoder );
    p->addAttr( "frameChanged", frameChanged );

    /*  position on tape */
    char time_str[20];
    err = dv_get_timestamp( _dv_decoder, time_str );
    if ( err>0 ) {
        p->addAttr( "timestamp", time_str );
    }

    /* date of the recording (in case the camera's clock was set
     * correctly ;-) 
     */
    err = dv_get_recording_datetime( _dv_decoder, time_str );
    if ( err>0 ) {
        p->addAttr( "recordingDatetime", time_str );
    }
}

/*
 * retrieves the audio data from the frame
 *  
 * The DV frame contains audio data mixed in the video data blocks,
 * which can be retrieved easily using this function.
 *
 * The audio data consists of 16 bit, two channel audio samples (a 16
 * bit word for channel 1, followed by a 16 bit word for channel 2
 * etc.)
 *
 * \return the number of bytes put into the buffer, or 0 if no audio
 * data could be retrieved
 */
void
LibDVDecoder::decodeAudio( Frame & f, uint8_t * dest )
{
    /*  already decoded  */
    if ( f.getAudioIsDecoded() ) return;
    /*  in DV format, we use the audio interleaved in the raw video data */
    if ( !f.getRawVBuf() ) return;
    if ( !isInitialized() ) {
        init( f.getRawVBuf()->buf() );
    }

    Buffer  *buffers[4];
    int16_t *audio[4];

    AudioBuffer * aBuf = f.getABuf();
    if ( dest == 0 ) {
        
        if ( !aBuf ) {
            /*  make a abuf if nobody else did already  */
            aBuf = new AudioBuffer();
            f.setABuf( aBuf );
        } else {
            aBuf->dropRef();
        }
        
        for ( int ch = 0; ch<4; ch++ ) {
//             buffers[ch] = BufferPool::getNewBuffer( 
//                 sizeof(int16_t) * DV_AUDIO_MAX_SAMPLES );
            buffers[ch] = BufferPool::getNewBuffer( 4000 );
            audio[ch] = (int16_t*)buffers[ch]->buf();
        }

    } else {
        WARN( "FIXME" );
        return;
    }

    /* call libdv  */
    dv_decode_full_audio( _dv_decoder, 
                          f.getRawVBuf()->buf(), 
                          audio );

    f.setAudioIsDecoded( true );

    f.getABuf()->getChannels()[0] = buffers[0];
    f.getABuf()->getChannels()[1] = buffers[1];
    f.getABuf()->getChannels()[2] = buffers[2];
    f.getABuf()->getChannels()[3] = buffers[3];

    if ( aBuf ) {
        aBuf->setNChannels( _dv_decoder->audio->num_channels );
        aBuf->setNSamples( _dv_decoder->audio->samples_this_frame );
        aBuf->setInterleaved( false );
    }

    return;
}

/*
 *
 *
 */
void
LibDVDecoder::decodeVideo( Frame & f, uint8_t * dest )
{
    /*  sanity checks  */
    if ( f.getVideoIsDecoded() ) return;
    if ( !f.getRawVBuf() ) return;
    if ( !isInitialized() ) {
        init( f.getRawVBuf()->buf() );
    }

    /* the VideoBuffer always has the size render_size or
     * display_size. If that is equal to size of the footage, we don't
     * need to scale the picture.
     */
    VideoBuffer * vBuf = 0;
    bool scale = false;
    if ( dest == 0 ) {
        /*  make a vbuf if nobody else did already  */
        vBuf = f.getVBuf();
        if ( !vBuf ) {
            vBuf = new VideoBuffer();
            f.setVBuf( vBuf );
        }
        scale = ( vBuf->getSizePixels() != getSizePixels() );
        if ( scale ) {
	    INFO( PRINTV(vBuf->getSizePixels())<<PRINTV(getSizePixels()));
            _pixels[0] = LibDV::tmp_buffer;
        } else {
            _pixels[0] = vBuf->getPixels();
        }
    } else {
        _pixels[0] = dest;
    }

    /* decode the frame  */
    _pitches[0] = getWidth() * 2;
    dv_decode_full_frame( _dv_decoder, f.getRawVBuf()->buf(),
                          e_dv_color_yuv, _pixels, _pitches );
    f.setVideoIsDecoded( true );

    /* scale to requested size  */
    if ( scale ) {
        WARN( "FIXME - nmot yet and not here. -> scaler filter." );
    }

    return;
};


bool 
LibDVDecoder::init( uint8_t * data ) throw (exception)
{
    INFO( "init decoder" );
    /*
     *   parse header and get file properties
     */
    if ( !data ) {
        return false; 
    }

    _dv_decoder = dv_decoder_new( TRUE, FALSE, FALSE );

    /// \todo test different speeds and make qual configurable
    _dv_decoder->quality = _dv_decoder->video->quality 
//        = DV_QUALITY_AC_2 | DV_QUALITY_COLOR;
        = DV_QUALITY_AC_1 | DV_QUALITY_COLOR;


    /* try to parse DV headers  */
    if( dv_parse_header( _dv_decoder, data )< 0) {
        INFO( "failed to parse DV header ... no DV file?" );
        throw codec_error( _("not a DV file") );
    }

    /*   check DV system  */
    unsigned char dsf = data[3] & 0x80;
    bool pal = ((data[3] & 0x80) == 0) ? false : true;
    if ( !pal ) {
        /*   really not PAL ???   */
        pal = ( dv_system_50_fields ( _dv_decoder ) == 1) ? true : false;
    }
    if ( pal ) {
        _dv_video_system = e_dv_system_625_50;
    } else {
        _dv_video_system = e_dv_system_525_60;
    }

    /*  chack video format  */
    bool wide = false;
    if ( dv_format_normal( _dv_decoder ) ) {
        wide = false;
    } else if ( dv_format_wide( _dv_decoder ) ) {
        wide = true;
    } else {
        WARN( "Arghhh  unknown format.. not 4x9, not 16:9 ...");
        throw codec_error( _("not a DV file") );
    }

    /*  setup parameters according to system and format  */
    if ( pal ) {
        /* PAL */
        _frameSize =  LibDV::PAL_FRAMESIZE;
        setFPS( LibDV::PAL_FPS ); 
        setWidth( LibDV::PAL_WIDTH );
        setHeight( LibDV::PAL_HEIGHT );
//        INFO( "System: PAL" );
    } else {
        /* NTSC */
        _frameSize =  LibDV::NTSC_FRAMESIZE;
        setFPS( LibDV::NTSC_FPS );
        setWidth( LibDV::NTSC_WIDTH );
        setHeight( LibDV::NTSC_HEIGHT );
//        INFO( "System: NTSC" );
    }

    if ( wide ) {
        setAspect( LibDV::ASPECT_16x9 );
//        INFO( "Format: 16x9" );
    } else {
        setAspect( LibDV::ASPECT_4x3 );
//        INFO( "Format: 4x3" );
    }



    setInitialized( true );
    INFO( "initialize decoder ok" );
    return true;
}

/*
 *
 *
 */
PropertyNode *
LibDVDecoder::getProperties( PropertyNode * p ) const
{
    if ( !p ) {
        p = new PropertyNode( "avcodec" );
    }
    PropertyNode * video = p->addDaughter( "codec" );
    video->addAttr( "type", "video" );
    video->addAttr( "name", "libdv_decoder" );
    video->addAttr( "format", "DV" );
    video->addAttr( "system", ((_dv_video_system==e_dv_system_625_50)?"PAL":"NTSC") );
    video->addAttr( "fps" , getFPS() );
    video->addAttr( "width" , getWidth() );
    video->addAttr( "height" , getHeight() );
    video->addAttr( "aspect" , getAspect() );

    // audio format
    PropertyNode * audio = p->addDaughter( "codec" );
    audio->addAttr( "type", "audio" );
    audio->addAttr( "name", "libdv_decoder" );
    audio->addAttr( "format", "PCL" );
    audio->addAttr( "frequency" , _dv_decoder->audio->frequency );
    audio->addAttr( "channels"  , _dv_decoder->audio->num_channels );
    audio->addAttr( "quantization" , _dv_decoder->audio->quantization );
    // FIXME how does the above fit to the following??????
    audio->addAttr( "bitspersample" , 16 );
    audio->addAttr( "emphasis" , (_dv_decoder->audio->emphasis ? "on" : "off") );
    audio->addAttr( "sampling" , _dv_decoder->sampling );

    return p;
}


/** checks whether this frame is the first in a new recording
 * To determine this, the function looks at the recStartPoint bit in
 * AAUX pack 51.
 *
 * \return true if this frame is the start of a new recording 
 */
bool 
LibDVDecoder::isNewRecording( Frame * f ) const
{
    dv_parse_header( _dv_decoder, f->getRawVBuf()->buf() );
    WARN( "FIXME   make sure we have data and have decoded headres" );
    return ( _dv_decoder->audio->aaux_asc.pc2.rec_st == 0 );
}







/*
 *
 *   DVEncoder
 *
 *
 *
 */
LibDVEncoder::LibDVEncoder() :
  AVEncoderIFace(),
  _dv_encoder( 0 ),
  _dv_video_system( e_dv_system_none ),
  _frameSize( 0 )
{
    memset( _pixels, 0, sizeof( _pixels ) );
    memset( _pitches, 0, sizeof( _pitches ) );
}


bool
LibDVEncoder::init( uint8_t * data ) throw (exception)
{
    /// \todo : make this configurable
    INFO( "Setup dv_encoder" );
    _dv_encoder = dv_encoder_new( FALSE, FALSE, FALSE );
    _dv_encoder->vlc_encode_passes = 3;
    _dv_encoder->static_qno = 0;
    _dv_encoder->force_dct = DV_DCT_AUTO;
    _dv_encoder->is16x9 = FALSE;
    _dv_encoder->clamp_luma = FALSE;
    _dv_encoder->clamp_chroma = FALSE;
    _dv_encoder->isPAL = 1;
    _frameSize = LibDV::PAL_FRAMESIZE;
    return true;
}



void 
LibDVEncoder::encodeVideo( Frame & f, uint8_t * dest )
{
    ABORT("FIXME");
#if 0
    if ( !_vBuf ) return;

    if ( _vBuf->hasAlpha() ) {
        WARN( "not supported" );
        flattenAlpha();
    }

    if ( !_dvFrame ) {
        DVFrame * dvf = FrameBufMgr::I()->getNewDVFrame(); 
        setDVFrame( dvf ); 
    }

    uint8_t   * pixels[3];
    pixels[0] = _vBuf->getPixBufPixels();
    pixels[1] = NULL;
    pixels[2] = NULL;

    dv_encode_full_frame( dv_encoder, pixels, e_dv_color_yuv,
                          _dvFrame->getBuffer() );
#endif
}

void 
LibDVEncoder::encodeAudio( Frame & f, uint8_t * dest )
{
    ABORT("FIXME");
}

PropertyNode *
LibDVEncoder::getProperties( PropertyNode * p ) const
{
    ABORT( " FIXME " );
    return 0;
}
