/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**
 * \file rawdvstream.cc implements rawdvstream.hh
 */

// config
#if defined( HAVE_CONFIG_H )
#    include "config.h"
#endif

// stdc
#include <errno.h>

// std++
#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <string>
using std::string;

// PIAVE
#include "libpiave/piave_base.hh"
#include "libpiave/frame.hh"
#include "libpiave/property.hh"
#include "libpiave/bufferpool.hh"


// this header
#include "rawdvstream.hh"
#include "libdv_codec.hh"

using namespace PIAVE;

/*
 *   Plugin for instream
 */
InRawDVStreamPlugin::InRawDVStreamPlugin() :
  Plugin( "InRawDVStream", PLUGIN_INSTREAM )
{
    PropertyNode * caps = new PropertyNode( "instream" );
    caps->addAttr( "name", getName() );

    PropertyNode * file = caps->addDaughter( "file" );

    PropertyNode * container = file->addDaughter( "container" );
    container->addAttr( "format", "rawdv" );
    container->addAttr( "extension", ".dv" );

    PropertyNode * video = container->addDaughter( "codec" );
    video->addAttr( "type","video" );
    video->addDaughter( "libdv_encoder" );

    PropertyNode * audio = container->addDaughter( "codec" );
    audio->addAttr( "type","audio" );

    PropertyNode * about = caps->addDaughter( "about" );
    about->setContent( "Can read raw DV files.\n" );

    setCapabilities( caps );
}

PluginBase * 
InRawDVStreamPlugin::newElement( const PropertyNode * p )
{
    if ( p ) {
        string filename;
        bool ok = p->getAttr( "filename", filename );
        if ( ok ) {
            return new InRawDVStream( filename );;
        }
    }

    WARN( "missing filename" );
    return 0;
}

/*
 *   Plugin for outstream
 */
OutRawDVStreamPlugin::OutRawDVStreamPlugin() :
  Plugin( "OutRawDVStream", PLUGIN_OUTSTREAM ) 
{
    PropertyNode * caps = new PropertyNode( "outstream" );
    caps->addAttr( "name", getName() );

    PropertyNode * file = caps->addDaughter( "file" );

    PropertyNode * container = file->addDaughter( "container" );
    container->addAttr( "format", "rawdv" );
    container->addAttr( "extension", ".dv" );

    PropertyNode * video = container->addDaughter( "codec" );
    video->addAttr( "type","video" );
    video->addDaughter( "libdv_encoder" );

    PropertyNode * audio = container->addDaughter( "codec" );
    audio->addAttr( "type","audio" );
    audio->addDaughter( "libdv_encoder" );

    PropertyNode * about = caps->addDaughter( "about" );
    about->setContent( "Can write raw DV files.\n" );

    setCapabilities( caps );
}

PluginBase * 
OutRawDVStreamPlugin::newElement( const PropertyNode * p )
{
    if ( !p ) {
        WARN( "missing filename" );
        return 0;
    }
    string filename;
    p->getAttr( "filename", filename );
    return new OutRawDVStream( filename );
}


/*
 *
 *  class InRawDVStream 
 *
 */
InRawDVStream::InRawDVStream( const string& fileName ) throw(exception)
  : InAVFStream( fileName ),
    _dvDecoder( 0 ),
    _frameSize( 0 )
{
    /*
     *    read the first frame to analyze it
     */
    static uint8_t buffer[ LibDV::PAL_FRAMESIZE ]; // enough for PAl or NTSC
    ssize_t ret;
    ret = lseek( _fd, 0, SEEK_SET );
    if ( ret == -1 ) {
        WARN( "could not read first frame in file "<<getFileName() );
        throw file_error( _("read error") );
    }

    ret = read( _fd, buffer, LibDV::PAL_FRAMESIZE );
    if ( ret != LibDV::PAL_FRAMESIZE ) {
        WARN( "could not read first frame in file "<<getFileName() );
        throw file_error( "read error" );
    }


    /*  init the decoder  */
    _dvDecoder = new LibDVDecoder();
    _dvDecoder->init( buffer );
    _frameSize = _dvDecoder->getFrameSize();

    setVideoFmt( _dvDecoder->getVideoFmt() );

    setLength( getFileSize() / _frameSize / _dvDecoder->getVideoFmt().fps );

    return;
}


Frame
InRawDVStream::getFrame( Time t )
{
    Frame f;

    if ( t>=length() ) t = length() - getVideoFmt().getFrameLength();
    if ( t<0 ) t = 0.0;

    int   offset_frames  = static_cast<int>( t.sec() * getVideoFmt().fps );
    off_t offset_bytes   = offset_frames * _frameSize;

    ssize_t ret;
    ret = lseek( _fd, offset_bytes, SEEK_SET );
    if ( ret == -1 ) {
        WARN( "could not find frame "<<t<<" in file "<<getFileName() );
        return f;
    }

    Buffer * buffer = BufferPool::getNewBuffer( _frameSize );
    ret = read( _fd, buffer->buf(), _frameSize );
    if ( ret != _frameSize ) {
        WARN( "error reading "<<getFileName() );
        WARN( PRINTV( ret )<<PRINTV( _frameSize )<<PRINTV( t.sec() )<<PRINTV( offset_bytes ) );
        WARN( "FIXME when does this happen? what to do?" );
        return f;
    }

    f.setRawVBuf( buffer ); 
    f.setAVContext( this );
    f.setPos( t );
    f.setLength( 1.0 / _dvDecoder->getFPS() );

    return f;
}

/* in a raw DV stream, each frame is completely independent of the
 * whole stream or other frames, hence, decoding of data and metadata
 * is independent of the stream, i.e. is handled by the Decoder.
 * 
 * in other streams these functions might not be simple call throughs
 * to the decoder, but actually deliver information that was read from
 * e.g. the file header
 */
void 
InRawDVStream::decodeMetaData( Frame & f )
{
    _dvDecoder->decodeMetaData( f, _frameProperties );
    f.setProperties( _frameProperties );
}

void 
InRawDVStream::decodeVideo( Frame & f, uint8_t * dest )
{
    /* don't do it here, we have a decoder plugin */
    _dvDecoder->decodeVideo( f, dest );
}

void 
InRawDVStream::decodeAudio( Frame & f, uint8_t * dest )
{
    /* don't do it here, we have a decoder plugin */
    _dvDecoder->decodeAudio( f, dest );
}


PropertyNode * 
InRawDVStream::getProperties( PropertyNode * p ) const
{
    if ( !_dvDecoder ) {
        WARN( "arghh!" );
        return 0;
    }

    // container format
    PropertyNode * stream = new PropertyNode( "stream" );
    stream->addAttr( "filename", getFileName() );
    stream->addAttr( "duration", length().sec() );

    PropertyNode * container = stream->addDaughter( "container" );
    container->addAttr( "format", "rawdv" );

    // audio and video format
    _dvDecoder->getProperties( container );

    return stream;
}


/*
 *  class OutRawDVStream 
 */
OutRawDVStream::OutRawDVStream( const string& fileName )
  : OutAVFStream( fileName ),
    _dvEncoder( 0 ),
    _frameSize( 0 )
{
    INFO( "opened raw dv file \""<<fileName<<"\" for writing." );

    /*  init the encoder  */
    _dvEncoder = new LibDVEncoder();
    _dvEncoder->init();
    _frameSize = _dvEncoder->getFrameSize();
    
    return;
}


void
OutRawDVStream::putFrame( Frame & f )
{
    if ( f.getVideoIsDecoded() ) {
        WARN( "FIXME use codec to encode");
    } else if ( f.getRawVBuf() ) {
//  FIXME         WARN( "FIXME check format of RawVBuf and see if we match" );
        int ret = write( _fd, f.getRawVBuf()->buf(),  _frameSize );
        int errnum = errno;
        if ( ret != _frameSize || errnum != 0 ) {
            WARN( PRINTV( ret )<<PRINTV( errnum ) );
        }
        _recievedFrames++;
    } else {
        WARN( "no video data" );
    }

    /* the frame is done, release A and V refereces */
    f.dropRefs();

    return;
}

PropertyNode * 
OutRawDVStream::getProperties( PropertyNode * p ) const
{
    ABORT( "FIXME" );
    return 0;
}
