/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**
 * \file sdloutstream.cc implements sdloutstream.hh
 */

// config
#if defined( HAVE_CONFIG_H )
#    include "config.h"
#endif


// this header
#include "sdlstream.hh"

// std++
#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <string>
using std::string;

// X
#include "X11/X.h"

// PIAVE
#include "libpiave/frame.hh"
#include "libpiave/property.hh"
#include <libdv/dv.h>
#include <libdv/dv_types.h>

using namespace PIAVE;


/*
 *   SDL Plugin
 *
 */

PIAVE_PLUGIN( SDLStreamPlugin );

SDLStreamPlugin::SDLStreamPlugin() :
  Plugin( "SDLStream", PLUGIN_OUTSTREAM ) 
{
    PropertyNode * caps = new PropertyNode( "outstream" );
    caps->addAttr( "name", getName() );

    PropertyNode * monitor = caps->addDaughter( "monitor" );
    PropertyNode * video = monitor->addDaughter( "video" );
    video->addAttr( "type", "xv" );
    video->addAttr( "fallback", "rgb" );

    video = monitor->addDaughter( "video" );
    video->addAttr( "type", "rgb" );

    PropertyNode * about = caps->addDaughter( "about" );
    about->setContent( "Use SDL to display video. Uses xv overlay if available.\n" );

    setCapabilities( caps );
}

PluginBase * 
SDLStreamPlugin::newElement( const PropertyNode * p )
{
    string type = "xv";
    double width = -1;
    double height = -1;
    if ( p ) {
        p->getAttrRec( "type", type );
        p->getAttrRec( "width", width );
        p->getAttrRec( "height", height );
    }
    
    if ( width<3 || height<3 ) {
        WARN( "insane value: "<<PRINTV( width )<<PRINTV(height) );
        return 0;
    }

    
    
    INFO( "trying to open SDL stream "<<PRINTV(width)<<PRINTV(height)<<PRINTV(type) );
    OutSDLStream * sdlStream = new OutSDLStream();
    sdlStream->_rect.x = 0;
    sdlStream->_rect.y = 0;
    sdlStream->_rect.w = (int)width;
    sdlStream->_rect.h = (int)height;

    if ( type == "rgb" ) {
        sdlStream->_fmt = OutSDLStream::rgb_surface;
    } else {
        sdlStream->_fmt = OutSDLStream::yuv_overlay;
    }

    sdlStream->init();

    return sdlStream;
}

/*
 *  \class OutSDLStream
 */

OutSDLStream::OutSDLStream() :
  OutAVStreamIFace(),
  _SDL_init_OK( false ),
  _screen( 0 ),
  _image( 0 ),
  _videoInfo( 0 ),
  _overlay( 0 ),
  _fmt( no_fmt )
{
}

OutSDLStream::~OutSDLStream()
{
    if(_overlay) {
        SDL_FreeYUVOverlay(_overlay);
        _overlay = 0;	
    }
}

void
OutSDLStream::init()
{
    INFO( "Initializing SDL" );
    _SDL_init_OK = (SDL_Init(SDL_INIT_VIDEO)==0);
 
    if ( !_SDL_init_OK ) {
        WARN( "Could not initialize SDL: "<<SDL_GetError() );
        return;
    }

    /* Clean up on exit */
    atexit(SDL_Quit);
 
    INFO( "setting SDL video mode" );
    _videoInfo = SDL_GetVideoInfo();
    int bpp;
    switch ( _videoInfo->vfmt->BitsPerPixel ) {
      case 8:
      case 16:
      case 24:
      case 32:
          bpp = _videoInfo->vfmt->BitsPerPixel;
          break;
      default:
          bpp = 32;
          break;
    }
    INFO( "try to get videomode: "<<PRINTV(bpp)<<PRINTV(_rect.w)<<PRINTV(_rect.h));

    _screen = SDL_SetVideoMode(  _rect.w, _rect.h, 
                                 bpp, SDL_HWSURFACE | SDL_DOUBLEBUF );
    if ( 0==_screen ) {
        WARN( "Couldn't set video mode:"<<SDL_GetError() );
        _SDL_init_OK = false;
        return;
    }

    SDL_WM_SetCaption( "PIAVE", "PIAVE" );
    
    if ( _fmt == rgb_surface ) {
        INFO( "Requested SDL RGB surface "<<PRINTV( _SDL_init_OK ) );
    } else {
        INFO( "Requested SDL YUV overlay "<<PRINTV( _SDL_init_OK ) );

        _overlay = SDL_CreateYUVOverlay( Global::renderFmt.width, Global::renderFmt.height,
                                         SDL_YUY2_OVERLAY, _screen );
        if( _overlay ) {
            INFO( "Got SDL overlay" );
            if( _overlay->hw_overlay ) {
                INFO( "Got SDL hardware overlay" );
                if ( SDL_LockYUVOverlay( _overlay ) >= 0 ) {
                    INFO( "lock overlay ok "<<PRINTV(_rect.w)<<PRINTV(_rect.h)  );

                } else {
                    INFO( "Couldn't lock YUVOverlay" );
                    SDL_FreeYUVOverlay( _overlay );
                    _overlay = 0;
                }

            } else {
                WARN( "No hardware overlay" );
            }
        } else {
            WARN( "Couldn't get overlay" );
        }
        if ( _overlay ) {
            INFO( "Using SDL YUV overlay "<<PRINTV( _SDL_init_OK ) );
        } else {
            WARN( "fallback: to RGB surface" );
        }

    }
    if ( ! _overlay ) {
        /*
         *       we are using RGB
         */
        Uint32 rmask, gmask, bmask, amask;
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
        rmask = 0xff000000;
        gmask = 0x00ff0000;
        bmask = 0x0000ff00;
        amask = 0x000000ff;
#else
        rmask = 0x000000ff;
        gmask = 0x0000ff00;
        bmask = 0x00ff0000;
        amask = 0xff000000;
#endif
        // FIXME ... think about how to handle alpha ...
        amask = 0x0;

        _image = SDL_CreateRGBSurface( SDL_HWSURFACE | SDL_DOUBLEBUF , 
                                       Global::renderFmt.width, Global::renderFmt.height,
                                       24, rmask, gmask, bmask, amask);

    }

    getSysWMinfo();

    setIgnoreInputEvents();

    return;
}

void
OutSDLStream::putFrame( Frame & f )
{
    ABORT_IF( SDL_QuitRequested() , "SDL_QuitReqiested" );
    if ( ! _SDL_init_OK ) return;

    VideoBuffer * vb = 0;

    if ( _overlay ) {
        if ( f.getVideoIsDecoded() ) {
            // FIXME  is this copy really necessary ???
            memcpy( _overlay->pixels[0], 
                    f.getVBuf()->getPixels(), 
                    f.getVBuf()->getSizeBytes() );
            vb = f.getVBuf();
        } else {
            if ( f.getAVContext() ) {
                f.getAVContext()->decodeVideo( f, _overlay->pixels[0] );
            } else if ( f.getRawVBuf() ) {
                WARN( " frame is not decoded and has no AVcontext ... arghh" );
                return ;
            } else {
                /* this frame looks very empty  */
                INFO( "ignoring empty frame" );
                return;
            }
        }

        SDL_UnlockYUVOverlay( _overlay );
        SDL_DisplayYUVOverlay( _overlay, &_rect );
        SDL_LockYUVOverlay( _overlay );

    } else {
        ABORT("FIXME implement YUV->RGB");
        
        if( SDL_BlitSurface( _image, 0, _screen, 0) < 0 ) {
            WARN( "BlitSurface error: "<<SDL_GetError() );
            return;
        }
        SDL_Flip( _screen );
     
    }

    /* the frame has been displayed release references to the
     * buffers 
     */
    f.dropVRefs();

    return;
}

const SDL_SysWMinfo &
OutSDLStream::getSysWMinfo()
{
    _sysWMinfo.version.major = SDL_MAJOR_VERSION;
    _sysWMinfo.version.minor = SDL_MINOR_VERSION;
    _sysWMinfo.version.patch = SDL_PATCHLEVEL;
    _sysWMinfo.subsystem = SDL_SYSWM_X11;
    _sysWMinfo.info.x11.wmwindow = 0xdeadbeef;
    SDL_GetWMInfo( &_sysWMinfo );
    return _sysWMinfo;
}


PropertyNode *
OutSDLStream::getProperties( PropertyNode * p ) const
{
	
    if ((_sysWMinfo.info.x11.lock_func != 0) && 
        (_sysWMinfo.info.x11.display != 0) &&
        (_sysWMinfo.info.x11.wmwindow != 0)) 
    {
        _sysWMinfo.info.x11.lock_func();
        XUnmapWindow(_sysWMinfo.info.x11.display, _sysWMinfo.info.x11.wmwindow);
        _sysWMinfo.info.x11.unlock_func();
    } else {
        WARN( "x11 info not setup correctly - could not hide window" );
    }

    if ( !p ) {
        p = new PropertyNode( "outstream" );
    }
    p->addAttr( "WinID", (int)_sysWMinfo.info.x11.wmwindow );

    if(SDL_GetVideoSurface()->flags & SDL_RESIZABLE) {
        p->addAttr( "resizeable", "yes" );
    } else {
        p->addAttr( "resizeable", "no" );
    }


    return p;
}


void
OutSDLStream::setIgnoreInputEvents()
{
    // Turn off keboard/mouse events & make them propagate to the parent window.

    XSelectInput( _sysWMinfo.info.x11.display, _sysWMinfo.info.x11.window, ExposureMask);
}
