/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**
 * \file sndfile.cc
 */

// config
#if defined( HAVE_CONFIG_H )
#    include "config.h"
#endif


// std c
#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>

#include <math.h>
// std c++
#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <string>
using std::string;

// sndfile API
#include "libsndfile/sndfile.h"

// this header
#include "sndfile.hh"

// PIAVE
#include "libpiave/frame.hh"
#include "libpiave/property.hh"

using namespace PIAVE;

PIAVE_PLUGIN( SndfilePlugin );


/*
 *    wrapper plugin
 */
SndfilePlugin::SndfilePlugin() :
  Plugin( "sndfile", PLUGIN_INSTREAM )
{
    PropertyNode * caps = new PropertyNode( "instream" );
    caps->addAttr( "name", getName() );

    PropertyNode * file = caps->addDaughter( "file" );

    PropertyNode * container = file->addDaughter( "container" );
    container->addAttr( "format", "wav" );
    container->addAttr( "extension", ".wav" );

    container = file->addDaughter( "container" );
    container->addAttr( "format", "au" );
    container->addAttr( "extension", ".au" );

    container = file->addDaughter( "container" );
    container->addAttr( "format", "aiff" );
    container->addAttr( "extension", ".aiff" );

    PropertyNode * about = caps->addDaughter( "about" );
    about->setContent( _("Read wav/au/aiff/etc. audio files with libsndfile.") );

    setCapabilities( caps );
}


PluginBase * 
SndfilePlugin::newElement( const PropertyNode * p )
{
    if ( p ) {
        string filename;
        bool ok = p->getAttr( "filename", filename );
        if ( ok ) {
            return new InSndfile( filename );;
        }
    }

    WARN( "missing filename" );
    return 0;
}






/**
 * \class InSndfile
 * use libsndfile to read and decode sound files.
 *
 */
InSndfile::InSndfile( const std::string& fileName ) throw(exception)
  : InAVFStream( fileName ),
    _sndfile( 0 ),
    _nChannels( 0 ),
    _frequency( 0 ),
    _minSec( 1 ),
    _maxSec( 3 )
{

    _sndfile = sf_open ( getFileName().c_str(), SFM_READ, &_sfinfo);
    if ( _sndfile == 0 )
    {
        string err( sf_strerror(NULL) );
        WARN( "sf_error(NULL)="<<err );
        // FIXME distinguish "read errror" from "codec error"
        throw file_error( _("codec  error") );
    }

    _nChannels = _sfinfo.channels;
    _frequency = _sfinfo.samplerate;

    setLength( (double)_sfinfo.frames / _frequency  );

    setState( state_open );

    _readBuffer = new ValRingBuffer<int16_t>( 
        (int)(_maxSec*_frequency*_nChannels)+4096 ); 
}

InSndfile::~InSndfile()
{
    sf_close( _sndfile );
}

Frame
InSndfile::getFrame( Time t )
{
    Frame f;

    /* since 'read' and 'decode' is a single step with libsndfile
     * there is not much to do here
     */

    /* audio doesn't define or change framelength */
    f.setLength( Global::renderFmt.getFrameLength() );
    if (  f.getLength()==0 )  
    {  /* this may happen if playing a single audio file FIXME */
        f.setLength( 0.1 ); 
    }

    f.setAudioContext( this );
    f.setPos( t );
    return f;
}

PropertyNode * 
InSndfile::getProperties( PropertyNode * p ) const
{
    if ( !p ) {
        p = new PropertyNode( "stream" );
    }

    // container format
    p->addAttr( "filename", getFileName() );
    p->addAttr( "duration", length().sec() );

    PropertyNode * container = p->addDaughter( "container" );
    container->addAttr( "format", getTypeStr() );

    // audio format
    PropertyNode * audio = container->addDaughter( "codec" );
    audio->addAttr( "type", "audio" );
    audio->addAttr( "name", "sndfile" );
    audio->addAttr( "format", getSubFormatStr() );
    audio->addAttr( "frequency" , _frequency );
    audio->addAttr( "channels"  , _nChannels );
    audio->addAttr( "bitspersample"  , 16 );
    audio->addAttr( "endianness", getEndiannessStr() );
    // FIXME
    audio->addAttr( "fps" , 11 );


    return p;
}


void 
InSndfile::decodeMetaData( Frame & f )
{
}

void 
InSndfile::decodeVideo( Frame & f, uint8_t * dest )
{
    WARN( "This is an audio file, i.e. a configuration problem!");
}

void 
InSndfile::decodeAudio( Frame & f, uint8_t * dest )
{
    /*  already decoded  */
    if ( f.getAudioIsDecoded() ) return;

    Buffer  *buffer;
    int16_t *audio;

    /* how much data do we need at least? */
    // FIXME: this way rounding errors will add up: FIXME 
    double wtd = f.getLength() * _frequency * _nChannels;
    int words_to_deliver = static_cast<int>( ceil( wtd ) );
    if ( words_to_deliver <= 1 ) return;

    /* allocate buffer for delivery */
    AudioBuffer * aBuf = f.getABuf();
    if ( dest == 0 ) {
        
        if ( !aBuf ) {
            /*  make a abuf if nobody else did already  */
            aBuf = new AudioBuffer();
            f.setABuf( aBuf );
        } else {
            aBuf->dropRef();
        }

        buffer = BufferPool::getNewBuffer( words_to_deliver*sizeof(int16_t) );
        audio = (int16_t*)buffer->buf();
        
    } else {
        WARN( "FIXME" );
        return;
    }

    /* check buffer fill status */
    double s = _readBuffer->getCurSize();
    double remainingSec = s / _nChannels / _frequency;
//    INFO( "before fill "<<PRINTV(s)<<PRINTV(remainingSec) );
    if ( remainingSec < _minSec ) refillBuffer();

    s = _readBuffer->getCurSize();
    remainingSec = s / _nChannels / _frequency;
//    INFO( "after fill  "<<PRINTV(s)<<PRINTV(remainingSec) );

    /* copy from readbuffer to deliverbuffer */
    int got_words = _readBuffer->getN( words_to_deliver, audio );
    if (got_words<words_to_deliver) 
    {
        INFO( "Buffer underrun");
    }

    s = _readBuffer->getCurSize();
    remainingSec = s / _nChannels / _frequency;
//    INFO( "after copy  "<<PRINTV(s)<<PRINTV(remainingSec) );

    /* setup frame */
    f.setRawABuf( buffer ); 
    f.getABuf()->getChannels()[0] = buffer;
    f.setAudioIsDecoded( true );

    if ( aBuf ) {
        aBuf->setNChannels( _nChannels );
        aBuf->setNSamples( got_words / _nChannels );
        aBuf->setInterleaved( true );
    }

    return;

}


void
InSndfile::refillBuffer()
{
#define BUFFER_LEN 4096    
    static int16_t tmpbuf[ BUFFER_LEN ];
    INFO( "refillBuffer" );

    static long max_bytes = (long)(_frequency * _maxSec * _nChannels 
        * sizeof(int16_t) - 4096);
    long bytes_in_buffer = _readBuffer->getCurSize() * sizeof(int16_t);
    long bytes_to_read = max_bytes - bytes_in_buffer;

    INFO(PRINTV(bytes_to_read));
    long bytes_delivered = 0;
    while ( bytes_delivered < bytes_to_read )
    {
        long nSamples = sf_read_short( _sndfile, (short int*)tmpbuf, BUFFER_LEN );

//        INFO(PRINTV(nSamples)<<PRINTV(bytes_delivered));
        if (nSamples == 0) {
            /* EOF */
            WARN( "trying to read beyond end of file" );
            return;
        } else if (nSamples < 0) {
            /* error in the stream. */
            WARN( "error in the Ogg/Vorbis stream" );
            return;
        }
        int ok = _readBuffer->putN( nSamples, tmpbuf );
        if (ok<=0) 
        {
            WARN( "Buffer overflow" );
        }
        bytes_delivered += nSamples * sizeof(int16_t);
    }
}

string
InSndfile::getTypeStr() const
{
    int type = _sfinfo.format & SF_FORMAT_TYPEMASK;
    switch ( type )
    {
      case SF_FORMAT_WAV   : return string( "Microsoft WAV format (little endian)." );
      case SF_FORMAT_AIFF  : return string( "Apple/SGI AIFF format (big endian)." );
      case SF_FORMAT_AU    : return string( "Sun/NeXT AU format (big endian)." );
      case SF_FORMAT_RAW   : return string( "RAW PCM data." );
      case SF_FORMAT_PAF   : return string( "Ensoniq PARIS file format." );
      case SF_FORMAT_SVX   : return string( "Amiga IFF / SVX8 / SV16 format." );
      case SF_FORMAT_NIST  : return string( "Sphere NIST format." );
      case SF_FORMAT_VOC   : return string( "VOC files." );
      case SF_FORMAT_IRCAM : return string( "Berkeley/IRCAM/CARL" );
      case SF_FORMAT_W64   : return string( "Sonic Foundry's 64 bit RIFF/WAV" );
      case SF_FORMAT_MAT4  : return string( "Matlab (tm) V4.2 / GNU Octave 2.0" );
      case SF_FORMAT_MAT5  : return string( "Matlab (tm) V5.0 / GNU Octave 2.1" );
      default: break;
    }	
    return string( "Unknown type" );
}


string
InSndfile::getSubFormatStr() const
{
    int sub = _sfinfo.format & SF_FORMAT_SUBMASK;
    switch ( sub )
    {
       case SF_FORMAT_PCM_S8	:return string( "Signed 8 bit data");
       case SF_FORMAT_PCM_16	:return string( "Signed 16 bit data");
       case SF_FORMAT_PCM_24	:return string( "Signed 24 bit data");
       case SF_FORMAT_PCM_32	:return string( "Signed 32 bit data");
       case SF_FORMAT_PCM_U8	:return string( "Unsigned 8 bit data (WAV and RAW only)");
       case SF_FORMAT_FLOAT		:return string( "32 bit float data");
       case SF_FORMAT_DOUBLE	:return string( "64 bit float data");
       case SF_FORMAT_ULAW		:return string( "U-Law encoded.");
       case SF_FORMAT_ALAW		:return string( "A-Law encoded.");
       case SF_FORMAT_IMA_ADPCM	:return string( "IMA ADPCM.");
       case SF_FORMAT_MS_ADPCM	:return string( "Microsoft ADPCM.");
       case SF_FORMAT_GSM610	:return string( "GSM 6.10 encoding.");
       case SF_FORMAT_VOX_ADPCM	:return string( "OKI / Dialogix ADPCM");
       case SF_FORMAT_G721_32	:return string( "32kbs G721 ADPCM encoding.");
       case SF_FORMAT_G723_24	:return string( "24kbs G723 ADPCM encoding.");
       case SF_FORMAT_G723_40	:return string( "40kbs G723 ADPCM encoding.");
       case SF_FORMAT_DWVW_12	:return string( "12 bit Delta Width Variable Word encoding.");
       case SF_FORMAT_DWVW_16	:return string( "16 bit Delta Width Variable Word encoding.");
       case SF_FORMAT_DWVW_24	:return string( "24 bit Delta Width Variable Word encoding.");
       case SF_FORMAT_DWVW_N	:return string( "N bit Delta Width Variable Word encoding.");
      default: break;
    }
    return string( "Unknown submask" );
}


string
InSndfile::getEndiannessStr() const
{
    int end = _sfinfo.format & SF_FORMAT_ENDMASK;
    switch ( end )
    {
           case SF_ENDIAN_FILE	:	return string( "Default file endian-ness.");
           case SF_ENDIAN_LITTLE:	return string( "Force little endian-ness.");
           case SF_ENDIAN_BIG	:	return string( "Force big endian-ness.");
           case SF_ENDIAN_CPU	:	return string( "Force CPU endian-ness.");
      default: break;
    }
    return string( "Unknown submask" );
}


