/*
 * PIAVE - PIAVE Is A Video Editor
 *
 * Copyright (C) 2002 Rolf Dubitzky, rolf@dubitzky.de
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**
 * \file vorbisfile.cc
 */

// config
#if defined( HAVE_CONFIG_H )
#    include "config.h"
#endif


// std c
#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>

// std c++
#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <string>
using std::string;
#include <math.h>
// this header
#include "vorbisfile.hh"

// PIAVE
#include "libpiave/frame.hh"
#include "libpiave/property.hh"

using namespace PIAVE;

PIAVE_PLUGIN( VorbisfilePlugin );


/*
 *    wrapper plugin
 */
VorbisfilePlugin::VorbisfilePlugin() :
  Plugin( "vorbisfile", PLUGIN_INSTREAM )
{
    PropertyNode * caps = new PropertyNode( "instream" );
    caps->addAttr( "name", getName() );

    PropertyNode * file = caps->addDaughter( "file" );

    PropertyNode * container = file->addDaughter( "container" );
    container->addAttr( "format", "vorbis I" );
    container->addAttr( "extension", ".ogg" );

    PropertyNode * video = container->addDaughter( "codec" );
    video->addAttr( "type","audio" );
    video->addDaughter( "vorbis" );

    PropertyNode * about = caps->addDaughter( "about" );
    about->setContent( _("Read simple Ogg/Vorbis audio files.") );

    setCapabilities( caps );
}


PluginBase * 
VorbisfilePlugin::newElement( const PropertyNode * p )
{
    if ( p ) {
        string filename;
        bool ok = p->getAttr( "filename", filename );
        if ( ok ) {
            return new InVorbisfile( filename );;
        }
    }

    WARN( "missing filename" );
    return 0;
}






/**
 * \class InVorbisfile
 * use libvorbisfile to read and decode Ogg/Vorbis I files.
 *
 */
InVorbisfile::InVorbisfile( const std::string& fileName ) throw(exception)
  : InAVFStream( fileName ),
    _file( 0 ),
    _vi( 0 ),
    _vc( 0 ),
    _current_section( 0 ),
    _nChannels( 0 ),
    _frequency( 0 ),
    _minSec( 1 ),
    _maxSec( 3 )
{
    _file = fdopen( _fd, "r" );
    if ( _file == 0 ) {
        WARN( "could not open buffered stream on file "<<getFileName() );
        setState( state_error );
        throw file_error( _("read error") );
    }
    

    if ( ov_open( _file, &_vf, NULL, 0) < 0) {
        WARN("Input does not appear to be an Ogg bitstream in file "<<getFileName() );
        setState( state_error );
        throw codec_error( _("not an Ogg/Vorbis file") );
    }
    
    /* get informatrion about the file content */
    _vi = ov_info( &_vf , -1 );
    _vc = ov_comment( &_vf, -1 );
    long pcm_total = ov_pcm_total( &_vf, -1 );

    _nChannels = _vi->channels;
    _frequency = _vi->rate;

    setLength( (double)pcm_total/_frequency );

    setState( state_open );

    _readBuffer = new ValRingBuffer<int16_t>( 
        (int)(_maxSec*_frequency*_nChannels)+4096 ); 
}

InVorbisfile::~InVorbisfile()
{
}

Frame
InVorbisfile::getFrame( Time t )
{
    Frame f;

    /* since 'read' and 'decode' is a single step with libvorbisfile
     * there is not much to do here
     */

    /* audio doesn't define or change framelength */
    f.setLength( Global::renderFmt.getFrameLength() );
    if (  f.getLength()==0 )  
    {  /* this may happen if playing a single audio file FIXME */
        f.setLength( 0.1 ); 
    }

    f.setAudioContext( this );
    f.setPos( t );
    return f;
}

PropertyNode * 
InVorbisfile::getProperties( PropertyNode * p ) const
{
    if ( !p ) {
        p = new PropertyNode( "stream" );
    }

    // container format
    p->addAttr( "filename", getFileName() );
    p->addAttr( "duration", length().sec() );

    PropertyNode * container = p->addDaughter( "container" );
    container->addAttr( "format", "Ogg" );

    // audio format
    PropertyNode * audio = container->addDaughter( "codec" );
    audio->addAttr( "type", "audio" );
    audio->addAttr( "name", "vorbisfile" );
    audio->addAttr( "format", "Vorbis I" );
    audio->addAttr( "frequency" , _frequency );
    audio->addAttr( "channels"  , _nChannels );
    audio->addAttr( "bitspersample"  , 16 );
    // FIXME
    audio->addAttr( "fps" , 11 );

    /* parse comment fields */
    if ( _vc )
    {
        audio->addAttr( "vendor", _vc->vendor );
        char **user_comment = _vc->user_comments;
        int i = 0;
        while ( *user_comment )
        {
            sprintf( Global::cbuf, "user_comment %d",i++);
            audio->addAttr( Global::cbuf, *user_comment );
            user_comment++;
        }
    }

    return p;
}


void 
InVorbisfile::decodeMetaData( Frame & f )
{
}

void 
InVorbisfile::decodeVideo( Frame & f, uint8_t * dest )
{
    WARN( "This is an audio file, i.e. a configuration problem!");
}

void 
InVorbisfile::decodeAudio( Frame & f, uint8_t * dest )
{
    /*  already decoded  */
    if ( f.getAudioIsDecoded() ) return;

    Buffer  *buffer;
    int16_t *audio;

    /* how much data do we need at least? */
    // FIXME: this way rounding errors will add up: FIXME 
    double wtd = f.getLength() * _frequency * _nChannels;
    int words_to_deliver = static_cast<int>( ceil( wtd ) );
    if ( words_to_deliver <= 1 ) return;

    /* allocate buffer for delivery */
    AudioBuffer * aBuf = f.getABuf();
    if ( dest == 0 ) {
        
        if ( !aBuf ) {
            /*  make a abuf if nobody else did already  */
            aBuf = new AudioBuffer();
            f.setABuf( aBuf );
        } else {
            aBuf->dropRef();
        }

        buffer = BufferPool::getNewBuffer( words_to_deliver*sizeof(int16_t) );
        audio = (int16_t*)buffer->buf();
        
    } else {
        WARN( "FIXME" );
        return;
    }

    /* check buffer fill status */
    double s = _readBuffer->getCurSize();
    double remainingSec = s / _nChannels / _frequency;
//    INFO( "before fill "<<PRINTV(s)<<PRINTV(remainingSec) );
    if ( remainingSec < _minSec ) refillBuffer();

    s = _readBuffer->getCurSize();
    remainingSec = s / _nChannels / _frequency;
//    INFO( "after fill  "<<PRINTV(s)<<PRINTV(remainingSec) );

    /* copy from readbuffer to deliverbuffer */
    int got_words = _readBuffer->getN( words_to_deliver, audio );
    if (got_words<words_to_deliver) 
    {
        INFO( "Buffer underrun");
    }

    s = _readBuffer->getCurSize();
    remainingSec = s / _nChannels / _frequency;
//    INFO( "after copy  "<<PRINTV(s)<<PRINTV(remainingSec) );

    /* setup frame */
    f.setRawABuf( buffer ); 
    f.getABuf()->getChannels()[0] = buffer;
    f.setAudioIsDecoded( true );

    if ( aBuf ) {
        aBuf->setNChannels( _nChannels );
        aBuf->setNSamples( got_words / _nChannels );
        aBuf->setInterleaved( true );
    }

    return;

}


void 
InVorbisfile::refillBuffer()
{
    static char tmpbuf[4096];
    INFO( "refillBuffer" );

    static long max_bytes = (long)(_frequency * _maxSec * _nChannels 
        * sizeof(int16_t) - 4096);
    long bytes_in_buffer = _readBuffer->getCurSize() * sizeof(int16_t);
    long bytes_to_read = max_bytes - bytes_in_buffer;

    INFO(PRINTV(bytes_to_read));
    long bytes_delivered = 0;
    while ( bytes_delivered < bytes_to_read )
    {
        long nSamples = 
            ov_read( &_vf, (char*)tmpbuf, 4096, 0, 2, 1, &_current_section);
//        INFO(PRINTV(nSamples)<<PRINTV(bytes_delivered));
        if (nSamples == 0) {
            /* EOF */
            WARN( "trying to read beyond end of file" );
            return;
        } else if (nSamples < 0) {
            /* error in the stream. */
            WARN( "error in the Ogg/Vorbis stream" );
            return;
        }
        int ok = _readBuffer->putN( nSamples/sizeof(int16_t), (int16_t*)tmpbuf );
        if (ok<=0) 
        {
            WARN( "Buffer overflow" );
        }
        bytes_delivered += nSamples;
    }
}
